/*
XLiFE++ is an extended library of finite elements written in C++
    Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file ShapeValues.hpp
  \authors D. Martin, N. Kielbasiewicz
  \since 01 jul 2005
  \date 10 may 2012

  \brief Definition of the xlifepp::ShapeValues class

  This class is used to create storage for finite element computation such as
    - Shape values at a given point (e.g. quadrature point)
    - Derivatives of Shape functions values stored in a vector<vector<real_t> dw as
      dw[0][i] stands for dx(wi), dw[1][i] stands for dy(wi) and dw[2][i] stands for dz(wi)
    - Mapping shape functions derivatives from Reference Element onto current element
*/

#ifndef SHAPE_VALUES_HPP
#define SHAPE_VALUES_HPP

#include "config.h"
#include "../geometry/GeomMapData.hpp"

namespace xlifepp
{

class RefElement;
class GeomMapData;

/*!
  \class ShapeValues
 */
class ShapeValues
{
  public:
    std::vector<real_t> w;                 //!< shape functions at a given point
    std::vector< std::vector<real_t> > dw; //!< derivatives of shape functions at a given point

    //--------------------------------------------------------------------------------
    // Constructors and Destructor
    //--------------------------------------------------------------------------------
    ShapeValues();                                 //!< default constructor
    ~ShapeValues();                                //!< destructor
    ShapeValues(const ShapeValues&);               //!< copy constructor
    ShapeValues& operator=(const ShapeValues&);    //!< assignment operator=
    ShapeValues(const RefElement&);                //!< constructor with associated ReferenceElement characteristics
    ShapeValues(const RefElement&, const dimen_t); //!< constructor with associated ReferenceElement characteristics and space dimension

    //--------------------------------------------------------------------------------
    // Public member functions
    //--------------------------------------------------------------------------------
    //! returns number of shape functions
    dimen_t dim() const { return static_cast<dimen_t>(dw.size()); }
    //! returns number of dofs
    number_t nbDofs() const { return static_cast<number_t>(w.size());}
    //! returns true if empty
    bool isEmpty() const {return w.size()==0;}

    void resize(const RefElement&, const dimen_t);      //!< defines shape functions size
    void resize(const number_t, const dimen_t);         //!< resize shape value vectors
    void set(const real_t);                             //!< set all shape functions to const value
    void extendToVector(dimen_t d);                     //!< extend scalar shape function values to vector shape function values
    void assign(const ShapeValues&);                    //!< fast assign of shapevalues into current shapevalues

    //shapevalues mapping
    void map(const ShapeValues&, const GeomMapData&, number_t der=1);                   //!< maps Reference Element shape functions onto current Geometric Element
    void contravariantPiolaMap(const ShapeValues&, const GeomMapData&, number_t der=1); //!< maps Reference Element shape functions onto current Geometric Element (contravariant Piola)
    void covariantPiolaMap(const ShapeValues&, const GeomMapData&, number_t der=1);     //!< maps Reference Element shape functions onto current Geometric Element (covariant Piola)
    void changeSign(const std::vector<real_t>&, dimen_t, number_t der=1);               //!< change sign of shape functions according to a sign vector

    friend std::ostream& operator<<(std::ostream&, const ShapeValues&);

}; // end of class ShapeValues

std::ostream& operator<<(std::ostream&, const ShapeValues&); //!< print operator

// fast assign of shape values into current shape values
inline void ShapeValues::assign(const ShapeValues& shv)
{
  const std::vector<real_t>& wc=shv.w;
  if(w.size()==wc.size()) //same size
  {
    std::vector<real_t>::iterator it=w.begin();
    std::vector<real_t>::const_iterator itc=wc.begin();
    for(;it!=w.end() && itc!=wc.end();++it, ++itc) *it=*itc;
    const std::vector< std::vector<real_t> >& dwc=shv.dw;
    std::vector< std::vector<real_t> >::iterator itd=dw.begin();
    std::vector< std::vector<real_t> >::const_iterator itdc=dwc.begin();
    for(;itd!=dw.end() && itdc!=dwc.end();++itd, ++itdc)
    {
        it=itd->begin(); itc=itdc->begin();
        for(;it!=itd->end();++it,++itc) *it=*itc;
    }
    return;
  }
  //usual copy
  w=shv.w;
  dw=shv.dw;
  return;
}

} // end of namespace xlifepp

#endif /* SHAPE_VALUES_HPP */
