/*
XLiFE++ is an extended library of finite elements written in C++
    Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file BilinearForm.hpp
  \author E. Lunéville
  \since 23 mar 2012
  \date 9 may 2012

  \brief Definition of the xlifepp::BilinearForm classes

  This file declares all the classes related to bilinear forms management.
  The more general bilinear form under consideration is a list of linear combinations of basic bilinear forms.
  The list is indexed by the unknowns of the linear combination.

  Basic bilinear forms are, for the moment, integral or double integral type.
  The bilinear form may have multiple unknowns but a linear combination is related to a unique unknown

  - xlifepp::BiLinearForm is the end user class, it is a map of <Unknown*,LCBilinearForm>
  - xlifepp::SuBilinearForm deals with linear combination of basic bilinear forms, it is a vector of pairs of <BasicBilinearForm*,complex>
  - xlifepp::BasicBilinearForm is the mother class (abstract) of classes
    - xlifepp::IntgBilinearForm for single integral bilinear form (generally in FEM method)
    - xlifepp::DoubleIngBilinearForm for double integral bilinear form (generally in BEM method)
    - xlifepp::ComposedBilinearForm for composition of bilinear forms (generally DtN operator)
    - xlifepp::MatrixBilinearForm for explicit bilinear form given by a TermMatrix

  As example :
  \code
  BilinearForm l=intg(Omega,f*u*v);                               //a single unknown (u) linear form
  BilinearForm l=intg(Omega,f*u*v)+2*intg(Omega,grad(u)|grad(w)); //a multiple unknowns (u,v), (u,w) bilinear form
  \endcode

  All linear algebraic operations are allowed on xlifepp::BilinearForm objects

  It is possible to impose the quadrature rule to an integral linear form by specifying a quadrature rule type and its order
  \code
  BilinearForm l=intg(Omega,f*u*v,Gauss_Legendre,5);                     //a single unknown (u) linear form
  \endcode
  see Quadrature class for the list of quadrature rules available
*/

#ifndef BILINEAR_FORM_HPP
#define BILINEAR_FORM_HPP

#include "config.h"
#include "space.h"
#include "operator.h"
#include "finiteElements.h"
#include "largeMatrix.h"

#include <utility>

namespace xlifepp
{

class IntgBilinearForm;
class DoubleIntgBilinearForm;
class ComposedBilinearForm;
class MatrixBilinearForm;

//=========================================================================================
/*!
  \class BasicBilinearForm
  describes the base class of bilinear form on a space.
  It is an abstract class
*/
//=========================================================================================
class BasicBilinearForm
{
protected:
  const Unknown* u_p;              //!< pointer to unknown u
  const Unknown* v_p;              //!< pointer to unknown v (test function)
  const GeomDomain* domainu_p;     //!< geometric domain of unknown u
  const GeomDomain* domainv_p;     //!< geometric domain of unknown v (test function)
  ComputationType compuType;       //!< computation type (_FEComputation, _IEComputation, _SPComputation, ...)
public:
  SymType symmetry;                //!< symmetry of the bilinear form ((_noSymmetry , _symmetric, _skewSymmetric, _selfAdjoint, _skewAdjoint)

public:
  //constructor/destructor
  BasicBilinearForm(const Unknown* up=0, const GeomDomain* dup=0,
                    const Unknown* vp=0, const GeomDomain* dvp=0) //! basic and default constructor
  :u_p(up), v_p(vp), domainu_p(dup), domainv_p(dvp), compuType(_undefComputation), symmetry(_undefSymmetry) {}
  virtual ~BasicBilinearForm() {}  //!< virtual destructor

  //accessors
  const Unknown& up() const        //! return the unknown
  {return *u_p;}
  const Unknown& vp() const        //! return the test function
  {return *v_p;}
  const GeomDomain& dom_up() const //! return the unknown domain
  {return *domainu_p;}
  const GeomDomain& dom_vp() const //! return the test function domain
  {return *domainv_p;}
  ComputationType computationType() const    //! return the computation type
  {return compuType;}

  Space* subSpace_up();   //!< return the u subspace of the bilinear form, construct it if does not exist
  Space* subSpace_vp();   //!< return the v subspace of the bilinear form, construct it if does not exist

  void checkUnknowns() const;   //!< check the Unknown/TestFunction status of u_p and v_p

  // virtual member functions
  virtual BasicBilinearForm* clone() const = 0; //!< clone of the bilinear form
  virtual LinearFormType type() const = 0;      //!< return the type of the bilinear form
  virtual ValueType valueType() const = 0;      //!< return the value type of the bilinear form
  IntgBilinearForm* asIntgForm();               //!< downcast as IntgBilinearForm*
  const IntgBilinearForm* asIntgForm() const;   //!< downcast as const IntgBilinearForm*
  DoubleIntgBilinearForm* asDoubleIntgForm();   //!< downcast as DoubleIntgBilinearForm*
  const DoubleIntgBilinearForm* asDoubleIntgForm() const; //!< downcast as const DoubleIntgBilinearForm*
  //virtual SymType symType() const =0;           //!< find symmetry property
  virtual void print(std::ostream&) const = 0;   //!< print utility
  virtual void print(PrintStream&) const = 0;
  virtual string_t asString() const = 0;         //!< interpret as string for print purpose
  virtual void setUnknowns(const Unknown&u, const Unknown&v) //! set (change) the unknowns
  {u_p=&u; v_p=&v;}
};

/*!
  \class IntgBilinearForm
  describes a bilinear form based on a single integral

       intg_domain opu aopu opv

     with opu, opv   : operator on u and v
          aopu, aopv : algebraic operation (*,|,%,^)
*/
class IntgBilinearForm : public BasicBilinearForm
{
protected :
  const OperatorOnUnknowns* opus_p;           //!< pointer to the operator on unknowns
  const IntegrationMethod* intgMethod_p;      //!< pointer to integration method

public :
  IntgBilinearForm(const GeomDomain&, const OperatorOnUnknowns&, const IntegrationMethod&, SymType);  //!< constructor from OperatorOnUnknowns
  IntgBilinearForm(const GeomDomain&, const OperatorOnUnknowns&, QuadRule, number_t, SymType );       //!< constructor from OperatorOnUnknowns
  IntgBilinearForm(const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,
                   const OperatorOnUnknown&, const IntegrationMethod&, SymType);                      //!< constructor from explicit operators
  IntgBilinearForm(const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,
                   const OperatorOnUnknown&, QuadRule, number_t, SymType);                            //!< constructor from explicit operators
  IntgBilinearForm(const IntgBilinearForm&);                                                          //!< copy constructor
  ~IntgBilinearForm();                                                                                //!< destructor
  IntgBilinearForm& operator=(const IntgBilinearForm&);                                               //!< assign operator

  void setIntegrationMethod(const GeomDomain&, const OperatorOnUnknowns&, QuadRule , number_t);       //!< set integration method
  void setComputationType();                //!< set computation type (_FEComputation,_SPComputation, _FESPComputation)

  const OperatorOnUnknown& opu() const      //! return the reference to the operator on unknown
  {return opus_p->opu();}
  const OperatorOnUnknown& opv() const      //! return the reference to the operator on unknown
  {return opus_p->opv();}
  const OperatorOnUnknown* opu_p() const    //! return the pointer to the operator on unknown
  {return opus_p->opu_p();}
  const OperatorOnUnknown* opv_p() const    //! return the pointer to the operator on unknown
  {return opus_p->opv_p();}
  AlgebraicOperator algop()const            //! return the algebraic operator
  {return opus_p->algop();}
  const GeomDomain* domain() const          //! return the pointer to the domain
  {return domainu_p;}
  virtual LinearFormType type() const       //! return the type of the linear form
  {return _intg;}
  virtual ValueType valueType() const       //! return the value type of the linear form
  {return opus_p->valueType();}
  const IntegrationMethod* intgMethod() const //! pointer to integration method object
  {return intgMethod_p;}
  SymType symType() const;                    //!< find symmetry property (used to set symmetry)
  virtual BasicBilinearForm* clone() const;   //!< clone of the linear form
  virtual void setUnknowns(const Unknown&, const Unknown&); //!< set (change) the unknowns

  //print stuff
  virtual void print(std::ostream&) const;   //!< print utility
  virtual void print(PrintStream& os) const {print(os.currentStream());}
  virtual string_t asString() const;         //!< interpret as string for print purpose
};

/*!
  \class DoubleIntgBilinearForm
  describes a bilinear form based on a double integral

     intg_domain_v  intg_domain_u  opu(y) aopu opker(x,y) aopv opv(x) dy dx

     with opu, opv   : operator on u and v
          aopu, aopv : algebraic operation (*,|,%,^)
          opker : operator on a kernel function

    note : when kernel = 1 the pointer to kernel function in opker is ker_p=0 !
    Be cautious with order of domain, first if for v-domain, second for u-domain

    Actually, two ways to handle some integration methods :
      by pointer to an IntegrationMethod object, so only one integration method (old way, should disappear in futur)
      by an IntegrationMethods object that handles some IntegrationMethod (new way)
*/
class DoubleIntgBilinearForm : public BasicBilinearForm
{
protected:
  const KernelOperatorOnUnknowns* kopus_p; //!< pointer to the operator on unknowns with kernel
  const IntegrationMethod* intgMethod_p;   //!< pointer to an integration method

public:
  IntegrationMethods intgMethods;          //!< list of integration methods

public:
  //copy constructor
  DoubleIntgBilinearForm(const DoubleIntgBilinearForm&);            //!< copy constructor
  DoubleIntgBilinearForm& operator=(const DoubleIntgBilinearForm&); //!< assign operator
  ~DoubleIntgBilinearForm();                                        //!< destructor

  //constructors from IntegrationMethod
  DoubleIntgBilinearForm(const GeomDomain&, const GeomDomain&, const KernelOperatorOnUnknowns&,
                         const IntegrationMethod&, SymType);      //!< basic constructor

  DoubleIntgBilinearForm(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator, const OperatorOnUnknown&,
                         const IntegrationMethod&, SymType);     //!< constructor from unknown operators (no kernel)

  DoubleIntgBilinearForm(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,const OperatorOnKernel&,
                         AlgebraicOperator, const OperatorOnUnknown&,
                         const IntegrationMethod&, SymType);     //!< constructor from unknown operators and operator on kernel

  DoubleIntgBilinearForm(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,const Kernel&,
                         AlgebraicOperator, const OperatorOnUnknown&,
                         const IntegrationMethod&, SymType);    //!< constructor from unknown operators and kernel

  //constructors from IntegrationMethods
  DoubleIntgBilinearForm(const GeomDomain&, const GeomDomain&, const KernelOperatorOnUnknowns&,
                         const IntegrationMethods&, SymType);      //!< basic constructor

  DoubleIntgBilinearForm(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator, const OperatorOnUnknown&,
                         const IntegrationMethods&, SymType);     //!< constructor from unknown operators (no kernel)

  DoubleIntgBilinearForm(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,const OperatorOnKernel&,
                         AlgebraicOperator, const OperatorOnUnknown&,
                         const IntegrationMethods&, SymType);     //!< constructor from unknown operators and operator on kernel

  DoubleIntgBilinearForm(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,const Kernel&,
                         AlgebraicOperator, const OperatorOnUnknown&,
                         const IntegrationMethods&, SymType);    //!< constructor from unknown operators and kernel

  //constructors from Quadrule's
  DoubleIntgBilinearForm(const GeomDomain&, const GeomDomain&, const KernelOperatorOnUnknowns&,
                         QuadRule, number_t, QuadRule, number_t, SymType);   //!< basic constructor

  DoubleIntgBilinearForm(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator, const OperatorOnUnknown&,
                         QuadRule, number_t, QuadRule, number_t, SymType);   //!< constructor from unknown operators (no kernel)

  DoubleIntgBilinearForm(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,const OperatorOnKernel&,
                         AlgebraicOperator, const OperatorOnUnknown&,
                         QuadRule, number_t, QuadRule, number_t, SymType);   //!< constructor from unknown operators and operator on kernel

  DoubleIntgBilinearForm(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,const Kernel&,
                         AlgebraicOperator, const OperatorOnUnknown&,
                         QuadRule, number_t, QuadRule, number_t, SymType);   //!< constructor from unknown operators and kernel//accessors

  void setIntegrationMethod(const GeomDomain&, const GeomDomain&, const KernelOperatorOnUnknowns&,
                            QuadRule, number_t, QuadRule, number_t);   //!< set IntegrationMethod for quadrature product rule
  void setComputationType();      //!< set computation type (_IEComputation,_IEExtComputation,IEHMatrixComputation, ..)
  void setIntegrationMethods();   //!< check and update quadrature pointers of IntegrationsMethods
  void setHMIntegrationMethods(); //!< check and update quadrature pointers of IntegrationsMethods when a HMatrixIM

  //accessors
  const KernelOperatorOnUnknowns& kopus()const //! return KernelOperatorOnUnknowns reference (const)
    {return *kopus_p;}
  const KernelOperatorOnUnknowns* kopusp()const //! return KernelOperatorOnUnknowns pointer (const)
    {return kopus_p;}
  const OperatorOnUnknown& opu() const      //! return the pointer to the operator on unknown
    {return kopus_p->opu();}
  const OperatorOnUnknown& opv() const      //! return the pointer to the operator on unknown
    {return kopus_p->opv();}
  const OperatorOnKernel& opker() const     //! return the pointer to the operator on kernel
    {return kopus_p->opker();}
  AlgebraicOperator algopu()const           //! return the algebraic operator on u
    {return kopus_p->algopu();}
  AlgebraicOperator algopv()const           //! return the algebraic operator on v
    {return kopus_p->algopv();}
  const IntegrationMethod* intgMethod() const //! return the pointer to the integration method
    {return intgMethod_p;}
  const GeomDomain* domainx() const         //! return the pointer to the x domain, say u-domain
    {return domainv_p;}
  const GeomDomain* domainy() const         //! return the pointer to the y domain, say v-domain
    {return domainu_p;}
  DomainPair domains() const                //! return the pair of domain pointers
    {return DomainPair(domainu_p,domainv_p);}
  virtual LinearFormType type() const       //! return the type of the linear form
    {return _doubleIntg;}
  virtual ValueType valueType() const       //! return the value type of the bilinear form
    {return kopus_p->valueType();}
  SymType symType() const;                  //!< return symmetry property (_noSymmetry , _symmetric, _skewSymmetric, _selfAdjoint, _skewAdjoint)
  virtual BasicBilinearForm* clone() const; //!< clone of the linear form
  virtual void setUnknowns(const Unknown&, const Unknown&); //!< set (change) the unknowns

  //print tools
  string_t asString() const;                //!< interpret as string for print purpose
  virtual void print(std::ostream&) const;  //!< print utility
  virtual void print(PrintStream& os) const {print(os.currentStream());}
  friend std::ostream& operator<<(std::ostream& os,const DoubleIntgBilinearForm& dibf)
  {dibf.print(os);return os;} //!< print operator
};

//@{
//! useful typedefs to MatrixBilinearForm class
typedef std::pair<BasicBilinearForm*, complex_t> blfPair;
typedef std::vector<blfPair>::iterator it_vblfp;
typedef std::vector<blfPair>::const_iterator cit_vblfp;
//@}

/*!
  \class MatrixBilinearForm
  describes a bilinear forms given by an explicit matrix
  explicit matrix is handled by a MatrixEntry object
  MatrixEntry pointer may be assigned from :
    - MatrixEntry pointer or reference : no hard copy to save memory
    - Matrix : matrix is copied as a dense matrix MatrixEntry
    - Vector : vector is copied as a diagonal MatrixEntry
  when hard copy is involved, newMatrixEntry flag is set to true in order to free memory when deleting MatrixBilinearForm
*/
class MatrixBilinearForm : public BasicBilinearForm
{
  protected :
    const MatrixEntry* matrix_p;  //!< pointer to a MatrixEntry
    bool newMatrixEntry_;         //!< useful flag to clean memory

  public :
    //constructors/destructor
    MatrixBilinearForm()                                                           //! default constructor
    : BasicBilinearForm(0,0), matrix_p(0), newMatrixEntry_(false) {}
    MatrixBilinearForm(const Unknown& u, const GeomDomain& du,
                       const Unknown& v, const GeomDomain& dv, const MatrixEntry* mp)  //! basic constructor
    : BasicBilinearForm(&u, &du, &v, &dv), matrix_p(mp), newMatrixEntry_(false) {}
    MatrixBilinearForm(const Unknown& u, const GeomDomain& du,
                       const Unknown& v, const GeomDomain& dv, const MatrixEntry & m)  //! basic constructor
    : BasicBilinearForm(&u, &du, &v, &dv), matrix_p(&m), newMatrixEntry_(false) {}
    template <typename T>
    MatrixBilinearForm(const Unknown&, const GeomDomain&,
                       const Unknown&, const GeomDomain&,
                       const Matrix<T>& , SymType=_noSymmetry);                    //!< constructor from a Matrix<T>
    template <typename T>
    MatrixBilinearForm(const Unknown&, const GeomDomain&,
                       const Matrix<T>& , SymType=_noSymmetry);                    //!< constructor from a Matrix<T>
    template <typename T>
    MatrixBilinearForm(const Unknown&, const GeomDomain&,const std::vector<T>&);   //!< constructor from a diagonal matrix(vector)
    ~MatrixBilinearForm();                                                         //!< destructor

    virtual LinearFormType type() const       //! return the type of the linear form
    {return _explicitLf;}
};

//-------------------------------------------------------------------------------
// External functions of MatrixBiLinearForm class
//-------------------------------------------------------------------------------
//constructor from a Matrix<T>
template <typename T>
MatrixBilinearForm::MatrixBilinearForm(const Unknown& u, const GeomDomain& du, const Matrix<T>& m, SymType sy)
: BasicBilinearForm(&u, &du, &u, &du)
{
  //create dense storage
  MatrixStorage* sto= new RowDenseStorage(m.numberOfRows(), m.numberOfColumns(),"");
  matrix_p= new MatrixEntry(m.valueType(), m.strucType(), sto, m.elementSize(), sy);
  newMatrixEntry_=true;
}

//constructor from a Matrix<T>
template <typename T>
MatrixBilinearForm::MatrixBilinearForm(const Unknown& u, const GeomDomain& du,
                         const Unknown& v, const GeomDomain& dv, const Matrix<T>& m, SymType sy)
: BasicBilinearForm(&u, &du, &v, &dv)
{
  //create dense storage
  MatrixStorage* sto= new RowDenseStorage(m.numberOfRows(), m.numberOfColumns(),"");
  matrix_p= new MatrixEntry(m.valueType(), m.strucType(), sto, m.elementSize(), sy);
  matrix_p->add(m.begin(),m.end()); //fill matrix
  newMatrixEntry_=true;
}

template <typename T>
MatrixBilinearForm::MatrixBilinearForm(const Unknown& u, const GeomDomain& d, const std::vector<T>& v)
{
  //create csr storage
  MatrixStorage* sto= new RowCsStorage(v.size(), v.size(),"");
  matrix_p= new MatrixEntry(v.valueType(), v.strucType(), sto, v.elementSize());
  matrix_p->add(v.begin(),v.end()); //fill matrix
  newMatrixEntry_=true;
}

/*!
  \class SuBilinearForm
  describes a linear combination of bilinear forms
  with the same pair of Unknowns.
*/
class SuBilinearForm
{
protected:
  std::vector<blfPair> blfs_;                       //!< list of pairs of bilinear form and coefficient
public:
  SuBilinearForm() {}                               //!< default constructor
  SuBilinearForm(const SuBilinearForm&);            //!< copy constructor
  SuBilinearForm(const std::vector<blfPair>&);      //!< full constructor from a list of BasicBilinearForms
  ~SuBilinearForm();                                //!< destructor
  SuBilinearForm& operator=(const SuBilinearForm&); //!< assignment operator

  number_t size() const                       //! number of elements of the linear combination
    {return blfs_.size();}
  std::vector<blfPair>& blfs()              //! access to the list of linear forms
    {return blfs_;}
  const std::vector<blfPair>& blfs() const  //! access to the list of linear forms
    {return blfs_;}
  blfPair& operator()(number_t n)             //! access to n-th pair
    {return blfs_[n - 1];}
  const blfPair& operator()(number_t n) const //! access to n-th pair
    {return blfs_[n - 1];}
  it_vblfp begin()                          //! iterator to the first element of vector of basic bilinear forms
    {return blfs_.begin();}
  cit_vblfp begin() const                   //! const iterator to the first element of vector of basic bilinear forms
    {return blfs_.begin();}
  it_vblfp end()                            //! iterator to the last+1 element of vector of basic bilinear forms
    {return blfs_.end();}
  cit_vblfp end() const                     //! const iterator to the last+1 element of vector of basic bilinear forms
    {return blfs_.end();}
  const Unknown* up() const;                //!< return the first unknown of the bilinear form
  const Unknown* vp() const;                //!< return the second unknown of the bilinear form
  const GeomDomain* dom_up() const;         //!< return the u-domain of the FIRST basic bilinear form
  const GeomDomain* dom_vp() const;         //!< return the v-domain of the FIRST basic bilinear form
  const Space* uSpace() const;              //!< return the u space of the bilinear form
  const Space* vSpace() const;              //!< return the v space of the bilinear form
  LinearFormType type() const;              //!< return the type of the bilinear form (intg,doubleIntg,linearCombination)
  ValueType valueType() const;              //!< return the value type of the bilinear form
  SymType symType() const;                  //!< return the symmetry of bilinear form
  void setUnknowns(const Unknown&, const Unknown&); //!< set unknowns of SuBilinearForm

  bool checkConsistancy(const SuBilinearForm&) const; //!< check consitancy of Unknown
  SuBilinearForm& operator +=(const SuBilinearForm&); //!< sum of bilinear forms
  SuBilinearForm& operator -=(const SuBilinearForm&); //!< difference of linear forms
  SuBilinearForm& operator *=(const complex_t&);      //!< multiply by a complex
  SuBilinearForm& operator /=(const complex_t&);      //!< divide by a complex

  void print(std::ostream&) const;                    //!< print utility
  string_t asString() const;                          //!< as symbolic string

}; // end of BilinearForm class

//-------------------------------------------------------------------------------
// External functions of SuBilinearForm class
//-------------------------------------------------------------------------------
std::ostream& operator<<(std::ostream&, const SuBilinearForm&); //!< output SuBilinearForm

// operations on bilinear form to construct linear combination of forms
SuBilinearForm operator-(const SuBilinearForm&);                        //!< opposite of bilinear form
SuBilinearForm operator+(const SuBilinearForm&, const SuBilinearForm&); //!< sum of bilinear forms
SuBilinearForm operator-(const SuBilinearForm&, const SuBilinearForm&); //!< difference of bilinear forms
SuBilinearForm operator*(const complex_t&, const SuBilinearForm&);        //!< multiply(left) by a scalar
SuBilinearForm operator*(const SuBilinearForm&, const complex_t&);        //!< multiply(right) by a scalar
SuBilinearForm operator/(const SuBilinearForm&, const complex_t&);        //!< divide by a scalar

//@{
//! useful typedefs to BilinearForm class
typedef std::pair<const Unknown*, const Unknown*> uvPair;
typedef std::map<uvPair, SuBilinearForm>::iterator it_mublc;
typedef std::map<uvPair, SuBilinearForm>::const_iterator cit_mublc;
//@}

/*!
  \class BilinearForm
  describes a general bilinear form, that is a list of linear combinations of basic bilinear forms.
  It is intend to collect linear forms with different unknowns, using a map<uvPair,SuBilinearForm>.
  It is the enduser's class
*/
class BilinearForm
{
protected:
  std::map<uvPair, SuBilinearForm> mlcblf_;              //!< list of linear combinations of basic bilinear forms

public:
  SuBilinearForm& operator[](const uvPair&);             //!< access to the bilinear form associated to an unknown
  const SuBilinearForm& operator[](const uvPair&) const; //!< access to the bilinear form associated to a pair of unknowns
  BilinearForm() {}                                                                  //!< default constructor
  BilinearForm(const SuBilinearForm& lc);                                            //!< constructor from a linear combination
  void clear()                                                                       //! clear the map of bilinear form
    {mlcblf_.clear();}
  bool isEmpty() const;                                                              //!< true if no bilinear form
  it_mublc begin()                                                                   //! iterator to the first element of map
    {return mlcblf_.begin();}
  cit_mublc begin() const                                                            //! const iterator to the first element of map
    {return mlcblf_.begin();}
  it_mublc end()                                                                     //! iterator to the last+1 element of map
    {return mlcblf_.end();}
  cit_mublc end() const                                                              //! const iterator to the last+1 element of map
    {return mlcblf_.end();}
  bool singleUnknown() const;                                                        //!< true if only one unknown
  const SuBilinearForm& first() const;                                               //!< return first linear combination
  BilinearForm operator()(const Unknown&, const Unknown&) const;                     //!< access to the bilinear form associated to a pair of unknowns
  BasicBilinearForm& operator()(const Unknown&, const Unknown&, number_t);             //!< access to the n-th basic bilinear form associated to a pair of unknowns
  const BasicBilinearForm& operator()(const Unknown&, const Unknown&, number_t) const; //!< access to the n-th basic bilinear form associated to a pair of unknowns
  const SuBilinearForm* subLfp(const uvPair& uv) const;                              //!< access to SuBiLinearForm pointers related to unknown
  SuBilinearForm* subLfp(const uvPair& uv);                                          //!< access to SuBiLinearForm pointers related to unknown (non const)
  void setUnknowns(const Unknown&, const Unknown&);                                  //!< set the unknowns of bilinear forms (only for single unknown bf)

  void print(std::ostream&) const;                                                   //!< print utility
  void print(PrintStream& os) const {print(os.currentStream());}

  BilinearForm& operator+=(const BilinearForm&); //!< sum of linear forms
  BilinearForm& operator-=(const BilinearForm&); //!< difference of linear forms
  BilinearForm& operator*=(const complex_t&);      //!< multiply by a complex
  BilinearForm& operator/=(const complex_t&);      //!< divide by a complex
};

//-------------------------------------------------------------------------------
// External functions of BiLinearForm class
//-------------------------------------------------------------------------------

// Single integral user functions like constructors
// ------------------------------------------------
// using IntegrationMethod object
BilinearForm intg(const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator, const OperatorOnUnknown&,
                  const IntegrationMethod&, SymType = _undefSymmetry);     //!< construct a simple intg bilinear form
BilinearForm intg(const GeomDomain&, const OperatorOnUnknowns&,
                  const IntegrationMethod&, SymType = _undefSymmetry);     //!< construct a simple intg bilinear form
// using explicit quadrature rule as IntegrationMethod
BilinearForm intg(const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator, const OperatorOnUnknown&,
                  QuadRule, number_t =0, SymType = _undefSymmetry); //!< construct a simple intg bilinear form
BilinearForm intg(const GeomDomain&, const OperatorOnUnknowns&,
                  QuadRule, number_t =0, SymType = _undefSymmetry); //!< construct a simple intg bilinear form
// using default quadrature rule as IntegrationMethod
BilinearForm intg(const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator, const OperatorOnUnknown&,
                  SymType = _undefSymmetry); //!< construct a simple intg bilinear form
BilinearForm intg(const GeomDomain&, const OperatorOnUnknowns&,
                  SymType = _undefSymmetry); //!< construct a simple intg bilinear form

// Double integral user functions like constructors
// ------------------------------------------------
// using IntegrationMethod object
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator, const OperatorOnUnknown&,
                  const IntegrationMethod&, SymType = _undefSymmetry);     //!< construct a double intg bilinear form (kernel=1)
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknowns&,
                  const IntegrationMethod&, SymType = _undefSymmetry);     //!< construct a double intg bilinear form (kernel=1)
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,
                  const Kernel&, AlgebraicOperator, const OperatorOnUnknown&,
                  const IntegrationMethod&, SymType = _undefSymmetry);     //!< construct a double intg bilinear form from kernel
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,
                  const OperatorOnKernel&, AlgebraicOperator, const OperatorOnUnknown&,
                  const IntegrationMethod&, SymType = _undefSymmetry);     //!< construct a double intg bilinear form from operator on kernel
BilinearForm intg(const GeomDomain&, const GeomDomain&, const KernelOperatorOnUnknowns&,
                  const IntegrationMethod&, SymType = _undefSymmetry);     //!< construct a double intg bilinear form from kernel operator
// using IntegrationMethods object
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator, const OperatorOnUnknown&,
                  const IntegrationMethods&, SymType = _undefSymmetry);     //!< construct a double intg bilinear form (kernel=1)
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknowns&,
                  const IntegrationMethods&, SymType = _undefSymmetry);     //!< construct a double intg bilinear form (kernel=1)
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,
                  const Kernel&, AlgebraicOperator, const OperatorOnUnknown&,
                  const IntegrationMethods&, SymType = _undefSymmetry);     //!< construct a double intg bilinear form from kernel
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,
                  const OperatorOnKernel&, AlgebraicOperator, const OperatorOnUnknown&,
                  const IntegrationMethods&, SymType = _undefSymmetry);     //!< construct a double intg bilinear form from operator on kernel
BilinearForm intg(const GeomDomain&, const GeomDomain&, const KernelOperatorOnUnknowns&,
                  const IntegrationMethods&, SymType = _undefSymmetry);     //!< construct a double intg bilinear form from kernel operator
// using an explicit quadrature rule
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator, const OperatorOnUnknown&,
                  QuadRule, number_t =0, SymType = _undefSymmetry); //!< construct a double intg bilinear form (kernel=1)
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknowns&,
                  QuadRule, number_t =0, SymType = _undefSymmetry); //!< construct a double intg bilinear form (kernel=1)
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,
                  const Kernel&, AlgebraicOperator, const OperatorOnUnknown&,
                  QuadRule, number_t =0, SymType = _undefSymmetry); //!< construct a double intg bilinear form from kernel
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,
                  const OperatorOnKernel&, AlgebraicOperator, const OperatorOnUnknown&,
                  QuadRule, number_t =0, SymType = _undefSymmetry); //!< construct a double intg bilinear form from operator on kernel
BilinearForm intg(const GeomDomain&, const GeomDomain&, const KernelOperatorOnUnknowns&,
                  QuadRule, number_t =0, SymType = _undefSymmetry); //!< construct a double intg bilinear form from kernel operator
// using default quadrature rule
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator, const OperatorOnUnknown&,
                  SymType = _undefSymmetry); //!< construct a double intg bilinear form (kernel=1)
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknowns&,
                  SymType = _undefSymmetry); //!< construct a double intg bilinear form (kernel=1)
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,
                  const Kernel&, AlgebraicOperator, const OperatorOnUnknown&,
                  SymType = _undefSymmetry); //!< construct a double intg bilinear form from kernel
BilinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&, AlgebraicOperator,
                  const OperatorOnKernel&, AlgebraicOperator, const OperatorOnUnknown&,
                  SymType = _undefSymmetry); //!< construct a double intg bilinear form from operator on kernel
BilinearForm intg(const GeomDomain&, const GeomDomain&, const KernelOperatorOnUnknowns&,
                  SymType = _undefSymmetry); //!< construct a double intg bilinear form from kernel operator

// -------------------------------------------------------------
// algebraic operations on BilinearForms
// -------------------------------------------------------------
const BilinearForm& operator+(const BilinearForm&);               //!< same bilinear form
BilinearForm operator-(const BilinearForm&);                      //!< opposite of a bilinear form
BilinearForm operator+(const BilinearForm&, const BilinearForm&); //!< sum of bilinear forms
BilinearForm operator-(const BilinearForm&, const BilinearForm&); //!< difference of bilinear forms
BilinearForm operator*(const int_t&, const BilinearForm&);        //!< product(left) by an integer scalar
BilinearForm operator*(const int&, const BilinearForm&);          //!< product(left) by an integer scalar
BilinearForm operator*(const number_t&, const BilinearForm&);     //!< product(left) by an integer scalar
BilinearForm operator*(const real_t&, const BilinearForm&);       //!< product(left) by a real scalar
BilinearForm operator*(const complex_t&, const BilinearForm&);    //!< product(left) by a complex scalar
BilinearForm operator*(const BilinearForm&, const int_t&);        //!< product(right) by an integer scalar
BilinearForm operator*(const BilinearForm&, const int&);          //!< product(right) by an integer scalar
BilinearForm operator*(const BilinearForm&, const number_t&);     //!< product(right) by an integer scalar
BilinearForm operator*(const BilinearForm&, const real_t&);       //!< product(right) by a real scalar
BilinearForm operator*(const BilinearForm&, const complex_t&);    //!< product(right) by a complex scalar
BilinearForm operator/(const BilinearForm&, const int_t&);        //!< division by an integer scalar
BilinearForm operator/(const BilinearForm&, const int&);          //!< division by an integer scalar
BilinearForm operator/(const BilinearForm&, const number_t&);     //!< division by an integer scalar
BilinearForm operator/(const BilinearForm&, const real_t&);       //!< division by a real scalar
BilinearForm operator/(const BilinearForm&, const complex_t&);    //!< division by a complex scalar

//composition of bilinear forms
BasicBilinearForm& operator*(const BasicBilinearForm&, const BasicBilinearForm&); //!< compose two basic bilinear forms
BilinearForm operator*(const BilinearForm&, const BilinearForm&); //!< compose two bilinear forms, have to be basic linear forms

// print utility
std::ostream& operator<<(std::ostream&, const BilinearForm&);     //!< output BilinearForm

} // end of namespace xlifepp

#endif /* BILINEAR_FORM_HPP */
