/*
XLiFE++ is an extended library of finite elements written in C++
    Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file LinearForm.hpp
  \author E. Lunéville
  \since 23 mar 2012
  \date  01 apr 2012

  \brief Definition of the xlifepp::LinearForm classes

  This file declares all the classes related to linear form managment.
  The more general linear form under consideration is a list of linear combinations of basic linear forms
  the list is indexed by the unknown of the linear combination
  basic linear forms are, for the moment, integral or double integral type
  The linear form may have multiple unknowns but a linear combination is related to a unique unknown

  xlifepp::LinearForm is the end user class, it deals with a map of <xlifepp::Unknown*,xlifepp::LCLinearForm>
  xlifepp::SuLinearForm deals with linear combination of basic linear forms (single unknown), a vector of pairs of <xlifepp::BasicLinearForm*,xlifepp::complex_t>

  xlifepp::BasicLinearForm is the mother class (abstract) of classes
    - xlifepp::IntgLinearForm for single integral linear form (generally in FEM method)
    - xlifepp::DoubleIngLinearForm for double integral linear form (generally in BEM method)

  As example :
    LinearForm l=intg(Omega,f*u);                   // a single unknown (u) linear form
    LinearForm l=intg(Omega,f*u)+2*intg(Omega,g|v); // a multiple unknowns (u,v) linear form

  all linear algebraic operations are allowed on xlifepp::LinearForm objects
*/


#ifndef LINEAR_FORM_HPP
#define LINEAR_FORM_HPP

#include "config.h"
#include "space.h"
#include "operator.h"
#include "finiteElements.h"

#include <utility>

namespace xlifepp
{
//forward declaration
class IntgLinearForm;
class DoubleIntgLinearForm;
class BilinearFormAsLinearForm;
class EssentialCondition;

/*!
  \class BasicLinearForm
  describes the base class of basic linear forms on a space
  It is an abstract class
*/
class BasicLinearForm
{
  protected:
    const Unknown* u_p;             //!< pointer to unknown
    ComputationType compuType;      //!< computation type (_FEComputation, _IEComputation, _SPComputation, ...)

  public :
    virtual ~BasicLinearForm() {}    //!< virtual destructor

    //accessor
    const Unknown& up() const         //! return the unknown
    {return *u_p;}
    ComputationType computationType() const    //! return the computation type
    {return compuType;}

    //virtual function
    virtual BasicLinearForm* clone() const = 0;   //!< clone of the linear form
    virtual LinearFormType type() const = 0;      //!< return the type of the linear form
    virtual ValueType valueType() const = 0;      //!< return the value type of the linear form
    virtual StrucType strucType() const = 0;      //!< return the struct type of the linear form
    IntgLinearForm* asIntgForm();                 //!< downcast to IntgLinearForm*
    const IntgLinearForm* asIntgForm() const;     //!< downcast to const IntgLinearForm*
    DoubleIntgLinearForm* asDoubleIntgForm();     //!< downcast to DoubleIntgLinearForm*
    const DoubleIntgLinearForm* asDoubleIntgForm() const;           //!< downcast to const DoubleIntgLinearForm*
    BilinearFormAsLinearForm* asBilinearAsLinearForm();             //!< downcast to BilinearFormAsLinearForm*
    const BilinearFormAsLinearForm* asBilinearAsLinearForm() const; //!< downcast to const BilinearFormAsLinearForm*

    virtual string_t asString() const = 0;         //!< interpret as string for print purpose
    virtual void print(std::ostream&) const = 0;   //!< print utility
    virtual void print(PrintStream& os) const =0;
};

/*!
   \class IntgLinearForm
   describes a linear form based on a single integral
*/
class IntgLinearForm : public BasicLinearForm
{
  protected :
    const GeomDomain* domain_p;                        //!< geometric domain of the integral
    const OperatorOnUnknown* opu_p;                    //!< pointer to the operator on unknown
    const IntegrationMethod* intgMethod_p;             //!< pointer to integration method

  public :
    IntegrationMethods intgMethods;                    //!< list of integration methods

    IntgLinearForm(const GeomDomain&, const OperatorOnUnknown&,
                   const IntegrationMethods&);                  //!< basic constructor from OperatorOnUnknown and intg methods
    IntgLinearForm(const GeomDomain&, const OperatorOnUnknown&,
                   const IntegrationMethod&);                  //!< basic constructor from OperatorOnUnknown and an intg method
    IntgLinearForm(const GeomDomain&, const OperatorOnUnknown&,
                   QuadRule = _defaultRule, number_t =0);      //!< basic constructor from OperatorOnUnknown and an explicit quadrature (shortcut)

    void setComputationType();                                 //!< set the computation type (_FEComputation, _FEextComputation, _SPComputation, ...)
    void setIntegrationMethods();                              //!< check or define integration methods

    const OperatorOnUnknown* opu() const           //! return the pointer to the operator on unknown
    {return opu_p;}
    const GeomDomain* domain() const               //! return the pointer to the domain
    {return domain_p;}
    const IntegrationMethod* intgMethod() const    //! pointer to integration method object
    {return intgMethod_p;}

    virtual BasicLinearForm* clone() const;        //!< clone of the linear form
    virtual LinearFormType type() const            //! return the type of the linear form
    {return _intg;}
    virtual ValueType valueType() const            //! return the value type of the linear form
    {return opu_p->valueType();}
    virtual StrucType strucType() const            //! return the structure type of the linear form
    {return opu_p->strucType();}
    bool normalRequired() const                    //! true if normal is required
    {return opu_p->normalRequired();}
    bool xnormalRequired() const                   //! true if x-normal is required
    {return opu_p->xnormalRequired();}
    bool ynormalRequired() const                   //! true if y-normal is required
    {return opu_p->ynormalRequired();}
    bool xnormalRequiredByOpKernel() const         //! true if x-normal is required by operator on Kernel
     {return opu_p->xnormalRequiredByOpKernel();}
    bool ynormalRequiredByOpKernel() const         //! true if y-normal is required by operator on Kernel
    {return opu_p->ynormalRequiredByOpKernel();}

    virtual string_t asString() const
    {return "intg_"+domain_p->name()+" "+opu_p->asString();}
    virtual void print(std::ostream&)const;        //!< print utility
    virtual void print(PrintStream& os) const {print(os.currentStream());}
};

/*!
   \class DoubleIntgLinearForm
   describes a linear form based on a double integral
*/
class DoubleIntgLinearForm : public BasicLinearForm
{
  protected :
    const OperatorOnUnknown* opu_p;  //!< pointer to the operator on unknown
    const GeomDomain* domainx_p;     //!< geometric domain of the integral (say x variable)
    const GeomDomain* domainy_p;     //!< geometric domain of the integral (say y variable)
  public :
    DoubleIntgLinearForm(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&,
                         QuadRule = _defaultRule, number_t =0);            //!< basic construct
    const OperatorOnUnknown* opu() const    //! return the pointer to the operator on unknown
    {return opu_p;}
    const GeomDomain* domainx() const       //! return the pointer to the first domain
    {return domainx_p;}
    const GeomDomain* domainy() const       //! return the pointer to the second domain
    {return domainy_p;}
    virtual BasicLinearForm* clone() const; //!< clone of the linear form
    virtual LinearFormType type() const     //! return the type of the linear form
    {return _doubleIntg;}
    virtual ValueType valueType() const     //! return the value type of the linear form
    {return opu_p->valueType();}
    virtual StrucType strucType() const     //! return the structure type of the linear form
    {return opu_p->strucType();}

    virtual string_t asString() const
      {return "intg_"+domainx_p->name()+"intg_"+domainy_p->name()+" "+opu_p->asString();}
    virtual void print(std::ostream&) const;//!< print utility
    virtual void print(PrintStream& os) const {print(os.currentStream());}
};

//@{
//! useful typedefs to SuLinearForm class
typedef std::pair<BasicLinearForm*, complex_t> lfPair;
typedef std::vector<lfPair>::iterator it_vlfp;
typedef std::vector<lfPair>::const_iterator cit_vlfp;
//@}

/*!
   \class SuLinearForm
   describes a linear combination of linear forms
   with the same unknown.
*/
class SuLinearForm
{
  protected:
    std::vector<lfPair> lfs_;                      //!< list of pairs of linear form and coefficient
  public :
    SuLinearForm() {}                             //!< default constructor
    SuLinearForm(const std::vector<lfPair>& );     //!< full constructor
    SuLinearForm(const SuLinearForm&);             //!< copy constructor
    ~SuLinearForm();                               //!< destructor
    SuLinearForm& operator=(const SuLinearForm&);  //!< assignment operator

    number_t size() const                        //! number of elements of the linear combination
    {return lfs_.size();}
    std::vector<lfPair>& lfs()                   //! access to the list of linear forms
    {return lfs_;}
    const std::vector<lfPair>& lfs() const       //! access to the list of linear forms
    {return lfs_;}
    lfPair& operator()(number_t n)               //! acces to n-th pair
    {return lfs_[n - 1];}
    const lfPair& operator()(number_t n) const   //! acces to n-th pair
    {return lfs_[n - 1];}
    it_vlfp begin()                              //! iterator to the first element of vector of basic linear forms
    {return lfs_.begin();}
    cit_vlfp begin() const                       //! const iterator to the first element of vector of basic linear forms
    {return lfs_.begin();}
    it_vlfp end()                                //! iterator to the last+1 element of vector of basic linear forms
    {return lfs_.end();}
    cit_vlfp end() const                         //! const iterator to the last+1 element of vector of basic linear forms
    {return lfs_.end();}
    const Unknown* unknown() const;    //!< return the unknown of the linear form
    const Space* space() const;        //!< return the space of the linear form
    LinearFormType type() const;       //!< return the type of the linear form (intg,doubleIntg,linearCombination)
    ValueType valueType() const;       //!< return the value type of the linear form
    StrucType strucType() const;       //!< return the structure type of the linear form

    bool checkConsistancy(const SuLinearForm&) const; //!< check consitancy of Unknown
    SuLinearForm& operator +=(const SuLinearForm&);   //!< sum of linear forms
    SuLinearForm& operator -=(const SuLinearForm&);   //!< difference of linear forms
    SuLinearForm& operator *=(const complex_t&);      //!< multiply by a complex
    SuLinearForm& operator /=(const complex_t&);      //!< divide by a complex

    bool normalRequired() const;            //!< true if normal is required
    bool xnormalRequired() const;           //!< true if x-normal is required
    bool ynormalRequired() const;           //!< true if y-normal is required
    bool xnormalRequiredByOpKernel() const; //!< true if x-normal is required by operator on Kernel
    bool ynormalRequiredByOpKernel() const; //!< true if y-normal is required by operator on Kernel

    string_t asString() const;              //!< as symbolic string
    void print(std::ostream&)const;         //!< print utility
    void print(PrintStream& os) const {print(os.currentStream());}

}; // end of SuLinearForm class

//-------------------------------------------------------------------------------
//External functions of SuLinearForm class
//-------------------------------------------------------------------------------

//operations on linear form to construct linear combination of forms
SuLinearForm operator-(const SuLinearForm&);                       //!< opposite of linear form
SuLinearForm operator+(const SuLinearForm&, const SuLinearForm&);  //!< sum of linear forms
SuLinearForm operator-(const SuLinearForm&, const SuLinearForm&);  //!< difference of linear forms
SuLinearForm operator*(const complex_t&, const SuLinearForm&);     //!< product(left) by a scalar
SuLinearForm operator*(const SuLinearForm&, const complex_t&);     //!< product(right) by a scalar
SuLinearForm operator/(const SuLinearForm&, const complex_t&);     //!< division by a scalar
std::ostream& operator<<(std::ostream&, const SuLinearForm&);      //!< output SuLinearForm

//@{
//! useful typedefs to LinearForm class
typedef std::map<const Unknown*, SuLinearForm>::iterator it_mulc;
typedef std::map<const Unknown*, SuLinearForm>::const_iterator cit_mulc;
//@}

/*!
   \class LinearForm
   describes a general linear form, that is
   a list of linear combinations of basic linear forms.
   It is intend to collect linear forms with different unknowns,
   using a map<const Unknown*,SuLinearForm>
   It is the enduser's class
*/
class LinearForm
{
  protected:
    std::map<const Unknown*, SuLinearForm> mlclf_;        //!< list of linear combinations of basic linear forms
    SuLinearForm& operator[](const Unknown*);             //!< access to the linear form associated to an unknown
    const SuLinearForm& operator[](const Unknown*) const; //!< access to the linear form associated to an unknown

  public:
    LinearForm() {}                                       //!< default constructor
    LinearForm(const SuLinearForm&);                      //!< constructor from a linear combination
    bool isEmpty() const;                                 //!< true if no linear form
    void clear()                                          //! clear the map of linear forms
    {mlclf_.clear();}
    it_mulc begin()                                       //! iterator to the first element of map
    {return mlclf_.begin();}
    cit_mulc begin() const                                //! const iterator to the first element of map
    {return mlclf_.begin();}
    it_mulc end()                                         //! iterator to the last+1 element of map
    {return mlclf_.end();}
    cit_mulc end() const                                  //! const iterator to the last+1 element of map
    {return mlclf_.end();}
    bool singleUnknown() const;                           //!< true if only one unknown involved
    number_t nbOfUnknowns() const;                        //!< number of Unknowns involved
    const Unknown* unknown() const;                       //!< return first unknown involved
    std::set<const Unknown*> unknowns() const;            //!< return set of Unknown's involved
    const SuLinearForm& first() const;                    //!< return first linear combination
    LinearForm operator()(const Unknown&) const;          //!< access to the linear form associated to an unknown
    BasicLinearForm& operator()(const Unknown&, number_t);//!< access to the n-th basic linear form associated to an unknown
    const BasicLinearForm& operator()(const Unknown&,
                                      number_t) const;    //!< access to the n-th basic linear form associated to an unknown
    const Unknown& firstUnknown() const                   //! returns first unknown
    {return *mlclf_.begin()->first;}
    ValueType valueType() const;                          //!< return the value type of the linear form
    StrucType strucType() const;                          //!< return the structure type of the linear form
    const SuLinearForm* subLfp(const Unknown*) const;     //!< access to SuLinearForm pointers related to unknown
    SuLinearForm* subLfp(const Unknown*);                 //!< access to SuLinearForm pointers related to unknown (non const)

    bool normalRequired() const;                          //!< true if normal is required
    bool xnormalRequired() const;                         //!< true if x-normal is required
    bool ynormalRequired() const;                         //!< true if y-normal is required
    bool xnormalRequiredByOpKernel() const;               //!< true if x-normal is required by operator on Kernel
    bool ynormalRequiredByOpKernel() const;               //!< true if y-normal is required by operator on Kernel

    void print(std::ostream&)const;                       //!< print utility
    void print(PrintStream& os) const {print(os.currentStream());}

    LinearForm& operator +=(const LinearForm&);           //!< sum of linear forms
    LinearForm& operator -=(const LinearForm&);           //!< difference of linear forms
    LinearForm& operator *=(const complex_t&);            //!< multiply by a complex
    LinearForm& operator /=(const complex_t&);            //!< divide by a complex

    //Equation constructor (implemented in EssentialCondition.cpp)
    EssentialCondition operator = (const complex_t &);    //!< construct essential condition lf(u)=c
};

//-------------------------------------------------------------------------------
//External functions of LinearForm class
//-------------------------------------------------------------------------------
//user functions like constructors of single integral involving Unknown : int_dom v
LinearForm intg(const GeomDomain&, const Unknown&,
                const IntegrationMethods&);                    //!< construct a simple intg linear form from Unknown and some intg methods
LinearForm intg(const GeomDomain&, const Unknown&,
                const IntegrationMethod&);                     //!< construct a simple intg linear form from Unknown and an intg method
LinearForm intg(const GeomDomain&, const Unknown&,
                QuadRule = _defaultRule, number_t =0);         //!< construct a simple intg linear form from Unknown and an explicit quadrature (shortcut)

//user functions like constructors of single integral involving OperatorOnUnknown : int_dom op(v)
LinearForm intg(const GeomDomain&, const OperatorOnUnknown&,
                const IntegrationMethods&);                    //!< construct a simple intg linear form from OperatorOnUnknown and some intg methods
LinearForm intg(const GeomDomain&, const OperatorOnUnknown&,
                const IntegrationMethod&);                     //!< construct a simple intg linear form from OperatorOnUnknown and an intg method
LinearForm intg(const GeomDomain&, const OperatorOnUnknown&,
                QuadRule = _defaultRule, number_t =0);         //!< construct a simple intg linear form from OperatorOnUnknown and an explicit quadrature (shortcut)

//user functions like constructors of single integral involving KernelOperatorOnUnknowns : int_dom opk(v)
LinearForm intg(const GeomDomain&, const KernelOperatorOnUnknowns&,
                const IntegrationMethods&);                    //!< construct a simple intg linear form from KernelOperatorOnUnknowns and some intg methods
LinearForm intg(const GeomDomain&, const KernelOperatorOnUnknowns&,
                const IntegrationMethod&);                     //!< construct a simple intg linear form from KernelOperatorOnUnknowns and an intg method
LinearForm intg(const GeomDomain&, const KernelOperatorOnUnknowns&,
                QuadRule = _defaultRule, number_t =0);         //!< construct a simple intg linear form from KernelOperatorOnUnknowns and an explicit quadrature (shortcut)

//user functions like constructors of double integral : int_domx int_domy opk(v)(x,y)
LinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&,
                const IntegrationMethod&);                     //!< construct a double intg linear form from OperatorOnUnknown and an intg method
LinearForm intg(const GeomDomain&, const GeomDomain&, const OperatorOnUnknown&,
                QuadRule = _defaultRule, number_t =0);         //!< construct a double intg linear form from OperatorOnUnknown and an explicit quadrature (shortcut)
LinearForm intg(const GeomDomain&, const GeomDomain&, const Unknown&,
                const IntegrationMethod&);                     //!< construct a double intg linear form from Unknown and an intg method
LinearForm intg(const GeomDomain&, const GeomDomain&, const Unknown&,
                QuadRule = _defaultRule, number_t =0);         //!< construct a double intg linear form from Unknown and an explicit quadrature (shortcut)

//algebraic operators on LinearForm
LinearForm operator-(const LinearForm&);                     //!< opposite of a linear form
LinearForm operator+(const LinearForm&, const LinearForm&);  //!< sum of linear forms
LinearForm operator-(const LinearForm&, const LinearForm&);  //!< difference of linear forms
LinearForm operator*(const int_t&, const LinearForm&);       //!< product(left) by a scalar
LinearForm operator*(const int&, const LinearForm&);         //!< product(left) by a scalar
LinearForm operator*(const number_t&, const LinearForm&);    //!< product(left) by a scalar
LinearForm operator*(const real_t&, const LinearForm&);      //!< product(left) by a scalar
LinearForm operator*(const complex_t&, const LinearForm&);   //!< product(left) by a scalar
LinearForm operator*(const LinearForm&, const int_t&);       //!< product(right) by a scalar
LinearForm operator*(const LinearForm&, const int&);         //!< product(right) by a scalar
LinearForm operator*(const LinearForm&, const number_t&);    //!< product(right) by a scalar
LinearForm operator*(const LinearForm&, const real_t&);      //!< product(right) by a scalar
LinearForm operator*(const LinearForm&, const complex_t&);   //!< product(right) by a scalar
LinearForm operator/(const LinearForm&, const int_t&);       //!< division by a scalar
LinearForm operator/(const LinearForm&, const int&);         //!< division by a scalar
LinearForm operator/(const LinearForm&, const number_t&);    //!< division by a scalar
LinearForm operator/(const LinearForm&, const real_t&);      //!< division by a scalar
LinearForm operator/(const LinearForm&, const complex_t&);   //!< division by a scalar

//print utilities
std::ostream& operator<<(std::ostream&, const LinearForm&);  //!< output LinearForm

} //end of namespace xlifepp

#endif /* LINEAR_FORM_HPP */
