/*
XLiFE++ is an extended library of finite elements written in C++
  Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file GeomDomain.hpp
  \author E. Lunéville
  \since 04 apr 2012
  \date 04 apr 2012

  \brief Definition of the classes related to geometric domains

  Geometric domains may have various representation :
  - a mesh description (a set of elementary patch, say a part of mesh)
  - a union of domains
  - an intersection of domains
  - an analytic description (say a parametrized geometric domain)

  the base class is GeomDomain whom inherit the classes :
  - xlifepp::MeshDomain for mesh type description
  - xlifepp::CompositeDomain for union or intersection of any domains
  - xlifepp::AnalyticDomain for analytic description (not implemented)

  astract/non abstract pattern is used  :
  \verbatim
       pointer to                        child
  GeomDomain ---------> GeomDomain -------| MeshDomain      (mesh)
                                          | CompositeDomain (union or intersection of domain)
                                          | AnalyticDomain  (analytic description, for future)
                                          | PointDomain     (points cloud, for future)
  \endverbatim
*/

#ifndef GEOM_DOMAIN_HPP
#define GEOM_DOMAIN_HPP

#include "config.h"
#include "utils.h"
#include "GeomElement.hpp"
#include "Geometry.hpp"

namespace xlifepp
{

class Mesh;

//--------------------------------------------------------------------------------

//backward declaration
class MeshDomain;
class CompositeDomain;
class AnalyticalDomain;
class PointsDomain;

//! type of domain
enum DomainType {_undefDomain = 0, _analyticDomain, _meshDomain, _compositeDomain, _pointsDomain};
//! type of composition of domains
enum SetOperationType {_union, _intersection};

/*!
   \class DomainInfo
   to store general information of geometric domain
*/
class DomainInfo
{
  public:
    string_t name;        //!< name of domain
    dimen_t dim;          //!< dimension of domain (the max when compositeDomain)
    DomainType domType;   //!< type of domain
    const Mesh* mesh_p;   //!< pointer to mesh
    string_t description; //!< description of the domain
    bool toSave;          //!< true if the domain is, by construction, intended to be saved into a file (by the saveToFile function)

    DomainInfo(const string_t& na, dimen_t d, DomainType dt, const Mesh* m, const string_t& desc="", bool ts=true) //! basic constructor
      : name(na), dim(d), domType(dt), mesh_p(m), description(desc), toSave(ts) {}
};

/*!
  \class GeomDomain
  base class of geometric domain
*/
class GeomDomain
{
  protected:
    DomainInfo* domainInfo_p;       //!< pointer to domain information class
    GeomDomain* domain_p;           //!< pointer to its child or itself if a child
    GeomDomain* up_p;               //!< pointer to the parent (0 if no parent)
    mutable Geometry* geometry_p;   //!< pointer to the supporting geometry (0 if not defined)

  public:
    static std::vector<const GeomDomain*> theDomains; //!< list of all existing domains
    static void clearGlobalVector();                  //!< delete all domain objects

    GeomDomain(const GeomDomain& d)              //! forbidden copy constructor
    {error("forbidden","GeomDomain::GeomDomain(const GeomDomain&)");}
    GeomDomain& operator=(const GeomDomain& d)   //! forbidden assign operator
    {error("forbidden","GeomDomain::operator=(const GeomDomain&)"); return *this;}

    //constructors
    GeomDomain(const string_t& na="", dimen_t d=0, const Mesh* m=0, Geometry* g=0);                          //!< default and basic constructor
    GeomDomain(const Mesh&, const string_t&, dimen_t, const string_t& desc="", Geometry* g=0, bool ts=true); //!< constructor of a meshdomain
    GeomDomain(SetOperationType sot, const GeomDomain&, const GeomDomain&, const string_t& na = "");  //!< constructor of a composite domain with two domains
    GeomDomain(SetOperationType sot, const std::vector<const GeomDomain*>&, const string_t& na = ""); //!< constructor of a composite domain with n domains
    GeomDomain(const std::vector<Point>&, const string_t& na = ""); //!< constructor of points domain
    virtual ~GeomDomain();                              //!< destructor

    //accessors
    const GeomDomain* up() const                              //! return up_p pointer (const)
    {return up_p;}
    const string_t& name() const                              //! return domain name (const)
    {return domainInfo_p->name;}
    void rename(const string_t & na)                          //! rename domain
    {domainInfo_p->name=na;}
    dimen_t dim() const                                       //! return domain dimension (const)
    {return domainInfo_p->dim;}
    dimen_t spaceDim() const;                                 //!< return the space dimension, say dim of points (const)
    DomainType domType() const                                //! return domain type (const)
    {return domainInfo_p->domType;}
    const string_t domTypeName() const                        //! return domain type name(const)
    {return words("domain type",domainInfo_p->domType);}
    const Mesh* mesh() const                                  //! return mesh pointer (const)
    {return domainInfo_p->mesh_p;}
    const string_t& description() const                       //! return domain description (const)
    {return domainInfo_p->description;}
    void setDescription(const string_t & desc)                //! set domain description
    { domainInfo_p->description=desc; }
    bool isToSave() { return domainInfo_p->toSave; }          //!< returns true if the domain is intended to be saved into a file
    void addSuffix(const string_t& s);                        //!< add a suffix to domain name
    Geometry* geometry() const;                               //!< get geometry, return geometry_p if not null else find the supporting geometry

    virtual bool isUnion() const;                             //!< true if the domain is an union of domains (const)
    virtual bool isIntersection() const;                      //!< true if the domain is a intersection of domains (const)
    virtual bool include(const GeomDomain&) const;            //!< true if the current domain includes a given domain (const)
    virtual MeshDomain* meshDomain();                         //!< access to child meshDomain
    virtual const MeshDomain* meshDomain() const;             //!< access to child MeshDomain (const)
    virtual CompositeDomain* compositeDomain();               //!< access to child CompositeDomain
    virtual const CompositeDomain* compositeDomain() const;   //!< access to child CompositeDomain (const)
    virtual AnalyticalDomain* analyticalDomain() ;            //!< access to child AnalyticalDomain
    virtual const AnalyticalDomain* analyticalDomain() const; //!< access to child AnalyticalDomain (const)
    virtual bool isSideDomain() const;                        //!< true if a side domain
    virtual number_t numberOfElements() const;                //!< number of geometric elements of domain
    virtual GeomElement* element(number_t);                   //!< access to k-th element (k>=1), 0 if not available
    virtual void clearGeomMapData();                          //!< clear temporary GeomData structures of MeshElement's
    virtual void setMaterialId(number_t);                     //!< set material id (>0) for all elements of domain
    virtual void setDomainId(number_t);                       //!< set the DomainId of all elements of domain
    virtual number_t order() const;                           //!< return order of element

    //specific MeshDomain virtual functions
    virtual real_t measure() const;                           //!< measure (length, surface or volume) of MeshDomain
    //@{
    //! set normal orientation of a meshdomain of dimension SpaceDim-1
    virtual void setNormalOrientation(OrientationType, const GeomDomain&) const;
    virtual void setNormalOrientation(OrientationType = _undefOrientationType, const GeomDomain* gp=0) const;
    //@}
    //! return the largest domain including list of subdomains (forbidden here)
    virtual const GeomDomain* largestDomain(std::vector<const GeomDomain*>) const
    {
      where("GeomDomain::largestDomain()");
      error("domain_type_not_handled",words("domain type",domType()));
      return 0;
    }
    virtual GeomDomain& extendDomain(bool useVertex=false, const GeomDomain& omega = GeomDomain()) const //! find or create extension of a side domain
    {
        where("GeomDomain::extendDomain()");
        error("domain_type_not_handled",words("domain type",domType()));
        return const_cast<GeomDomain&>(omega);
    }
    const GeomDomain* extendedDomain(bool useVertex=false, const GeomDomain& omega = GeomDomain()) const; //!< find extension of a side domain
    void updateParentOfSideElements();                                //!< update parent sides info of a side domain (only for mesh domain)

    //print and export stuff
    virtual void print(std::ostream&) const;                          //!< print GeomDomain
    virtual void print(PrintStream& os) const {print(os.currentStream());}
    void saveNormalsToFile(const string_t&, IOFormat iof=_vtk) const; //!< export normal to file
    void saveNormalsToVtk(const string_t&) const;                     //!< export normal to vtk file
    void saveColorsToFile(const string_t& filename) const ;           //!< export element colors to vtk file

    //static functions
    static const GeomDomain* findDomain(const GeomDomain*);      //!< find a domain of mesh by its pointer
    static const GeomDomain* findDomain(SetOperationType, const std::vector<const GeomDomain*>&); //!< find an union/intersection domain by the domain pointers
    static const GeomDomain* findUnionDomain(const GeomDomain*); //!< find mesh domain corresponding to an union domain
    static void printTheDomains(std::ostream&);                  //!< print all domains
};

//! print geomdomain
std::ostream& operator<<(std::ostream&, const GeomDomain&);
//! construct or identify the geometrical symbolic union of domains
const GeomDomain* geomUnionOf(std::vector<const GeomDomain*>&, const GeomDomain*);
//! merge some geometrical domains (true union of elements)
template<typename S_> GeomDomain& merge(const std::vector<const GeomDomain*>&, S_ name);
//! merge some geometrical domains (true union of elements), non const version
template<typename S_> GeomDomain& merge(const std::vector<GeomDomain*>&, S_ name);
//! merge 2 geometrical domains (true union of elements)
template<typename S_> GeomDomain& merge(const GeomDomain&, const GeomDomain&, S_ name);
//! merge 3 geometrical domains (true union of elements)
template<typename S_> GeomDomain& merge(const GeomDomain&, const GeomDomain&, const GeomDomain&, S_ name);
//! merge 4 geometrical domains (true union of elements)
template<typename S_> GeomDomain& merge(const GeomDomain&, const GeomDomain&, const GeomDomain&, const GeomDomain&, S_ name);
//! merge 5 geometrical domains (true union of elements)
template<typename S_> GeomDomain& merge(const GeomDomain&, const GeomDomain&, const GeomDomain&, const GeomDomain&,
                                        const GeomDomain&, S_ name);
//! merge 6 geometrical domains (true union of elements)
template<typename S_> GeomDomain& merge(const GeomDomain&, const GeomDomain&, const GeomDomain&, const GeomDomain&,
                                        const GeomDomain&, const GeomDomain&, S_ name);
void locateError(const string_t&, const GeomDomain&, const Point&); //!< error management for locate tool
inline real_t measure(const GeomDomain& g)  //! measure (length, surface or volume) of Domain
{return g.measure();}

/*!
   \class MeshDomain
   class to handle domain with mesh description
*/
class MeshDomain : public GeomDomain
{
  protected:
    MeshDomain* parent_p;                          //!< parent mesh domain if submesh domain (not fully managed !!!)
  public:
    std::vector<GeomElement*> geomElements;        //!< list of geometric elements
    std::set<ShapeType> shapeTypes;                //!< list of element shape types in mesh domain (counted once)
    const MeshDomain* extensionof_p;               //!< side domain pointer, if domain is an extension of a side domain, else 0
    mutable std::pair<OrientationType, const GeomDomain*> normalOrientationRule;  //!< the normal orientation rule

    //additional structures created if required, declared as mutable to allow construction on fly by const methods
    mutable std::map<Point,std::list<GeomElement*> > vertexElements;  //!< list of elements by vertex
    mutable KdTree<Point> kdtree;                                     //!< kdtree structure using to fast locate element containing a given point
    mutable std::map<GeomElement*, std::list<GeoNumPair> > sideToExt; //!< map side element -> list of extended elements (for a side domain!)
    mutable bool parentSidesUptodate;                                 //!< true :  parents of side elements are up to date
    mutable bool orientationComputed;                                 //!< true :  orientations are computed
    mutable bool jacobianComputed;                                    //!< true :  jacobians  are computed
    mutable bool diffEltComputed;                                     //!< true :  differential elements  are computed
    mutable bool normalComputed;                                      //!< true :  normals  are computed
    mutable bool inverseJacobianComputed;                             //!< true :  inverse of jacobians  are computed

    //constructor
    MeshDomain() : parent_p(0), extensionof_p(0) {}
    MeshDomain(const Mesh&, const string_t&, dimen_t, const string_t&); //!< constructor

    //accessors and basic tools
    MeshDomain* parent() const {return parent_p;}  //!< parent mesh domain if submesh

    void setShapeTypes();                          //!< set the element shape types in mesh domain
    virtual MeshDomain* meshDomain()               //! access to meshDomain pointer
    {return this;}
    virtual const MeshDomain* meshDomain() const   //! access to meshDomain pointer (const)
    {return this;}
    virtual bool isUnion() const                   //! MeshDomain is not an union of domains
    {return false;}
    virtual bool isIntersection() const            //! MeshDomain is not an intersection of domains
    {return false;}
    virtual number_t numberOfElements() const      //! number of geometric elements of meshdomain
    {return geomElements.size();}
    virtual GeomElement* element(number_t k)       //! access to k-th element (k>=1), 0 if not available
    {
      if(k<1 || k>geomElements.size()) error("index_out_of_range","k", 1, geomElements.size());
      return geomElements[k-1];
    }
    virtual const GeomElement* element(number_t k) const      //! access to k-th element (k>=1), 0 if not available
    {
      if(k<1 || k>geomElements.size()) error("index_out_of_range","k", 1, geomElements.size());
      return geomElements[k-1];
    }
    virtual number_t order() const;                //!< order of geometric elements of domain

    virtual void setMaterialId(number_t);          //!< set material id (>0) for all geom elements of domain
    virtual void setDomainId(number_t);            //!< set the DomainId of all elements of domain

    //various tools
    virtual bool include(const GeomDomain&) const; //!< true if the current domain includes a given domain (const)
    bool include(const MeshDomain& d) const;       //!< true if the current domain includes a given mesh domain (const)
    bool isSideDomain() const;                     //!< true if MeshDomain is a side of meshdomain (first element has one parent)
    bool isInterface() const;                      //!< true if MeshDomain is an interface between domains (first element has more than one parent)
    std::set<number_t> nodeNumbers() const;        //!< return the set of all node numbers of meshdomain, ascending order
    std::set<number_t> vertexNumbers() const;      //!< return the set of all vertex numbers of meshdomain, ascending order
    std::vector<Point> nodes() const;              //!< return the set of all nodes of meshdomain, numbers ascending order
    std::vector<Point> vertices() const;           //!< return the set of all vertices of meshdomain, numbers ascending order
    real_t measure() const;                        //!< measure (length, surface or volume) of MeshDomain
    bool isUnionOf(const std::vector<const GeomDomain*>&) const;           //!< check if an union of subdomains is equal to the current mesh domain
    const GeomDomain* largestDomain(std::vector<const GeomDomain*>) const; //!< return the largest domain including list of subdomains
    GeomDomain& extendDomain(bool useVertex=false,
                             const GeomDomain& omega = GeomDomain()) const;//!< find or create extension of a side domain

    //tools to update informations
    void buildGeomData() const;                    //!< build measure and orientation of domain elements
    void updateSides(std::map<string_t,std::vector<GeoNumPair> >&);  //!< update parent sides of elements of a side domain
    void buildVertexElements() const;              //!< build VertexElementsMap from geomElements
    void buildKdTree() const;                      //!< build KdTree
    const std::map<GeomElement*, std::list<GeoNumPair> >& buildSideToExt( const GeomDomain& = GeomDomain()) const; //!< build sideToExt map

    //tools to locate a point in domain
    GeomElement* locate(const Point&) const;       //!< locate element containing a given point
    GeomElement* nearest(Point& pt,real_t&) const; //!< locate nearest element of a given point
    std::list<GeomElement*> elementsCloseTo(Point& pt) const; //!< locate elements containing a given point

    // ================ tools to rebuild the domain
    //! set color of geom elements of domain according to vertex values and a coloring rule
    void setColor(const std::vector<real_t>&, const std::map<number_t,number_t>&, ColoringRule=defaultColoringRule) const;
    //! rebuild  by collecting elements when color satisfies a criterium given by ComparisonFunction
    void rebuild(const ComparisonFunction<>&, bool restrictedToDomain=false);
    //! rebuild a side domain from plain domain elements
    void rebuild(const std::map<string_t, std::vector<GeoNumPair> >&, std::map<string_t, GeomElement*>&,
                 const std::map<GeomDomain*,std::set<GeomElement*>*>&);

    // =========== tools to deal with normal orientations
    //! set normal orientation of a meshdomain of dimension SpaceDim-1
    void setNormalOrientation(OrientationType, const GeomDomain&) const;
    //! set normal orientation of a meshdomain of dimension SpaceDim-1
    void setNormalOrientation(OrientationType = _undefOrientationType, const GeomDomain* gp=0) const;
    //! set orientation of MeshElement's of a manifold say a mesh domain of dimension SpaceDim-1 domain
    void setOrientationForManifold(OrientationType = _towardsInfinite) const;
    //! set orientation of MeshElement's of a boundary say a mesh domain of dimension SpaceDim-1 domain
    void setOrientationForBoundary(OrientationType ort =_outwardsDomain, const GeomDomain*  =0) const;
    void reverseOrientations() const; //!< reverse all the element orientations

    void clearGeomMapData(); //!< clear temporary GeomData structures of MeshElement's

    //in/out stuff
    virtual void print(std::ostream&) const; //!< print MeshDomain informations
    virtual void print(PrintStream& os) const {print(os.currentStream());}
};

std::ostream& operator<<(std::ostream&, const MeshDomain&); //!< print operator
// void addNode(Node<GeomElement>*, GeomElement*, std::map<GeomElement*,bool>&,std::map<string_t, std::list<GeomElement*> >&); //!< utility
void createPath(Node<GeomElement>*, std::map<GeomElement*,bool>&,std::map<string_t, std::list<GeomElement*> >&); //!< utility

/*!
   \class CompositeDomain
   class to handle union or intersection of domains
*/
class CompositeDomain : public GeomDomain
{
  protected:
    SetOperationType setOpType_;                            //!< type of the set operation
    std::vector<const GeomDomain*> domains_;                //!< list of domains
  public:
    CompositeDomain(SetOperationType, const std::vector<const GeomDomain*>&,
                    const string_t&);                       //!< constructor
    SetOperationType setOpType() const                      //! return type of operation (_union,_intersection)
    {return setOpType_;}
    const std::vector<const GeomDomain*>& domains() const   //! access to list of domains (const)
    {return domains_;}
    virtual CompositeDomain* compositeDomain()              //! access to CompositeDomain pointer
    {return this;}
    virtual const CompositeDomain* compositeDomain() const  //! access to CompositeDomain pointer (const)
    {return this;}
    virtual bool isUnion()const                             //! true if an union of domains (const)
    {return setOpType_ == _union;}
    virtual bool isIntersection()const                      //! true if an intersection of domains (const)
    {return setOpType_ == _intersection;}
    virtual bool include(const GeomDomain&) const;          //!< true if the current domain includes a given domain (const)
    virtual bool isSideDomain() const;                      //!< true if composite domain is aside domain
    std::vector<const GeomDomain*> basicDomains() const;    //!< return the list of basic domains (recursive)
    virtual number_t numberOfElements() const;              //!< number of geometric elements of composite domain is not available
    virtual void print(std::ostream&) const;                //!< print CompositeDomain informations
    virtual void print(PrintStream& os) const {print(os.currentStream());}
};

std::ostream& operator<<(std::ostream&, const CompositeDomain&); //!< print operator

/*!
   \class AnalyticalDomain
   class to handle domain with analytical description
   to be developed later
*/
class AnalyticalDomain : public GeomDomain
{
  public:
    AnalyticalDomain* analyticalDomain()              //! access to AnalyticalDomain pointer
    {return this;}
    const AnalyticalDomain* analyticalDomain() const  //! access to AnalyticalDomain pointer (const)
    {return this;}
    virtual bool include(const GeomDomain&) const    //! true if the current domain includes a given domain (const)
    {return false;}                                  // no inclusion managed
};

/*!
   \class PointsDomain
   class handles a list of points (cloud)
*/
class PointsDomain : public GeomDomain
{
  public:
    std::vector<Point> points;                     //!< list of points

    PointsDomain(const std::vector<Point>&, const string_t& = "");  //!< PointsDomain constructor

    PointsDomain* pointsDomain()                   //! access to PointsDomain pointer
    {return this;}
    const PointsDomain* pointsDomain()  const      //! access to PointsDomain pointer (const)
    {return this;}
    const Point& point(number_t n) const           //! access to n-th point (n>=1) (const)
    {return points[n-1];}
    Point& point(number_t n)                       //! access to n-th point (n>=1)
    {return points[n-1];}
    virtual bool include(const GeomDomain&) const  //! true if the current domain includes a given domain (const)
    {return false;}                                  // no inclusion managed
    virtual void print(std::ostream&) const;       //!< print PointsDomain informations
    virtual void print(PrintStream& os) const {print(os.currentStream());}
};

std::ostream& operator<<(std::ostream&, const PointsDomain&); //!< print operator

//! useful typedef for domain objects
typedef std::pair<const GeomDomain*,const GeomDomain*> DomainPair;

//=================================================================================
// utilities related to GeomDomain or GeomElement
//=================================================================================

//@{
/*!
  extract GeomDomain or GeomElement information from Parameters
  \note if pointer parameter does not exist, it is created with 0 value (see Parameters.get member function)
*/
inline string_t domainNameFromParameters(Parameters& pars)
{
  const void* p=0;
  const GeomDomain* dom=static_cast<const GeomDomain*>(pars.get("Domain pointer",p));
  if(dom!=0) return dom->name();
  return "?";
}

inline number_t materialIdFromParameters(Parameters& pars)
{
  const void* p=0;
  const GeomElement* gelt=static_cast<const GeomElement*>(pars.get("GeomElement pointer",p));
  if(gelt!=0) return gelt->materialId;
  return 0;
}

inline number_t domainIdFromParameters(Parameters& pars)
{
  const void* p=0;
  const GeomElement* gelt=static_cast<const GeomElement*>(pars.get("GeomElement pointer",p));
  if(gelt!=0) return gelt->domainId;
  return 0;
}

inline const GeomDomain* domainFromParameters(Parameters& pars)
{
  const void* p=0;
  return static_cast<const GeomDomain*>(pars.get("Domain pointer",p));
}

inline const GeomElement* geomElementFromParameters(Parameters& pars)
{
  const void* p=0;
  return static_cast<const GeomElement*>(pars.get("GeomElement pointer",p));
}
//@}

//! extract normal vector from Parameters
inline Vector<real_t>& normalVectorFromParameters(const Parameters& pa)
{
  return const_cast<Vector<real_t>&>(*static_cast<const Vector<real_t>*>(pa("_n").p_));
}


//=================================================================================
//geometric facilities (to be moved elsewhere)
//=================================================================================
bool intersectionSegments(const Point& P, const Point& Q, const Point& A, const Point& B,
                          Point& I);  //!< intersection of [P,Q] and [A,B] (2D)
bool intersectionSegmentTriangle(const Point& P, const Point& Q, const Point& A, const Point& B, const Point& C,
                                 Point& I);  //!< intersection of [P,Q] and [A,B,C] (2D,3D)
bool intersectionSegmentQuadrangle(const Point& P, const Point& Q, const Point& A, const Point& B, const Point& C, const Point& D,
                                   Point& I);  //!< intersection of [P,Q] and [A,B,C] (2D,3D)

typedef PCollection<GeomDomain> Domains;  //!< collection of GeomDomain pointers

} // end of namespace xlifepp

#endif // GEOM_DOMAIN_HPP
