/*
XLiFE++ is an extended library of finite elements written in C++
  Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file Geometry.hpp
  \authors D. Martin, E. Lunéville, N. Kielbasiewicz
  \since 10 dec 2007
  \date 17 oct 2014

  \brief Definition of the xlifepp::Geometry class and childs

  xlifepp::Geometry class represents geometric data of the "physical" problem
  It provides the following minimum informations :
  - the physical space dimension
  - the names of space variables (for documentation purpose)
  - a bounding box
  - a minimal box (a sort of box convex hull)
  In future, this class should be abble to handle parametrisations of domain computation
*/

#ifndef GEOMETRY_HPP
#define GEOMETRY_HPP

#include "config.h"
#include "utils.h"

namespace xlifepp
{

/*!
   \class BoundingBox
  utility class to describe a bounding box
 */
typedef std::pair<real_t, real_t> RealPair;

class BoundingBox
{
  private:
    std::vector<RealPair> bounds_; //!< Bounding values of the box [xmin,xmax]x[ymin,ymax]x[zmin,zmax]

  public:
    BoundingBox() { bounds_.resize(0); }                                 //!< void constructor
    BoundingBox(const std::vector<RealPair>& bds) : bounds_(bds) {}      //!< basic constructor
    BoundingBox(real_t, real_t, real_t, real_t, real_t, real_t);         //!< basic constructor 3D box
    BoundingBox(real_t, real_t, real_t, real_t);                         //!< basic constructor 2D box
    BoundingBox(real_t, real_t);                                         //!< basic constructor 1D box
    BoundingBox(const Point&, const Point&, const Point&, const Point&); //!< general constructor 3D box
    BoundingBox(const Point&, const Point&, const Point&);               //!< general constructor 2D box
    BoundingBox(const Point&, const Point&);                             //!< general constructor 1D box
    BoundingBox(const std::vector<Point>& vp);                           //!< genereal constructor from a set of points
    dimen_t dim() const { return bounds_.size(); }                       //!< dimension of the bounding box
    std::vector<RealPair> bounds() const { return bounds_; }             //!< return bound in all direction
    RealPair bounds(dimen_t i) const;                                    //!< return box bounds in direction i (from 1 to dim)
    RealPair& bounds(dimen_t i) {return bounds_[i-1];}                   //!< return box bounds in direction i (from 1 to dim), non const
    RealPair xbounds() const { return bounds(1); }                       //!< return box bounds in first direction, if exists
    RealPair ybounds() const { return bounds(2); }                       //!< return box bounds in second direction, if exists
    RealPair zbounds() const { return bounds(3); }                       //!< return box bounds in third direction, if exists
    Point minPoint() const;                                              //!< return the min point (left,bottom,front)
    Point maxPoint() const;                                              //!< return the max point (right,top,back)
    std::vector<Point> points();                                         //!< return the vertices of the bounding box
    real_t diameter() const;                                             //!< return the max of (xmax-xmin, ymax -ymin, zmax-zmin);
    real_t diameter2() const;                                            //!< return sqrt((xmax-xmin)^2+(ymax -ymin)^2+ (zmax-zmin)^2)
    BoundingBox& operator +=(const BoundingBox&);                        //!< merge a bounding box
    void print(std::ostream&) const;                                     //!< print BoundingBox
    void print(PrintStream& os) const {print(os.currentStream());}
    string_t asString() const;                                           //!< format as string : [a,b]x[c,d]x[e,f]

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a BoundingBox
    BoundingBox& transform(const Transformation& t);
    //! apply a translation on a BoundingBox (vector version)
    BoundingBox& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a BoundingBox (3 reals version)
    BoundingBox& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a BoundingBox
    BoundingBox& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a BoundingBox
    BoundingBox& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.), real_t angle = 0.);
    //! apply a rotation on a BoundingBox
    BoundingBox& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a BoundingBox
    BoundingBox& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a BoundingBox
    BoundingBox& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a BoundingBox
    BoundingBox& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a BoundingBox
    BoundingBox& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a BoundingBox
    BoundingBox& homothetize(real_t factor);
    //! apply a point reflection on a BoundingBox
    BoundingBox& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a BoundingBox
    BoundingBox& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a BoundingBox
    BoundingBox& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a BoundingBox
    BoundingBox& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a BoundingBox
    BoundingBox& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

std::ostream& operator<<(std::ostream&, const BoundingBox&); //!< output BoundingBox
std::ostream& operator<<(std::ostream&, const RealPair&);    //!< output pair of reals
real_t dist(const BoundingBox&, const BoundingBox&);         //!< distance from two bounding boxes

/*!
   \class MinimalBox
  utility class to describe a minimal box
  This is like the bounding box, but independant from the orientation.
 */
class MinimalBox
{
  protected:
    std::vector<Point> bounds_;              //!< Bounding values of the box [xmin,xmax]x[ymin,ymax]x[zmin,zmax]

  public:
    MinimalBox() { bounds_.resize(0); }                                 //!< void constructor
    MinimalBox(const std::vector<Point>& bds) : bounds_(bds) {}         //!< basic constructor
    MinimalBox(const std::vector<RealPair>&);                           //!< basic constructor
    MinimalBox(real_t, real_t, real_t, real_t, real_t, real_t);         //!< basic constructor 3D box
    MinimalBox(real_t, real_t, real_t, real_t);                         //!< basic constructor 2D box
    MinimalBox(real_t, real_t);                                         //!< basic constructor 1D box
    MinimalBox(const Point&, const Point&, const Point&, const Point&); //!< general constructor 3D box
    MinimalBox(const Point&, const Point&, const Point&);               //!< general constructor 2D box
    MinimalBox(const Point&, const Point&);                             //!< general constructor 1D box
    dimen_t dim() const;                                                //!< dimension of the bounding box
    std::vector<Point> bounds() const { return bounds_; }               //!< return bounds_
    std::vector<Point> vertices() const;                                //!< returns vertices
    Point boundPt(dimen_t i) const;                                     //!< return box bounds in direction i (from 1 to dim)
    Point minPoint() const { return bounds_[0]; }                       //!< return the min point (left,bottom,front)
    Point maxPoint() const;                                             //!< return the max point (right,top,back)
    void print(std::ostream&) const;                                    //!< print MinimalBox
    void print(PrintStream& os) const {print(os.currentStream());}
    string_t asString() const;                                          //!< format as string : [a,b]x[c,d]x[e,f]

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a MinimalBox
    MinimalBox& transform(const Transformation& t);
    //! apply a translation on a MinimalBox (vector version)
    MinimalBox& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a MinimalBox (3 reals version)
    MinimalBox& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a MinimalBox
    MinimalBox& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a MinimalBox
    MinimalBox& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.), real_t angle = 0.);
    //! apply a rotation on a MinimalBox
    MinimalBox& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a MinimalBox
    MinimalBox& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a MinimalBox
    MinimalBox& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a MinimalBox
    MinimalBox& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a MinimalBox
    MinimalBox& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a MinimalBox
    MinimalBox& homothetize(real_t factor);
    //! apply a point reflection on a MinimalBox
    MinimalBox& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a MinimalBox
    MinimalBox& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a MinimalBox
    MinimalBox& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a MinimalBox
    MinimalBox& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a MinimalBox
    MinimalBox& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

std::ostream& operator<<(std::ostream&, const MinimalBox&); //!< output MinimalBox

// forward declarations
class Segment;
class EllArc;
class CircArc;
class Polygon;
class Triangle;
class Quadrangle;
class Parallelogram;
class Rectangle;
class Square;
class Ellipse;
class Disk;
class SetOfElems;
class SetOfPoints;
class Polyhedron;
class Tetrahedron;
class Hexahedron;
class Parallelepiped;
class Cuboid;
class Cube;
class Ellipsoid;
class Ball;
class Trunk;
class Cylinder;
class Prism;
class Cone;
class Pyramid;
class RevTrunk;
class RevCylinder;
class RevCone;

/*!
   \class Geometry
  handles geometric data of the physical problem
 */
class Geometry
{
  public:
    BoundingBox boundingBox;                    //!< the bounding box
    MinimalBox minimalBox;                      //!< the smallest box containing the geometry
    mutable bool force;                         //!< tag to force inclusion when geometry engines fails
    mutable bool crackable;                     //!< tag to define if the geometry should be cracked or not
    mutable CrackType crackType;                //!< type of crack (open or closed)
    mutable string_t crackDomNameToOpen;        //!< when open crack, name to the side domain to open.
    bool isPlaneSurface;                        //!< used only for 2D loop geometries
  protected:
    string_t domName_;                          //!< geometry domain name
    dimen_t dim_;                               //!< physical space dimension
    ShapeType shape_;                           //!< geometrical shape
    std::vector<string_t> sideNames_;           //!< list of side domains' names
    std::vector<string_t> theNamesOfVariables_; //!< names of variables (x,y,z), (r,theta)
    string_t teXFilename_;                      //!< name of the file containing the tex code drawing the geometry
  private:
    mutable std::map<number_t, Geometry*> components_;      //!< composite or loop geometry if not empty
    std::map<number_t, std::vector<number_t> > geometries_; //!< composite geometry if not empty
    std::map<number_t, std::vector<number_t> > loops_;      //!< loop geometry if not empty
    mutable Transformation* extrusion_;                     //!< extrusion geometry is not null
    number_t layers_;                                       //!< only for extrusion geometry if not 0

  public:
    //! void constructor
    Geometry(): force(false), crackable(false), crackType(_noCrack), crackDomNameToOpen(""), isPlaneSurface(true), domName_(""), dim_(0),
                shape_(_noShape), extrusion_(0), layers_(0)
    { sideNames_.resize(0); }
    Geometry(const Geometry& g);               //!< copy constructor
    Geometry(const BoundingBox&, const string_t& na = "", ShapeType sht=_noShape,
             const string_t& nx = "x", const string_t& ny = "y",
             const string_t& nz = "z");          //!< basic constructor from bounding box
    Geometry(const BoundingBox&, dimen_t, const string_t& na = "", ShapeType sht=_noShape,
             const string_t& nx = "x", const string_t& ny = "y",
             const string_t& nz = "z");          //!< basic constructor from bounding box and dim
    Geometry(dimen_t, const string_t& na = "", ShapeType sht=_noShape,
             const string_t& nx = "x", const string_t& ny = "y",
             const string_t& nz = "z");          //!< basic constructor from (unit bounding box)

  protected:
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();

  public:
    virtual Geometry* clone() const { return new Geometry(*this); } //!< virtual copy constructor

    virtual ~Geometry();

    Geometry & operator=(const Geometry& g);  //!< assign operator
    bool operator==(const Geometry& g) const; //!< comparison operator

    dimen_t dim() const { return dim_; }                  //!< returns the physical space dimension
    dimen_t dimPoint() const;                             //!< dimension of points, may differ from dimension of the geometry
    ShapeType shape() const { return shape_; }            //!< returns the geometrical shape
    void shape(ShapeType sh) { shape_=sh; }               //!< sets the geometrical shape
    string_t domName() const { return domName_; }         //!< get the domain name
    string_t& domName() { return domName_; }              //!< get the domain name (writable)
    void domName(const string_t& nm) { domName_ = nm; }   //!< set the domain name
    string_t teXFilename() const { return teXFilename_; } //!< returns tex file name
    void teXFilename(const string_t& fn) { teXFilename_ = fn; } //!< set filename to get a TeX output (drawing of the mesh)

    const std::vector<string_t> sideNames() const { return sideNames_; } //!< returns the list of side domains' names
    std::vector<string_t> sideNames() { return sideNames_; }            //!< returns the list of side domains' names
    std::map<number_t, Geometry*> components() { return components_;}             //!< accessor to components_
    const std::map<number_t, Geometry*> components() const { return components_;} //!< accessor to components_ (const)
    const std::map<number_t, std::vector<number_t> > geometries() const { return geometries_; } //!< accessor to geometries_ (const)
    std::map<number_t, std::vector<number_t> > loops() const { return loops_; }                 //!< accessor to loops_ (const)
    
    //! accessor to number of nodes on the Geometry (read only)
    virtual number_t n(number_t i=0) const { error("not_handled"); return 2;}

    void checkSideNamesAndUpdate(number_t n); //!< check is n side names are given and update size when single
    void sideNames(const std::vector<string_t> sn) { sideNames_=sn; }

    number_t layers() const                     //! return the number of layers
    { return layers_; }
    Transformation* extrusion()                 //! return the extrusion transformation (non const)
    { return extrusion_; }
    const Transformation* extrusion() const     //! return the extrusion transformation (const)
    { return extrusion_; }

    Geometry* find(const string_t& n) const;    //!< return geometry with name n (a canonical geometry or a composite of canonical geometry's)
    virtual void collect(const string_t& n, std::list<Geometry*>&) const; //!< collect in a list all canonical geometry's with name n

    void print(std::ostream&) const; //!< print Geometry
    void print(PrintStream& os) const {print(os.currentStream());}
    void printBoundNodes(std::ostream&) const; //!< print Geometry boundNodes() vector
    void printBoundNodes(PrintStream& os) const {printBoundNodes(os.currentStream());}
    void printNodes(std::ostream&) const;      //!< print Geometry nodes() vector
    void printNodes(PrintStream& os) const {printNodes(os.currentStream());}
    void printCurves(std::ostream&) const;     //!< print Geometry curves() vector
    void printCurves(PrintStream& os) const {printCurves(os.currentStream());}
    void printSurfs(std::ostream&) const;      //!< print Geometry surfs() vector
    void printSurfs(PrintStream& os) const {printSurfs(os.currentStream());}
    friend std::ostream& operator<<(std::ostream&, const Geometry&);
    void addSuffix(const string_t& s); //!< add a suffix to all names (geometry names and side domain names)
    //! format as string
    virtual string_t asString() const { return string_t(); }

    //! compute the bounding box for a composite/loop geometry
    void computeBB();
    //! compute the minimal box for a composite/loop geometry
    virtual void computeMB() {} //!< compute the minimal box
    void updateBB(const std::vector<RealPair>& bb) { boundingBox = BoundingBox(bb); } //!< update the bounding box

    //! return the list of boundary nodes defining the Geometry
    virtual std::vector<const Point*> boundNodes() const;
    //! return the list of nodes defining the Geometry
    virtual std::vector<Point*> nodes();
    //! return the list of nodes defining the Geometry (const)
    virtual std::vector<const Point*> nodes() const;
    //! return a node belonging to the Geometry (const)
    virtual Point firstNode() const;
    //! return the list of curves defining the borders of a 2D Geometry
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > curves() const;
    //! return the list of surfaces defining the borders of a 3D Geometry
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const;
    //! return the number of sides of the geometry
    virtual number_t nbSides() const;
    virtual bool withNnodes() const //! check if geometry is defined only with _nnodes or with _hsteps option
    { error("not_handled" "Geometry::withNnodes()"); return false; }
    //! i-th value is nnodes of the Geometry's i-th border, else -1 if defined with hsteps
    virtual std::vector<int_t> nnodesPerBorder()
    { error("not_handled", "Geometry::nnodesPerBorder()"); return std::vector<int_t>(); }
    //! return the length/area/volume of the geometry
    virtual real_t measure() const
    { error("not_handled", "Geometry::measure()"); return 0.; }

    //! return the number of sides of the geometry
    number_t nbLateralSidesForExtrusion() const;

    bool isCoplanar(const Geometry& g) const; //!< check if geometry is coplanar with another geometry (g)
    bool isInside(const Geometry& g) const; //!< check if geometry is inside another geometry (g)


    virtual const Segment* segment() const //! access to child Segment object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _segment)); return 0; }
    virtual Segment* segment() //! access to child Segment object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _segment)); return 0; }

    virtual const EllArc* ellArc() const //! access to child EllArc object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _ellArc)); return 0; }
    virtual EllArc* ellArc() //! access to child EllArc object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _ellArc)); return 0; }

    virtual const CircArc* circArc() const //! access to child CircArc object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _circArc)); return 0; }
    virtual CircArc* circArc() //! access to child CircArc object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _circArc)); return 0; }

    virtual const Polygon* polygon() const //! access to child Polygon object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _polygon)); return 0; }
    virtual Polygon* polygon() //! access to child Polygon object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _polygon)); return 0; }

    virtual const Triangle* triangle() const //! access to child Triangle object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _triangle)); return 0; }
    virtual Triangle* triangle() //! access to child Triangle object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _triangle)); return 0; }

    virtual const Quadrangle* quadrangle() const //! access to child Quadrangle object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _quadrangle)); return 0; }
    virtual Quadrangle* quadrangle() //! access to child Quadrangle object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _quadrangle)); return 0; }

    virtual const Parallelogram* parallelogram() const //! access to child Parallelogram object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _parallelogram)); return 0; }
    virtual Parallelogram* parallelogram() //! access to child Parallelogram object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _parallelogram)); return 0; }

    virtual const Rectangle* rectangle() const //! access to child Rectangle object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _rectangle)); return 0; }
    virtual Rectangle* rectangle() //! access to child Rectangle object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _rectangle)); return 0; }

    virtual const Square* square() const //! access to child Square object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _square)); return 0; }
    virtual Square* square() //! access to child Square object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _square)); return 0; }

    virtual const Ellipse* ellipse() const //! access to child Ellipse object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _ellipse)); return 0; }
    virtual Ellipse* ellipse() //! access to child Ellipse object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _ellipse)); return 0; }

    virtual const Disk* disk() const //! access to child Disk object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _disk)); return 0; }
    virtual Disk* disk() //! access to child Disk object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _disk)); return 0; }

    virtual const Polyhedron* polyhedron() const //! access to child Polyhedron object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _polyhedron)); return 0; }
    virtual Polyhedron* polyhedron() //! access to child Polyhedron object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _polyhedron)); return 0; }

    virtual const Tetrahedron* tetrahedron() const //! access to child tetrahedron object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _tetrahedron)); return 0; }
    virtual Tetrahedron* tetrahedron() //! access to child Tetrahedron object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _tetrahedron)); return 0; }

    virtual const Hexahedron* hexahedron() const //! access to child Hexahedron object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _hexahedron)); return 0; }
    virtual Hexahedron* hexahedron() //! access to child Hexahedron object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _hexahedron)); return 0; }

    virtual const Parallelepiped* parallelepiped() const //! access to child Parallelepiped object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _parallelepiped)); return 0; }
    virtual Parallelepiped* parallelepiped() //! access to child Parallelepiped object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _parallelepiped)); return 0; }

    virtual const Cuboid* cuboid() const //! access to child Cuboid object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _cuboid)); return 0; }
    virtual Cuboid* cuboid() //! access to child Cuboid object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _cuboid)); return 0; }

    virtual const Cube* cube() const //! access to child Cube object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _cube)); return 0; }
    virtual Cube* cube() //! access to child Cube object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _cube)); return 0; }

    virtual const Ellipsoid* ellipsoid() const //! access to child Ellipsoid object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _ellipsoid)); return 0; }
    virtual Ellipsoid* ellipsoid() //! access to child Ellipsoid object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _ellipsoid)); return 0; }

    virtual const Ball* ball() const //! access to child Ball object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _ball)); return 0; }
    virtual Ball* ball() //! access to child Ball object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _ball)); return 0; }

    virtual const RevTrunk* revTrunk() const //! access to child RevTrunk object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _revTrunk)); return 0; }
    virtual RevTrunk* revTrunk() //! access to child RevTrunk object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _revTrunk)); return 0; }

    virtual const Trunk* trunk() const //! access to child Trunk object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _trunk)); return 0; }
    virtual Trunk* trunk() //! access to child Trunk object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _trunk)); return 0; }

    virtual const Cylinder* cylinder() const //! access to child Cylinder object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _cylinder)); return 0; }
    virtual Cylinder* cylinder() //! access to child Cylinder object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _cylinder)); return 0; }

    virtual const RevCylinder* revCylinder() const //! access to child RevCylinder object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _revCylinder)); return 0; }
    virtual RevCylinder* revCylinder() //! access to child RevCylinder object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _revCylinder)); return 0; }

    virtual const Prism* prism() const //! access to child Prism object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _prism)); return 0; }
    virtual Prism* prism() //! access to child Prism object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _prism)); return 0; }

    virtual const Cone* cone() const //! access to child Cone object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _cone)); return 0; }
    virtual Cone* cone() //! access to child Cone object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _cone)); return 0; }

    virtual const RevCone* revCone() const //! access to child RevCone object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _revCone)); return 0; }
    virtual RevCone* revCone() //! access to child RevCone object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _revCone)); return 0; }

    virtual const Pyramid* pyramid() const //! access to child Pyramid object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _pyramid)); return 0; }
    virtual Pyramid* pyramid() //! access to child Pyramid object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _pyramid)); return 0; }

    virtual const SetOfElems* setofelems() const //! access to child SetOfElems object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _setofelems)); return 0; }
    virtual SetOfElems* setofelems() //! access to child SetOfElems object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _setofelems)); return 0; }

    virtual const SetOfPoints* setofpoints() const //! access to child SetOfPoints object (const)
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _setofpoints)); return 0; }
    virtual SetOfPoints* setofpoints() //! access to child SetOfPoints object
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _setofpoints)); return 0; }

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a Geometry
    virtual Geometry& transform(const Transformation& t);
    //! apply a translation on a Geometry (vector version)
    virtual Geometry& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a Geometry (3 reals version)
    virtual Geometry& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a Geometry
    virtual Geometry& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a Geometry
    virtual Geometry& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                               real_t angle = 0.);
    //! apply a rotation on a Geometry
    virtual Geometry& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Geometry
    virtual Geometry& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a Geometry
    virtual Geometry& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Geometry
    virtual Geometry& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a Geometry
    virtual Geometry& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a Geometry
    virtual Geometry& homothetize(real_t factor);
    //! apply a point reflection on a Geometry
    virtual Geometry& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a Geometry
    virtual Geometry& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a Geometry
    virtual Geometry& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a Geometry
    virtual Geometry& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a Geometry
    virtual Geometry& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);

    Geometry& crack(CrackType ct, string_t domNameToOpen); //!< force geometry to crack
    Geometry& uncrack(); //!< disable geometry to crack
    Geometry& operator+(); //!< force behavior in union of geometries
    Geometry& operator-(); //!< disable force behavior in union of geometries
    void cleanInclusions(); //!< clean geometries_ so that inclusions are not included by other ones in the list

    // addition of 2 geometries
    Geometry& operator+=(const Geometry& g); //!< union of g1 and g2 (general case)
    friend Geometry addCompositeAndComposite(const Geometry& g1, const Geometry& g2);
    friend Geometry addCompositeAndLoop(const Geometry& g1, const Geometry& g2);
    friend Geometry addCompositeAndCanonical(const Geometry& g1, const Geometry& g2);
    friend Geometry addLoopAndLoop(const Geometry& g1, const Geometry& g2);
    friend Geometry addLoopAndCanonical(const Geometry& g1, const Geometry& g2);
    friend Geometry addCanonicalAndCanonical(const Geometry& g1, const Geometry& g2);
    friend Geometry operator+(const Geometry& g1, const Geometry& g2);

    // substraction of 2 geometries
    Geometry& operator-=(const Geometry& g); //!< definition of a hole (general case)
    friend Geometry substrCompositeAndComposite(const Geometry& g1, const Geometry& g2);
    friend Geometry substrCompositeAndLoop(const Geometry& g1, const Geometry& g2);
    friend Geometry substrCompositeAndCanonical(const Geometry& g1, const Geometry& g2);
    friend Geometry substrLoopAndComposite(const Geometry& g1, const Geometry& g2);
    friend Geometry substrLoopAndLoop(const Geometry& g1, const Geometry& g2);
    friend Geometry substrLoopAndCanonical(const Geometry& g1, const Geometry& g2);
    friend Geometry substrCanonicalAndComposite(const Geometry& g1, const Geometry& g2);
    friend Geometry substrCanonicalAndLoop(const Geometry& g1, const Geometry& g2);
    friend Geometry substrCanonicalAndCanonical(const Geometry& g1, const Geometry& g2);
    friend Geometry operator-(const Geometry& g1, const Geometry& g2);

    friend Geometry surfaceFrom(const Geometry& c, string_t domName, bool isPlaneSurface);
    friend Geometry planeSurfaceFrom(const Geometry& c, string_t domName);
    friend Geometry ruledSurfaceFrom(const Geometry& c, string_t domName);
    friend Geometry volumeFrom(const Geometry& s, string_t domName);
    friend Geometry extrude(const Geometry& g, const Transformation& t, number_t layers, string_t domName,
                            std::vector<string_t> sidenames);
    friend Geometry extrude(const Geometry& g, const Transformation& t, number_t layers, std::vector<string_t> sidenames);
    friend Geometry extrude(const Geometry& g, const Transformation& t, string_t domName, std::vector<string_t> sidenames);
    friend Geometry extrude(const Geometry& g, const Transformation& t, std::vector<string_t> sidenames);

    friend Geometry toComposite(const Geometry& g);
};

std::ostream& operator<<(std::ostream&, const Geometry&); //!< output Geometry

// addition of 2 geometries
Geometry addCompositeAndComposite(const Geometry& g1, const Geometry& g2); //!< union of composite g1 and g2
Geometry addCompositeAndLoop(const Geometry& g1, const Geometry& g2);      //!< union of composite g1 and loop g2
Geometry addCompositeAndCanonical(const Geometry& g1, const Geometry& g2); //!< union of composite g1 and canonical g2
Geometry addLoopAndLoop(const Geometry& g1, const Geometry& g2);           //!< union of loop g1 and g2
Geometry addLoopAndCanonical(const Geometry& g1, const Geometry& g2);      //!< union of loop g1 and canonical g2
Geometry addCanonicalAndCanonical(const Geometry& g1, const Geometry& g2); //!< union of canonical g1 and g2
Geometry operator+(const Geometry& g1, const Geometry& g2);                //!< union of g1 and g2 (general case)
// substraction of 2 geometries
Geometry substrCompositeAndComposite(const Geometry& g1, const Geometry& g2); //!< g2 (composite) hole of g1 (composite)
Geometry substrCompositeAndLoop(const Geometry& g1, const Geometry& g2);      //!< g2 (composite) hole of g1 (loop)
Geometry substrCompositeAndCanonical(const Geometry& g1, const Geometry& g2); //!< g2 (composite) hole of g1 (canonical)
Geometry substrLoopAndComposite(const Geometry& g1, const Geometry& g2);      //!< g2 (loop) hole of g1 (composite)
Geometry substrLoopAndLoop(const Geometry& g1, const Geometry& g2);           //!< g2 (loop) hole of g1 (loop)
Geometry substrLoopAndCanonical(const Geometry& g1, const Geometry& g2);      //!< g2 (loop) hole of g1 (canonical)
Geometry substrCanonicalAndComposite(const Geometry& g1, const Geometry& g2); //!< g2 (canonical) hole of g1 (composite)
Geometry substrCanonicalAndLoop(const Geometry& g1, const Geometry& g2);      //!< g2 (canonical) hole of g1 (loop)
Geometry substrCanonicalAndCanonical(const Geometry& g1, const Geometry& g2); //!< g2 (canonical) hole of g1 (canonical)
Geometry operator-(const Geometry& g1, const Geometry& g2);                   //!< g2 hole of g1 (general case)

Geometry toComposite(const Geometry& g); //!< return a composite geometry with only one component : g

//! definition of a geometry 2D from an union of boundaries 1D
Geometry surfaceFrom(const Geometry& c, string_t domName = "", bool isPlaneSurface = true);
//! definition of a geometry 2D from an union of boundaries 1D
inline Geometry planeSurfaceFrom(const Geometry& c, string_t domName = "")
{ return surfaceFrom(c, domName, true); }
//! definition of a geometry 2D from an union of boundaries 1D
inline Geometry ruledSurfaceFrom(const Geometry& c, string_t domName = "")
{ return surfaceFrom(c, domName, false); }
//! definition of a geometry 3D from an union of boundaries 2D
Geometry volumeFrom(const Geometry& s, string_t domName = "");
//! definition of a geometry by extrusion of another geometry, with name and side names
Geometry extrude(const Geometry& g, const Transformation& t, number_t layers, string_t domName,
                 std::vector<string_t> sidenames = std::vector<string_t>());
//! definition of a geometry by extrusion of another geometry, with name and side name
inline Geometry extrude(const Geometry& g, const Transformation& t, number_t layers, string_t domName, string_t sidenames)
{ return extrude(g, t, layers, domName, std::vector<string_t>(1,sidenames)); }
//! definition of a geometry by extrusion of another geometry, with side names
Geometry extrude(const Geometry& g, const Transformation& t, number_t layers,
                 std::vector<string_t> sidenames = std::vector<string_t>());
//! definition of a geometry by extrusion of another geometry, with name and side names
inline Geometry extrude(const Geometry& g, const Transformation& t, string_t domName,
                        std::vector<string_t> sidenames = std::vector<string_t>())
{ return extrude(g, t, 0, domName, sidenames); }
//! definition of a geometry by extrusion of another geometry, with name and side name
inline Geometry extrude(const Geometry& g, const Transformation& t, string_t domName, string_t sidenames)
{ return extrude(g, t, 0, domName, sidenames); }
//! definition of a geometry by extrusion of another geometry, with side names
inline Geometry extrude(const Geometry& g, const Transformation& t, std::vector<string_t> sidenames = std::vector<string_t>())
{ return extrude(g, t, 0, sidenames); }

//! user shortcut to crack one geometry
void crack(Geometry& g1, CrackType ct=_closedCrack, string_t domNameToOpen = string_t());
//! user shortcut to crack 2 geometries
void crack(Geometry& g1, Geometry& g2, CrackType ct=_closedCrack, string_t domNameToOpen = string_t());
//! user shortcut to crack 3 geometries
void crack(Geometry& g1, Geometry& g2, Geometry& g3, CrackType ct=_closedCrack, string_t domNameToOpen = string_t());
//! user shortcut to crack 4 geometries
void crack(Geometry& g1, Geometry& g2, Geometry& g3, Geometry& g4, CrackType ct=_closedCrack, string_t domNameToOpen = string_t());
//! user shortcut to crack 5 geometries
void crack(Geometry& g1, Geometry& g2, Geometry& g3, Geometry& g4, Geometry& g5, CrackType ct=_closedCrack,
           string_t domNameToOpen = string_t());
//! user shortcut to crack 6 geometries
void crack(Geometry& g1, Geometry& g2, Geometry& g3, Geometry& g4, Geometry& g5, Geometry& g6, CrackType ct=_closedCrack,
           string_t domNameToOpen = string_t());
//! user shortcut to crack 7 geometries
void crack(Geometry& g1, Geometry& g2, Geometry& g3, Geometry& g4, Geometry& g5, Geometry& g6, Geometry& g7,
           CrackType ct=_closedCrack, string_t domNameToOpen = string_t());

//! user shortcut to crack one geometry
inline void openCrack(Geometry& g1, string_t domNameToOpen)
{ crack(g1, _openCrack, domNameToOpen); }
//! user shortcut to crack 2 geometries
inline void openCrack(Geometry& g1, Geometry& g2, string_t domNameToOpen)
{ crack(g1, g2, _openCrack, domNameToOpen); }
//! user shortcut to crack 3 geometries
inline void openCrack(Geometry& g1, Geometry& g2, Geometry& g3, string_t domNameToOpen)
{ crack(g1, g2, g3, _openCrack, domNameToOpen); }
//! user shortcut to crack 4 geometries
inline void openCrack(Geometry& g1, Geometry& g2, Geometry& g3, Geometry& g4, string_t domNameToOpen)
{ crack(g1, g2, g3, g4, _openCrack, domNameToOpen); }
//! user shortcut to crack 5 geometries
inline void openCrack(Geometry& g1, Geometry& g2, Geometry& g3, Geometry& g4, Geometry& g5, string_t domNameToOpen)
{ crack(g1, g2, g3, g4, g5, _openCrack, domNameToOpen); }
//! user shortcut to crack 6 geometries
inline void openCrack(Geometry& g1, Geometry& g2, Geometry& g3, Geometry& g4, Geometry& g5, Geometry& g6, string_t domNameToOpen)
{ crack(g1, g2, g3, g4, g5, g6, _openCrack, domNameToOpen); }
//! user shortcut to crack 7 geometries
inline void openCrack(Geometry& g1, Geometry& g2, Geometry& g3, Geometry& g4, Geometry& g5, Geometry& g6, Geometry& g7,
                      string_t domNameToOpen)
{ crack(g1, g2, g3, g4, g5, g6, g7, _openCrack, domNameToOpen); }

//! user shortcut to crack one geometry
inline void closedCrack(Geometry& g1)
{ crack(g1, _closedCrack); }
//! user shortcut to crack 2 geometries
inline void closedCrack(Geometry& g1, Geometry& g2)
{ crack(g1, g2, _closedCrack); }
//! user shortcut to crack 3 geometries
inline void closedCrack(Geometry& g1, Geometry& g2, Geometry& g3)
{ crack(g1, g2, g3, _closedCrack); }
//! user shortcut to crack 4 geometries
inline void closedCrack(Geometry& g1, Geometry& g2, Geometry& g3, Geometry& g4)
{ crack(g1, g2, g3, g4, _closedCrack); }
//! user shortcut to crack 5 geometries
inline void closedCrack(Geometry& g1, Geometry& g2, Geometry& g3, Geometry& g4, Geometry& g5)
{ crack(g1, g2, g3, g4, g5, _closedCrack); }
//! user shortcut to crack 6 geometries
inline void closedCrack(Geometry& g1, Geometry& g2, Geometry& g3, Geometry& g4, Geometry& g5, Geometry& g6)
{ crack(g1, g2, g3, g4, g5, g6, _closedCrack); }
//! user shortcut to crack 7 geometries
inline void closedCrack(Geometry& g1, Geometry& g2, Geometry& g3, Geometry& g4, Geometry& g5, Geometry& g6, Geometry& g7)
{ crack(g1, g2, g3, g4, g5, g6, g7, _closedCrack); }

//! find if a curve/surf is in a list of curves/surfs
int_t findBorder(std::pair<ShapeType, std::vector<const Point*> > border, std::vector<std::pair<ShapeType,
                 std::vector<const Point*> > > borders);
//! determine if 2 geometries (list of vertices) have the same orientation
bool sameOrientation(std::vector<const Point*>,std::vector<const Point*>);

//================================================
//         transformations facilities
//================================================
//! apply a geometrical transformation on a Geometry (external)
Geometry transform(const Geometry& g, const Transformation& t);
//! apply a translation on a Geometry (vector version) (external)
Geometry translate(const Geometry& g, std::vector<real_t> u = std::vector<real_t>(3,0.));
//! apply a translation on a Geometry (3 reals version) (external)
Geometry translate(const Geometry& g, real_t ux, real_t uy = 0., real_t uz = 0.);
//! apply a rotation on a Geometry (external)
Geometry rotate2d(const Geometry& g, const Point& c = Point(0.,0.), real_t angle = 0.);
//! apply a rotation on a Geometry (external)
Geometry rotate3d(const Geometry& g, const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                  real_t angle = 0.);
//! apply a rotation on a Geometry (external)
Geometry rotate3d(const Geometry& g, real_t dx, real_t dy, real_t angle);
//! apply a rotation on a Geometry (external)
Geometry rotate3d(const Geometry& g, real_t dx, real_t dy, real_t dz, real_t angle);
//! apply a rotation on a Geometry (external)
Geometry rotate3d(const Geometry& g, const Point& c, real_t dx, real_t dy, real_t angle);
//! apply a rotation on a Geometry (external)
Geometry rotate3d(const Geometry& g, const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
//! apply a homothety on a Geometry (external)
Geometry homothetize(const Geometry& g, const Point& c = Point(0.,0.,0.), real_t factor = 1.);
//! apply a homothety on a Geometry (external)
Geometry homothetize(const Geometry& g, real_t factor);
//! apply a point reflection on a Geometry (external)
Geometry pointReflect(const Geometry& g, const Point& c = Point(0.,0.,0.));
//! apply a reflection2d on a Geometry (external)
Geometry reflect2d(const Geometry& g, const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
//! apply a reflection2d on a Geometry (external)
Geometry reflect2d(const Geometry& g, const Point& c, real_t dx, real_t dy = 0.);
//! apply a reflection3d on a Geometry (external)
Geometry reflect3d(const Geometry& g, const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
//! apply a reflection3d on a Geometry (external)
Geometry reflect3d(const Geometry& g, const Point& c, real_t nx, real_t ny, real_t nz = 0.);

} // end of namespace xlifepp

#endif // GEOMETRY_HPP
