/*
XLiFE++ is an extended library of finite elements written in C++
    Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file Mesh.hpp
  \authors D. Martin, Y. Lafranche, E. Lunéville, N. Kielbasiewicz
  \since 26 may 2010
  \date 17 oct 2014

  \brief Definition of the xlifepp::Mesh class

  xlifepp::Mesh represents the main class to deal with mesh data
  either created by xlife++ meshing tools or read from mesh files
  generated by external mesh tools (gmsh, TetMesh-GHS3D, ...)

  This xlifepp::Mesh class supports any kind of 1D, 2D, 3D meshes made of
    - segments, triangles, quadrangles, tetraedron, hexaedron, prisms, pyramids
    - of order 1 (plane elements) 2, 3, ... (curved elements)
    - elements of same dimension can be mixed

  These elements are GeomElements object in xlife++ terminology
  and should not be confused with Element object supporting finite element interpolation in addition
  Note that the dimension of elements may be less than the space dimension

  A Mesh consists of :
    - a vector of points : vertices of elements and additonnal nodes for element of order greater than 1
    - a list of xlifepp::GeomElement : list of vertices, a xlifepp::RefElement describing geometrical interpolation, ...
    - a list of sides (edge in 2D, faces in 3D) seen as xlifepp::GeomElement (xlifepp::MeshSideElement)
    - a list of side of sides (edges in 3D) seen as xlifepp::GeomElement (xlifepp::MeshSideElement)
    - a list of vertices indices
    - a xlifepp::Geometry object collecting global geometric data such as bounding box, boundary parametrisations,...
    - a list of xlifepp::GeomDomain describing physical meaning part of the mesh, in particular boundaries

  Contrary to a xlifepp::MeshElement which carries a lot of data, a MeshSideElement contains only its parents (GeomElement)
  and the related side numbers

  The class manages a pointer (firstOrderMesh_p) to its underlying first order mesh. If the mesh is itself a first order mesh
  (meaning that all elements are first order elements), firstOrderMesh_p is equal to the mesh pointer (this), else firstOrderMesh_p=0
  until its underlying first order mesh is built. The underlying first order mesh is required by some tools visualization
  and therefore it is generally built by mesh export functions.
*/

#ifndef MESH_HPP
#define MESH_HPP

#include "config.h"
#include "GeomElement.hpp"
#include "GeomDomain.hpp"
#include "Geometry.hpp"
#include "geometries_utils.hpp"
#include "geometries1D.hpp"
#include "geometries2D.hpp"
#include "geometries3D.hpp"
#include "subdivision/subUtil/HexahedronMesh.hpp"
#include "subdivision/subUtil/TetrahedronMesh.hpp"
#include "subdivision/subUtil/QuadrangleMesh.hpp"
#include "subdivision/subUtil/TriangleMesh.hpp"
#include "utils.h"

namespace xlifepp
{

/*!
  \class CrackData
  store data related to a crack : domain name, domain id and dimension
 */

class CrackData
{
  public:
    string_t domainName;    //!< name of the domain supporting the crack
    number_t id;            //!< id of the domain supporting the crack
    number_t dim;           //!< dimension of the crack
    MinimalBox minimalBox;  //!< minimal box of the geometry supporting the crack
    number_t domIdToOpen;   //!< stricly positive is open crack
    CrackData()             //! default constructor
    {
      domainName="Crack";
      id=1;
      dim=1;
      domIdToOpen=0;
      minimalBox=MinimalBox();
    }
    //! constructor
    CrackData(string_t name, number_t i, number_t d, const MinimalBox& mb, number_t domId = 0)
    : domainName(name), id(i), dim(d), minimalBox(mb), domIdToOpen(domId) {}
    bool operator<(const CrackData& cd) const; //!< comparison operator
    std::vector<real_t> computeNormal() const; //!< compute the normal vector for a 1D (in plane) or 2D (in space) cracked domain
    void print(std::ostream&) const;                        //!< print mesh data
    void print(PrintStream& os) const {print(os.currentStream());}
    friend std::ostream& operator<<(std::ostream&, const CrackData&);
};

std::ostream& operator<<(std::ostream&, const CrackData&); //!< print operator

//! finds a crack data in a list
short int findId(std::vector<CrackData>::const_iterator it_b, std::vector<CrackData>::const_iterator it_e, number_t id);

/*!
  \class Mesh
  handles mesh data either created by internal mesh tools
  or read from file generated by external mesh tools
 */
class Mesh
{
  public:
    Geometry* geometry_p;                       //!< global geometric information (space dimension, variable names, bounding box, ...)
    std::set<CrackData> cracks;                 //!< list of domains to be cracked (with gmsh generator)
    std::vector<Point> nodes;                   //!< list of all mesh nodes (vertices and other nodes)
    mutable number_t lastIndex_;                //!< last index of GeomElements (= max of id numbers over all GeomElements in the mesh)

  protected:
    string_t name_;                             //!< a name for documentation purposes
    string_t comment_;                          //!< comment documentation purposes ("load from file data.msh",...)
    std::vector<GeomElement*> elements_;        //!< list of geometric elements (MeshElement)
    std::vector<GeomDomain*> domains_;          //!< list of geometric domains (part of the mesh)
    std::vector<number_t> vertices_;            //!< list of vertices indexed by their number in nodes vector
    bool isMadeOfSimplices_;                    //!< only composed of segments (1D mesh), triangles (2D mesh) or tetrahedra (3D mesh)
    dimen_t order_;                             //!< order of the mesh = max of element orders

    // built if required
    std::vector<GeomElement*> sides_;           //!< list of sides (edges in 2D, face in 3D), built if required
    std::vector<GeomElement*> sideOfSides_;     //!< list of sides of sides (edges in 3D), built if required
    std::vector<std::vector<GeoNumPair> > vertexElements_;  //!< for each vertex v, list of elements having vertex v, built if required
    mutable Mesh* firstOrderMesh_p;             //!< underlying first order mesh when mesh has elements of order greater than one

  public:
    // constructors
    Mesh(); //!< void constructor
    //! Mesh from a file of MeshFormat type
    Mesh(const string_t& filename, const string_t& na, IOFormat mft = _undefFormat, number_t nodesDim=0);
    void copyAllButNodes(const Mesh&); //!< deep copy of all mesh attributes, nodes excepted
    void copy(const Mesh&);            //!< deep copy of all mesh attributes
    Mesh(const Mesh& m) { copy(m); }   //!< copy constructor
    Mesh& operator=(const Mesh&);      //!< assign operator
    void clear();                      //!< clear all data and delete pointers
    ~Mesh() { clear(); }               //!< destructor

    void build1DMesh(const Geometry& g, number_t order, MeshGenerator mg);
    void buildMesh(const Geometry& g, ShapeType sh, number_t order, MeshGenerator mg, std::set<MeshOption> mos);
    //! constructor from any 1D geometry
    Mesh(const Geometry& g, number_t order = 1, MeshGenerator mg = _defaultGenerator, const string_t& name = "")
    : name_(name) { build1DMesh(g, order, mg); }
    //! constructor from any 2D and 3D geometries
    Mesh(const Geometry& g, ShapeType sh, number_t order = 1, MeshGenerator mg = _defaultGenerator, const string_t& name = "")
    : name_(name) {
      std::set<MeshOption> mos; mos.insert(_defaultMeshOption);
      buildMesh(g, sh, order, mg, mos);
    }
    //! constructor from any 2D and 3D geometries
    Mesh(const Geometry& g, ShapeType sh, number_t order, MeshGenerator mg, MeshOption mo, const string_t& name = "")
    : name_(name) {
      std::set<MeshOption> mos; mos.insert(mo);
      buildMesh(g, sh, order, mg, mos);
    }
    //! constructor from any 2D and 3D geometries
    Mesh(const Geometry& g, ShapeType sh, number_t order, MeshGenerator mg, MeshOption mo1, MeshOption mo2, const string_t& name = "")
    : name_(name) {
      std::set<MeshOption> mos; mos.insert(mo1); mos.insert(mo2);
      buildMesh(g, sh, order, mg, mos);
    }
    //! constructor from any 2D and 3D geometries
    Mesh(const Geometry& g, ShapeType sh, number_t order, MeshGenerator mg, MeshOption mo1, MeshOption mo2, MeshOption mo3,
         const string_t& name = "")
    : name_(name) {
      std::set<MeshOption> mos; mos.insert(mo1); mos.insert(mo2); mos.insert(mo3);
      buildMesh(g, sh, order, mg, mos);
    }
    //! constructor from a mesh to be subdivided
    Mesh(const Mesh& msh, number_t nbsubdiv, number_t order = 1);
    //! elementary mesh constructor with only one reference element for test purpose
    Mesh(ShapeType shape);
    //! constructor from extrusion
    Mesh(const Mesh& ms, const Point& O, const Point& d , number_t nbl,
         number_t namingDomain=0, number_t namingSection=0, number_t namingSide=0,
         const string_t& meshName="")
    : name_(meshName) {buildExtrusion(ms,O,d,nbl,namingDomain, namingSection, namingSide);}
    //! build new mesh converting type of element of a given mesh, if possible
    Mesh(const Mesh& mesh, ShapeType sh, const string_t name="");

    //! choose mesh generator from Geometry and element shape
    MeshGenerator defaultMeshGenerator(const Geometry&, ShapeType = _noShape);

    //construct additionnal data on demand
    void createSideIndex(std::map<string_t,std::vector<GeoNumPair> >& sideIndex) const //! index all the sides of mesh
    { sideIndex.clear(); xlifepp::createSideIndex(elements_, sideIndex);}
    void createSideEltIndex(std::map<string_t, GeomElement*>& sideEltIndex) const;     //!< index all the side elements of mesh
    void buildSides();                //!< build the sides elements
    void buildSideOfSides();          //!< build the sides of sides elements
    void buildVertexElements();       //!< build for each vertex v, the list of elements having v as vertex
    void buildGeomData();             //!< compute element measures and orientation
    void setShapeTypes();             //!< set the shape types for all mesh domains
    void mergeDomainsWithSameName();  //!< merge all domains having the same name (internal tool)
    void updateSideDomains();         //!< update parent sides of elements involved in all side domains

    //================================================
    //                   accessors
    //================================================
    const std::vector<GeomElement*>& elements() const       //! access to the elements list
      { return elements_; }
    const std::vector<GeomDomain*>& domains()   const       //! access to the domains list
      { return domains_; }
    const GeomDomain& domain(const string_t&) const;        //!< access to a domain by its name (const)
    const GeomDomain& domain(number_t) const;               //!< access to a domain by its number (const)
    GeomDomain& domain(const string_t&);                    //!< access to a domain by its name (no const)
    GeomDomain& domain(number_t);                           //!< access to a domain by its number (no const)
    number_t domainNumber(const string_t&) const;           //!< access to a domain number by its name (const)
    number_t domainNumber(const string_t&);                 //!< access to a domain number by its name (no const)
    string_t domainName(number_t) const;                    //!< access to a domain name by its number (const)
    string_t domainName(number_t);                          //!< access to a domain name by its number (no const)
    const std::vector<GeomElement*>& sides()const           //! access to the sides list
      { return sides_; }
    const std::vector<GeomElement*>& sideOfSides()const     //! access to the sides of sides list
      { return sideOfSides_; }
    const std::vector<number_t>& vertices()const            //! access to the vertices list
      { return vertices_; }
    const std::vector<std::vector<GeoNumPair> >& vertexElements() const //! access to the vertexElements list
      { return vertexElements_; }
    const string_t& name() const                            //! access to the mesh name
      { return name_; }
    void name(const string_t& n)                                     //! set to the mesh name
      { name_=n; }
    bool isMadeOfSimplices()const                           //! true if the mesh is composed only with simplices
      { return isMadeOfSimplices_; }
    const GeomElement& element(number_t i) const            //! access to the i-th element (i>0; no control)
      { return *elements_[i - 1]; }
    const GeomDomain& domains(number_t i) const             //! access to the i-th domain (i>0; no control)
      { return *domains_[i - 1]; }
    const GeomElement& side(number_t i) const               //! access to the i-th side (i>0, no control)
      { return *sides_[i - 1]; }
    const GeomElement& sideOfSide(number_t i) const         //! access to the i-th side of side (i>0, no control)
      { return *sideOfSides_[i - 1]; }
    number_t vertex(number_t i) const                       //! access to the i-th vertex as a MeshSideElement (i>0, no control)
      { return vertices_[i - 1]; }
    dimen_t spaceDim() const                                //! return the dimension of physical space (dimension of node)
      { if(nodes.size()>0) return nodes[0].size(); else return 0; }
    dimen_t geometryDim() const                             //! return the dimension of geometry
      { return geometry_p->dim(); }
    dimen_t meshDim() const                                 //! return the dimension of mesh element (may be less than space dimension)
      { return elements_[0]->elementDim(); }
    number_t nbOfElements() const                           //!return the number of elements
      { return elements_.size(); }
    number_t nbOfDomains() const                            //! return the number of domains
      { return domains_.size(); }
    number_t nbOfSides() const                              //! return the number of sides
      { return sides_.size(); }
    number_t nbOfSideOfSides() const                        //! return the number of side of sides
      { return sideOfSides_.size(); }
    number_t nbOfVertices() const                           //! return the number of vertices
      { return vertices_.size(); }
    number_t nbOfNodes() const                              //! return the number of nodes
      { return nodes.size(); }
    dimen_t order() const                                   //! return the order of the mesh (max of element orders)
      { return order_; }
    Mesh& firstOrderMesh()                                  //! return underlying first order mesh as reference
      { return *firstOrderMesh_p; }
    const Mesh& firstOrderMesh() const                      //! return underlying first order mesh as const reference
      { return *firstOrderMesh_p; }

    //================================================
    //                  mesh tools                  
    //================================================
    Mesh* createFirstOrderMesh() const;                       //!< create underlying first order mesh if necessary
    void merge(const Mesh&, bool mergeSharedBoundary = true, const string_t& nd="main domain"); //!< "add" mesh to the current one

    void renameDomain(number_t n, const string_t& newname)               //! rename a domain given by its number
      { domain(n).rename(newname); }
    void renameDomain(const string_t& oldname, const string_t& newname)  //! rename a domain given by its name
      { domain(oldname).rename(newname); }
    void removeDomain(const string_t&);                       //!< remove a domain given by its name, use with cautious

    //================================================
    //               print facilities                  
    //================================================
    void printInfo(std::ostream& os=std::cout) const;            //!< print general informations about what the mesh is made of
    void printInfo(PrintStream& os) const
     {printInfo(os.currentStream());}
    void print(std::ostream&) const;                             //!< print mesh data
    void print(PrintStream& os) const {print(os.currentStream());}
    void addSuffix(const string_t& s);                           //!< add a suffix to all names (mesh name and domain names)
    friend std::ostream& operator<<(std::ostream&, const Mesh&);

    //================================================
    //           input/output facilities                  
    //================================================
    void saveToFile(const string_t &, IOFormat mf=_undefFormat, bool withDomains= false) const; //!< save mesh to file
    void saveToVtk(const string_t& fname, bool withDomains) const;       //!< save to vtk file format
    void vtkExport(std::ostream& out) const;                             //!< export mesh as vtk file format to ostream
    void vtkExport(const GeomDomain& dom, std::ostream& out) const;      //!< export mesh domain as vtk file format
    void saveToMsh(const string_t& fname, bool withDomains=false) const; //!< save to msh file format
    void mshExport(std::ostream& out) const;                             //!< export mesh as msh file format to ostream
    void mshExport(const GeomDomain& dom, std::ostream& out) const;      //!< export mesh domain as msh file format
    void saveToMel(const string_t& fname, bool withDomains) const;       //!< save to mel file format
    void melExport(std::ostream& out) const;                             //!< export mesh as mel file format to ostream
    void exportNodes(const string_t&) const;                             //!< export nodes to a file

    //================================================
    //         transformations facilities                  
    //================================================
    //! apply a geometrical transformation on a Mesh
    Mesh& transform(const Transformation& t);
    //! apply a translation on a Mesh (vector version)
    Mesh& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a Mesh (3 reals version)
    Mesh& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a Mesh
    Mesh& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a Mesh
    Mesh& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> u = std::vector<real_t>(3,0.), real_t angle = 0.);
    //! apply a rotation on a Mesh
    Mesh& rotate3d(real_t ux, real_t uy, real_t angle);
    //! apply a rotation on a Mesh
    Mesh& rotate3d(real_t ux, real_t uy, real_t uz, real_t angle);
    //! apply a rotation on a Mesh
    Mesh& rotate3d(const Point& c, real_t ux, real_t uy, real_t angle);
    //! apply a rotation on a Mesh
    Mesh& rotate3d(const Point& c, real_t ux, real_t uy, real_t uz, real_t angle);
    //! apply a homothety on a Mesh
    Mesh& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a Mesh
    Mesh& homothetize(real_t factor);
    //! apply a point reflection on a Mesh
    Mesh& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a Mesh
    Mesh& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> u = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a Mesh
    Mesh& reflect2d(const Point& c, real_t ux, real_t uy = 0.);
    //! apply a reflection3d on a Mesh
    Mesh& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a Mesh
    Mesh& reflect3d(const Point& c, real_t ux, real_t uy, real_t uz = 0.);

  private:
    //================================================
    // build functions using a structured algorithm
    //================================================

    //! Pk regular mesh of a segment
    void meshP1Segment(Segment&, number_t, const std::vector<string_t>&);
    //! Pk regular mesh of a parallelogram
    void meshP1Parallelogram(Parallelogram&, number_t, number_t, const std::vector<string_t>&);
    //! Qk regular mesh of a parallelogram
    void meshQ1Parallelogram(Parallelogram&, number_t, number_t, const std::vector<string_t>&);
    //! P1 regular mesh of a parallelepiped
    void meshP1Parallelepiped(Parallelepiped& , number_t, number_t, number_t, const std::vector<string_t>&);
    //! Q1 regular mesh of a parallelepiped
    void meshQ1Parallelepiped(Parallelepiped& , number_t, number_t, number_t, const std::vector<string_t>&);
    //! Prisme regular mesh of a parallelepiped
    void meshPr1Parallelepiped(Parallelepiped& , number_t, number_t, number_t, const std::vector<string_t>&);
    //! Pyramid regular mesh of a parallelepiped
    void meshPy1Parallelepiped(Parallelepiped& , number_t, number_t, number_t, const std::vector<string_t>&);

    //! P1 mesh of a polygonal line
    void meshP1Line(SetOfPoints& sp);
    //! Utility function used to build 1D mesh
    void complete1Dmesh(const string_t& domname, const std::vector<string_t>& sideNames);

    //================================================
    // build functions using a subdivision algorithm
    //================================================

    //! Pk mesh of a (part of a) sphere (surface or volume)
    void subdvMesh(Ball& sph, const ShapeType shape, const int nboctants,
                   const number_t nbsubdiv=0, const number_t order=1,
                   const number_t type=1, const string_t& TeXFilename = "");

    //! Pk mesh of a (part of a) cube (volume)
    void subdvMesh(Cube& cube, const ShapeType shape, const int nboctants,
                   const number_t nbsubdiv=0, const number_t order=1,
                   const string_t& TeXFilename = "");

    //! Pk mesh of a truncated cone, a cone or a cylinder (volume)
    void subdvMesh(RevTrunk& cyl, const ShapeType shape, const number_t nbSubDomains=1,
                   const number_t nbsubdiv=0, const number_t order=1,
                   const number_t type=1, const string_t& TeXFilename = "");

    //! Pk or Qk mesh of a portion of disk (surface)
    void subdvMesh(Disk& carc, const ShapeType shape,
                   const number_t nbsubdiv, const number_t order,
                   const number_t type, const string_t& TeXFilename = "");

    //! Pk or Qk mesh starting from an initial mesh (surface or volume)
    void subdvMesh(const std::vector<Point>& pts, const std::vector<std::vector<number_t> >& elems,
                   const std::vector<std::vector<number_t> >& bounds, const ShapeType shape,
                   const number_t nbsubdiv=0, const number_t order=1, const string_t& TeXFilename = "");

  protected:
    //! Construction of a mesh of order k in the volume of a 3D domain from a mesh built with the help of a subdivision algorithm
    template <class ST_> void build3Delements(subdivision::GeomFigureMesh<ST_> *TM_p, const ShapeType elShape, number_t nbsubdom = 1);
    //! Construction of a mesh of order k in 2D or over a surface in R3 from a mesh built with the help of a subdivision algorithm
    template <class ST_> void build2Delements(subdivision::GeomFigureMesh<ST_> *TM_p, const ShapeType elShape, number_t nbsubdom = 1);
    //! Construction of an extruded mesh from section mesh
    void buildExtrusion(const Mesh& ms, const Point& O, const Point& d , number_t nbl,
                        number_t namingDomain=0, number_t namingSection=0, number_t namingSide=0);
    //! Construction of an pyramid mesh from hexaedron mesh
    void buildPyramidFromHexadron(const Mesh& hexMesh);
    //! Compute the bounding box containing the mesh
    const std::vector<RealPair> computeBB();

  private:
    //! Utilitary function to copy information from subdivision mesh to XLiFE++
    template <class ST_> void copyPtsEltsDoms(subdivision::GeomFigureMesh<ST_> *TM_p, const ShapeType elShape,
                                              const dimen_t elementDim, const dimen_t spaceDim, number_t& el_k,
                                              const std::vector<string_t>& subdomainNames, const string_t& domName);
    //! Manages construction of domains of a mesh computed using a subdivision algorithm.
    void manageDomains(subdivision::SubdivisionMesh *SM_p, number_t nbsubdom, dimen_t elementDim,
                       std::vector<string_t>& boundaryNames, std::vector<string_t>& interfaceNames,
                       std::vector<string_t>& subdomainNames, string_t& domName);

    //================================================
    //           input/output facilities
    //================================================
    //! Read mesh from input file filename according to Gmsh mesh format version 2.2 at least
    void loadGmsh(const string_t& filename, number_t nodesDim=0);
    //! Read mesh from input file filename according to Gmsh geometry file format (calls and runs gmsh)
    void loadGeo(const string_t& filename, number_t nodesDim=0);
    //! Read mesh from input file filename according to Medit Format
    void loadMedit(const string_t& filename, number_t nodesDim=0);
    //! Read mesh from input file filename according to Melina mesh format
    void loadMelina(const string_t& filename, number_t nodesDim=0);
    //! Read mesh from input file filename according to VTK mesh format
    void loadVtk(const string_t& filename, number_t nodesDim=0);
    //! loads mesh from input stream according to POLYDATA dataset in vtk mesh format
    dimen_t loadPolyDataVtk(std::istream& data, number_t nodesDim=0);
    //! loads mesh from input stream according to RECTILINEAR_GRID dataset in vtk mesh format
    dimen_t loadRectilinearVtk(std::istream& data, number_t nodesDim=0);
    //! loads mesh from input stream according to STRUCTURED_GRID dataset in vtk mesh format
    dimen_t loadStructuredGridVtk(std::istream& data, number_t nodesDim=0);
    //! loads mesh from input stream according to UNSTRUCTURED_GRID dataset in vtk mesh format
    dimen_t loadUnstructuredGridVtk(std::istream& data, number_t nodesDim=0);
    //! Read mesh from input file filename according to VTP mesh format
    void loadVtp(const string_t& filename, number_t nodesDim=0);
    //! Read mesh from input file filename according to VTR mesh format
    void loadVtr(const string_t& filename, number_t nodesDim=0);
    //! Read mesh from input file filename according to VTS mesh format
    void loadVts(const string_t& filename, number_t nodesDim=0);
    //! Read mesh from input file filename according to VTU mesh format
    void loadVtu(const string_t& filename, number_t nodesDim=0);
    //! Read mesh from input file filename according to PLY mesh format
    void loadPly(const string_t& filename, number_t nodesDim=0);

    //! writing a 1D geometry in a geo file
    void saveToGeo(Geometry& g, number_t order, const string_t& filename);
    //! writing a 2D/3D geometry in a geo file
    void saveToGeo(Geometry& g, ShapeType sh, number_t order, std::set<MeshOption> mos, const string_t& filename);

  public:
    //! adaptative domains for levelset methods
    void adaptDomains(number_t mainDomNum, const std::vector<number_t>& elementsColorMap);
};

std::ostream& operator<<(std::ostream&, const Mesh&); //!< print operator

//! set the characteristic length
inline void setDefaultCharacteristicLength(real_t h) { theDefaultCharacteristicLength=h; }

//! merge two meshes in a new one
Mesh merge(const Mesh&, const Mesh&);

//! returns the id msh number of a shape, according to the finite elements order
number_t mshType(ShapeType sht, number_t order);

//@{
//! input/output function
inline void saveToFile(const string_t& fname, const Mesh& m, IOFormat iof=_undefFormat, bool withDomains=false)
{ m.saveToFile(fname, iof, withDomains); }
void saveToFile(const string_t&, const GeomDomain&, IOFormat iof=_undefFormat);
//@}

//! export a split mesh of a domain to msh format
void mshExport(const GeomDomain& dom, const std::vector<Point>& coords,
               const splitvec_t& elementsInfo, std::ostream& out);
//! export a split mesh of a domain to Matlab - Octave format
void mtlbExport(const GeomDomain& dom, const std::vector<Point>& coords,
                const splitvec_t& elementsInfo, std::ostream& out);
//! export a split mesh of a domain to vtk format
void vtkExport(const GeomDomain& dom, const std::vector<Point>& coords,
               const splitvec_t& elementsInfo, std::ostream& out);

//================================================
//         transformations facilities
//================================================
//! apply a geometrical transformation on a Mesh (external)
Mesh transform(const Mesh& m, const Transformation& t);
//! apply a translation on a Mesh (vector version) (external)
Mesh translate(const Mesh& m, std::vector<real_t> u = std::vector<real_t>(3,0.));
//! apply a translation on a Mesh (3 reals version) (external)
Mesh translate(const Mesh& m, real_t ux, real_t uy = 0., real_t uz = 0.);
//! apply a rotation on a Mesh (external)
Mesh rotate2d(const Mesh& m, const Point& c = Point(0.,0.), real_t angle = 0.);
//! apply a rotation on a Mesh (external)
Mesh rotate3d(const Mesh& m, const Point& c = Point(0.,0.,0.), std::vector<real_t> u = std::vector<real_t>(3,0.), real_t angle = 0.);
//! apply a rotation on a Mesh (external)
Mesh rotate3d(const Mesh& m, real_t ux, real_t uy, real_t angle);
//! apply a rotation on a Mesh (external)
Mesh rotate3d(const Mesh& m, real_t ux, real_t uy, real_t uz, real_t angle);
//! apply a rotation on a Mesh (external)
Mesh rotate3d(const Mesh& m, const Point& c, real_t ux, real_t uy, real_t angle);
//! apply a rotation on a Mesh (external)
Mesh rotate3d(const Mesh& m, const Point& c, real_t ux, real_t uy, real_t uz, real_t angle);
//! apply a homothety on a Mesh (external)
Mesh homothetize(const Mesh& m, const Point& c = Point(0.,0.,0.), real_t factor = 1.);
//! apply a homothety on a Mesh (external)
Mesh homothetize(const Mesh& m, real_t factor);
//! apply a point reflection on a Mesh (external)
Mesh pointReflect(const Mesh& m, const Point& c = Point(0.,0.,0.));
//! apply a reflection2d on a Mesh (external)
Mesh reflect2d(const Mesh& m, const Point& c = Point(0.,0.), std::vector<real_t> u = std::vector<real_t>(2,0.));
//! apply a reflection2d on a Mesh (external)
Mesh reflect2d(const Mesh& m, const Point& c, real_t ux, real_t uy = 0.);
//! apply a reflection3d on a Mesh (external)
Mesh reflect3d(const Mesh& m, const Point& c = Point(0.,0.,0.), std::vector<real_t> u = std::vector<real_t>(3,0.));
//! apply a reflection3d on a Mesh (external)
Mesh reflect3d(const Mesh& m, const Point& c, real_t ux, real_t uy, real_t uz = 0.);

} // end of namespace xlifepp

#endif // MESH_HPP
