/*
XLiFE++ is an extended library of finite elements written in C++
    Copyright (C) 2014  Lunville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file Parametrization.hpp
  \authors E. Lunville
  \since 10 jan 2017
  \date 10 jan 2017

  \brief Definition of the xlifepp::Parametrization class

  The Parametrization class handles a general parametrization
  that is an application p from a geometric domain d to the space R^n
  It mainly manages
      geometry_p : a pointer to a Geometry object d
      fun_p      : a pointer to a function of the form Point& f(const Point&, DiffOpType)
      par_fun_p  : a pointer to a function of the form Point& f(const Point&, Parameters&, DiffOpType)
        only one of the function pointers should be allocated

  besides it manages the data:
      name : the name of the parametrization
      dim  : the dimension of the arrival space (initialized at construction by calling once the parametrization function)

  and offers
      some constructors
      the computation of parametrization at a point using operator (), may be with a differential operator of dim 1 (_dx,_dy,_dz, ...)
  Example :
    Point& myfun(const Point& P, DiffOpType dif=_id)
    {
     Real t=P(1);
     return Point(cos(t),sin(t));  //unit circle
    }
    Parametrization pcircle(0,2*pi_,myfun);
    Point q=pcircle(pi_);
 */

#ifndef PARAMETRIZATION_HPP_INCLUDED
#define PARAMETRIZATION_HPP_INCLUDED

#include "config.h"
#include "Geometry.hpp"
#include "geometries_utils.hpp"
#include "geometries1D.hpp"
#include "geometries2D.hpp"
#include "geometries3D.hpp"

namespace xlifepp
{
typedef Point(*par_fun_t)(const Point&, DiffOpType);                    //!< alias of parametrization function
typedef Point(*par_fun_par_t)(const Point&, Parameters&, DiffOpType);   //!< alias of parametrization function with a Parameters

//===========================================================================================
//                  Parametrization class handling any parametrization
//===========================================================================================
class Parametrization
{
  protected:
    Geometry* geom_p;         //!< pointer to the geometry support
    par_fun_t fun_p;          //!< pointer to the parametrization function
    par_fun_par_t fun_par_p;  //!< pointer to the parametrization function with parameters
  public:
    string_t name;            //!< a parametrization name useful for printing
    dimen_t dim;              //!< dimension of the arrival space, set by init()
    Parameters params;        //!< optional parameter list

    //constructors
    Parametrization()
      : geom_p(0), fun_p(0), fun_par_p(0), name(""), dim(0) {}
    Parametrization(Geometry* geo, par_fun_t, string_t na="");                            //!< parametrization from a geometry and a function
    Parametrization(Geometry* geo, par_fun_par_t, const Parameters&, string_t na="");     //!< parametrization from a geometry and a fonction with parameters
    Parametrization(real_t a, real_t b, par_fun_t, string_t na="");                       //!< 1d parametrization from segment [a,b]
    Parametrization(real_t a, real_t b, par_fun_par_t, const Parameters&, string_t na="");//!< 1d parametrization from segment [a,b], fonction with parameters
    Parametrization(real_t a, real_t b, real_t c, real_t d, par_fun_t, string_t na="");   //!< 2d parametrization from rectangle [a,b]x[c,d]
    Parametrization(real_t a, real_t b, real_t c, real_t d, par_fun_par_t, const Parameters&, string_t na="");//!< 2d parametrization from rectangle [a,b]x[c,d], fonction with parameters

    void init();     //!< initialization, set dim

    //operations
    Point operator()(const Point& p, DiffOpType d=_id)
    {
      if(fun_p!=0) return fun_p(p,d);
      else if(fun_par_p!=0) return fun_par_p(p,params,d);
      error("null_pointer","fun_p/fun_par_p");
      return Point(); // dummy return
    }
};

//===========================================================================================
//                  Parametrizations class handling a list of parametrizations
//===========================================================================================
/*
    to deal with complex parametrization made of patches
*/
class Parametrizations: std::vector<Parametrization*>
{
public:
    Parametrizations(){}
    Parametrizations(Parametrization* p1){push_back(p1);}
    Parametrizations(Parametrization* p1, Parametrization* p2){push_back(p1);push_back(p2);}

    ~Parametrizations()
    {
        std::vector<Parametrization*>::iterator it=begin();
        for(;it!=end();++it)
            if(*it!=0) delete *it;
    }

    friend Parametrizations& operator<<(Parametrizations& ps,const Parametrization& p)
    {
        ps.push_back(new Parametrization(p));
        return ps;
    }

    Point operator()(number_t i, const Point& p, DiffOpType d=_id)  //index i starts at 0
    {
        if(i<size()) return (*(*this)[i])(p,d);
        where("Parametrizations::operator()(Number, Point, DiffOpType)");
        error("index_out_of_range","Parametrizations", 1, size());
        return Point();   //fake return
    }

};

} // end of namespace xlifepp

#endif // PARAMETRIZATION_HPP_INCLUDED
