/*
  XLiFE++ is an extended library of finite elements written in C++
  Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file geometries1D.hpp
  \authors Y. Lafranche, N. Kielbasiewicz
  \since 18 oct 2012
  \date 30 jul 2015

  \brief Definition of the classes corresponding to 1D canonical geometrical shapes.

*/

#ifndef GEOMETRIES_1D_HPP
#define GEOMETRIES_1D_HPP

#include "Geometry.hpp"

namespace xlifepp
{

/*!
  \class Curve
  base class for 1D geometries
*/
class Curve : public Geometry
{
  public:
    //! default constructor
    Curve();

  protected:
    //@{
    //! true constructor functions
    void buildParam(const Parameter& p);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    Curve(const Curve& c) : Geometry(c) {} //!< copy constructor
    virtual Geometry* clone() const { return new Curve(*this); } //!< virtual copy constructor for Geometry
    virtual ~Curve() {} //!< virtual destructor

    //! format as string
    virtual string_t asString() const { error("shape_not_handled", words("shape",shape_)); return string_t(); }

    //! reverse orientation of curve
    Curve reverse() const { error("shape_not_handled", words("shape",shape_)); return Curve();}

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a Curve
    virtual Curve& transform(const Transformation& t);
    //! apply a translation on a Curve (vector version)
    virtual Curve& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a Curve (3 reals version)
    virtual Curve& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a Curve
    virtual Curve& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a Curve
    virtual Curve& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                            real_t angle = 0.);
    //! apply a rotation on a Curve
    virtual Curve& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Curve
    virtual Curve& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a Curve
    virtual Curve& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Curve
    virtual Curve& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a Curve
    virtual Curve& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a Curve
    virtual Curve& homothetize(real_t factor);
    //! apply a point reflection on a Curve
    virtual Curve& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a Curve
    virtual Curve& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a Curve
    virtual Curve& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a Curve
    virtual Curve& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a Curve
    virtual Curve& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

/*!
  \class Segment
  definition of a segment geometry in R^3
 */
class Segment : public Curve
{
  private:
    //! first vertex of the segment
    mutable Point p1_;
    //! second vertex of the segment
    mutable Point p2_;
    //! number of nodes on the segment
    number_t n_;
    //! local steps on vertices
    std::vector<real_t> h_;

  public:
    //! default constructor
    Segment();
    //! constructor with 2 Parameter
    Segment(const Parameter& p1, const Parameter& p2);
    //! constructor with 3 Parameter
    Segment(const Parameter& p1, const Parameter& p2, const Parameter& p3);
    //! constructor with 4 Parameter
    Segment(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4);
    //! constructor with 5 Parameter
    Segment(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4,
            const Parameter& p5);

  private:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~Segment() {}

    //@{
    //! accessor to point
    Point p1() const {return p1_; }
    Point p2() const {return p2_; }
    //@}

    //! accessor to number of nodes on segment (read only)
    number_t n(number_t i=0) const { return n_;}
    //! accessor to number of nodes on segment (read/write)
    number_t& n(number_t i) { return n_;}
    //! return the local steps on the segment
    std::vector<real_t> h() const {return h_;}
    //! return the local step on the first vertex
    real_t h1() const {return h_[0];}
    //! return the local step on the second vertex
    real_t h2() const {return h_[1];}

    //! format as string
    string_t asString() const;

    //! reverse orientation of segment
    Segment& reverse();

    virtual Geometry* clone() const { return new Segment(*this); } //!< virtual copy constructor for Geometry

    virtual void computeMB() { minimalBox= MinimalBox(p1_,p2_); } //!< computes the minimal box

    bool withNnodes() const { return (h_.size() == 0); } //!< check if Segment is defined only with _nnodes or with _hsteps option 
    virtual std::vector<const Point*> boundNodes() const; //!< returns list of points on boundary (const)
    virtual std::vector<Point*> nodes(); //!< returns list of every node (non const)
    virtual std::vector<const Point*> nodes() const; //!< returns list of every node (const)
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > curves() const; //!< returns list of curves (const)
    virtual number_t nbSides() const { return 2; } //!< returns the number of sides
    real_t measure() const { return p1_.distance(p2_); }

    //! access to child Segment object (const)
    virtual const Segment* segment() const {return this;}
    //! access to child Segment object
    virtual Segment* segment() {return this;}

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a Segment
    virtual Segment& transform(const Transformation& t);
    //! apply a translation on a Segment (vector version)
    virtual Segment& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a Segment (3 reals version)
    virtual Segment& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a Segment
    virtual Segment& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a Segment
    virtual Segment& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.);
    //! apply a rotation on a Segment
    virtual Segment& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Segment
    virtual Segment& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a Segment
    virtual Segment& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Segment
    virtual Segment& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a Segment
    virtual Segment& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a Segment
    virtual Segment& homothetize(real_t factor);
    //! apply a point reflection on a Segment
    virtual Segment& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a Segment
    virtual Segment& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a Segment
    virtual Segment& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a Segment
    virtual Segment& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a Segment
    virtual Segment& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

Segment operator~(const Segment& s); //!< copy a Segment and reverse its orientation

/*!
  \class EllArc
  definition of an elliptic arc geometry in R^3 (curve)
 */
class EllArc : public Curve
{
  protected:
    //! center of the ellipse arc
    Point c_;
    //! first apogee of the ellipse arc
    Point a_;
    //! first vertex of the arc
    Point p1_;
    //! second vertex of the arc
    Point p2_;
    //! number of nodes on the arc
    number_t n_;
    //! local steps on vertices
    std::vector<real_t> h_;

  public:
    //! default constructor with side names
    EllArc();
    //! constructor with 3 Parameter
    EllArc(const Parameter& p1, const Parameter& p2, const Parameter& p3);
    //! constructor with 4 Parameter
    EllArc(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4);
    //! constructor with 5 Parameter
    EllArc(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4,
            const Parameter& p5);
    //! constructor with 6 Parameter
    EllArc(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4,
            const Parameter& p5, const Parameter& p6);
    //! constructor with 7 Parameter
    EllArc(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4,
            const Parameter& p5, const Parameter& p6, const Parameter& p7);

  private:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& p);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~EllArc() {}

    //@{
    //! accessor to point
    Point center() const { return c_; }
    Point apogee() const { return a_; }
    Point p1() const { return p1_; }
    Point p2() const { return p2_; }
    //@}

    //! accessor to number of nodes on ellipse arc (read only)
    number_t n(number_t i=0) const { return n_;}
    //! accessor to number of nodes on ellipse arc (read/write)
    number_t& n(number_t i) { return n_;}
    //! return the local steps on the ellipse arc
    std::vector<real_t> h() const {return h_;}
    //! return the local step on the first vertex
    real_t h1() const {return h_[0];}
    //! return the local step on the second vertex
    real_t h2() const {return h_[1];}

    //! format as string
    virtual string_t asString() const;

    //! reverse orientation of arc
    EllArc& reverse();

    virtual Geometry* clone() const { return new EllArc(*this); } //!< virtual copy constructor for Geometry

    bool withNnodes() const { return (h_.size() == 0); } //!< check if EllArc is defined only with _nnodes or with _hsteps option 

    virtual std::vector<const Point*> boundNodes() const; //!< returns list of points on boundary (const)
    virtual std::vector<Point*> nodes(); //!< returns list of every point (non const)
    virtual std::vector<const Point*> nodes() const; //!< returns list of every point (const)
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > curves() const; //!< returns list of curves (const)
    virtual number_t nbSides() const { return 2; } //!< returns the number of sides

    //! computes the minimal box
    virtual void computeMB();
    //! compute the bounding box
    virtual void computeBB();

    //! access to child EllArc object (const)
    virtual const EllArc* ellArc() const {return this;}
    //! access to child EllArc object
    virtual EllArc* ellArc() {return this;}

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a EllArc
    virtual EllArc& transform(const Transformation& t);
    //! apply a translation on a EllArc (vector version)
    virtual EllArc& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a EllArc (3 reals version)
    virtual EllArc& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a EllArc
    virtual EllArc& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a EllArc
    virtual EllArc& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                             real_t angle = 0.);
    //! apply a rotation on a EllArc
    virtual EllArc& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a EllArc
    virtual EllArc& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a EllArc
    virtual EllArc& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a EllArc
    virtual EllArc& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a EllArc
    virtual EllArc& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a EllArc
    virtual EllArc& homothetize(real_t factor);
    //! apply a point reflection on a EllArc
    virtual EllArc& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a EllArc
    virtual EllArc& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a EllArc
    virtual EllArc& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a EllArc
    virtual EllArc& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a EllArc
    virtual EllArc& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

EllArc operator~(const EllArc& e); //!< copy a EllArc and reverse its orientation

/*!
  \class CircArc
  definition of a circular arc geometry in R^3 (curve)
 */
class CircArc : public Curve
{
  protected:
    //! center of the circle arc
    Point c_;
    //! first vertex of the arc
    Point p1_;
    //! second vertex of the arc
    Point p2_;
    //! number of nodes on the arc
    number_t n_;
    //! local steps on vertices
    std::vector<real_t> h_;

  public:
    //! default constructor with side names
    CircArc();
    //! constructor with 3 Parameter
    CircArc(const Parameter& p1, const Parameter& p2, const Parameter& p3);
    //! constructor with 4 Parameter
    CircArc(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4);
    //! constructor with 5 Parameter
    CircArc(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4,
            const Parameter& p5);
    //! constructor with 6 Parameter
    CircArc(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4,
            const Parameter& p5, const Parameter& p6);

  private:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& p);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~CircArc() {}

    //@{
    //! accessor to point
    Point center() const { return c_; }
    Point p1() const { return p1_; }
    Point p2() const { return p2_; }
    //@}

    //! accessor to number of nodes on circular arc (read only)
    number_t n(number_t i=0) const { return n_;}
    //! accessor to number of nodes on circular arc (read/write)
    number_t& n(number_t i) { return n_;}

    //! return the local steps on the circular arc
    std::vector<real_t> h() const {return h_;}
    //! return the local step on the first vertex
    real_t h1() const {return h_[0];}
    //! return the local step on the second vertex
    real_t h2() const {return h_[1];}

    //! format as string
    virtual string_t asString() const;

    //! reverse orientation of arc
    CircArc& reverse();

    virtual Geometry* clone() const { return new CircArc(*this); } //!< virtual copy constructor for Geometry

    bool withNnodes() const { return (h_.size() == 0); } //!< check if CircArc is defined only with _nnodes or with _hsteps option 

    virtual std::vector<const Point*> boundNodes() const; //!< returns list of points on boundary (const)
    virtual std::vector<Point*> nodes(); //!< returns list of every point (non const)
    virtual std::vector<const Point*> nodes() const; //!< returns list of every point (const)
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > curves() const; //!< returns list of curves (const)
    virtual number_t nbSides() const { return 2; } //!< returns the number of sides
    real_t measure() const;

    //! computes the minimal box
    virtual void computeMB();
    //! computes the bounding box
    virtual void computeBB();

    //! access to child CircArc object (const)
    virtual const CircArc* circArc() const {return this;}
    //! access to child CircArc object
    virtual CircArc* circArc() {return this;}

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a CircArc
    virtual CircArc& transform(const Transformation& t);
    //! apply a translation on a CircArc (vector version)
    virtual CircArc& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a CircArc (3 reals version)
    virtual CircArc& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a CircArc
    virtual CircArc& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a CircArc
    virtual CircArc& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.);
    //! apply a rotation on a CircArc
    virtual CircArc& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a CircArc
    virtual CircArc& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a CircArc
    virtual CircArc& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a CircArc
    virtual CircArc& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a CircArc
    virtual CircArc& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a CircArc
    virtual CircArc& homothetize(real_t factor);
    //! apply a point reflection on a CircArc
    virtual CircArc& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a CircArc
    virtual CircArc& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a CircArc
    virtual CircArc& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a CircArc
    virtual CircArc& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a CircArc
    virtual CircArc& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

CircArc operator~(const CircArc& e); //!< copy a CircArc and reverse its orientation

/*!
   \class SetOfPoints
   This class defines a set of points. The points are implicitly ordered according to their position in the vector.
   This allows the definition of a polygonal line made of successive segments from arbitrary points in 1D, 2D or 3D.
*/
class SetOfPoints : public Geometry {
  private:
    std::vector<Point> pts_; //!< list of points

  public:
    //! default constructor
    SetOfPoints(const std::vector<Point>& pts, const std::vector<string_t>& names)
    : Geometry(BoundingBox(pts), BoundingBox(pts).dim(), string_t("Omega"), _setofpoints), pts_(pts)
    { sideNames_ = names; }

    //! format as string
    virtual string_t asString() const;

    virtual Geometry* clone() const { return new SetOfPoints(*this); } //!< virtual copy constructor

    virtual std::vector<Point*> nodes(); //!< list of every point (non const)
    virtual std::vector<const Point*> nodes() const; //!< list of every point (const)
    real_t measure() const;

    // access functions to data members
    std::vector<Point>& pts() { return pts_; } //!< return list of points (non const)
    const std::vector<Point>& pts() const { return pts_; } //!< returns list of points (const)

    //! access to child SetOfPoints object (const)
    virtual const SetOfPoints* setofpoints() const {return this;}
    //! access to child SetOfPoints object
    virtual SetOfPoints* setofpoints() {return this;}
/*
    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a SetOfPoints
    virtual SetOfPoints& transform(const Transformation& t);
    //! apply a translation on a SetOfPoints (vector version)
    virtual SetOfPoints& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a SetOfPoints (3 reals version)
    virtual SetOfPoints& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a SetOfPoints
    virtual SetOfPoints& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a SetOfPoints
    virtual SetOfPoints& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                                 real_t angle = 0.);
    //! apply a rotation on a SetOfPoints
    virtual SetOfPoints& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a SetOfPoints
    virtual SetOfPoints& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a SetOfPoints
    virtual SetOfPoints& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a SetOfPoints
    virtual SetOfPoints& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a SetOfPoints
    virtual SetOfPoints& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a SetOfPoints
    virtual SetOfPoints& homothetize(real_t factor);
    //! apply a point reflection on a SetOfPoints
    virtual SetOfPoints& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a SetOfPoints
    virtual SetOfPoints& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a SetOfPoints
    virtual SetOfPoints& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a SetOfPoints
    virtual SetOfPoints& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a SetOfPoints
    virtual SetOfPoints& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
*/
};

} // end of namespace xlifepp

#endif // GEOMETRIES_1D_HPP
