/*
  XLiFE++ is an extended library of finite elements written in C++
  Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file geometries2D.hpp
  \authors Y. Lafranche, N. Kielbasiewicz
  \since 18 oct 2012
  \date 30 jul 2015

  \brief Definition of the classes corresponding to 2D canonical geometrical shapes.

  It is quite easy to find in which category a geometry is, according to their mathematical
  definition, but they are some exceptions :
  - Cylinder and Cone are surfaces but can also represent volumes inside
  - Ellipse is a closed curve or the surface inside (whereas there is the
    distinction circle / disk)
  - Ellipsoid is a closed surface or the volume inside (whereas there is the
    distinction Sphere / Ball)

  It becomes all the more tricky than if we respect the geometrical hierarchy, we will derive a
  prism (semantically 3D) from a cylinder (2D or 3D) or a pyramid (semantically 3D) from a cone.
  Another example is the general definition of a cylinder, from a surface and an axis : it supposes
  that the basis is a surface. But ellipses are suposed to be curves !!!

  This is the reason why we chose to defines ellipses as surfaces, ellipsoids as volumes (coherence) and cylinders
  and cones as volumes.

*/

#ifndef GEOMETRIES_2D_HPP
#define GEOMETRIES_2D_HPP

#include "Geometry.hpp"

namespace xlifepp
{

/*!
  \class Surface
  base class for 2D geometries
*/
class Surface : public Geometry
{
  public:
    //! default constructor for 2D geometries
    Surface();

  protected:
    //@{
    //! true constructor functions
    void buildParam(const Parameter& p);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    Surface(const Surface& s) : Geometry(s) {} //!< copy constructor
    virtual Geometry* clone() const { return new Surface(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Surface(*this); } //!< virtual copy constructor for Surface
    virtual ~Surface() {} //!< virtual destructor

    virtual std::vector<Point> p() const
    { error("not_yet_implemented","std::vector<Point> Surface::p() const"); return std::vector<Point>(); }
    virtual std::vector<number_t> n() const
    { error("not_yet_implemented","std::vector<Point> Surface::n() const"); return std::vector<number_t>(); }
    virtual Point p(number_t i) const
    { error("not_yet_implemented","Point Surface::p(Number i) const"); return Point(); }
    virtual number_t n(number_t i) const
    { error("not_yet_implemented","Number Surface::n(Number i) const"); return 0; }
    virtual number_t& n(number_t i)
    { error("not_yet_implemented","Number& Surface::n(Number i)"); return *new number_t(0); }
    virtual std::vector<real_t> h() const
    { error("not_yet_implemented","std::vector<Real> Surface::h() const"); return std::vector<real_t>(); }
    virtual real_t h(number_t i) const
    { error("not_yet_implemented","Number Surface::h(Number i) const"); return 0.; }
    virtual real_t& h(number_t i)
    { error("not_yet_implemented","Number& Surface::h(Number i)"); return *new real_t(0); }

    //! test if a surface is a polygon
    bool isPolygon()
    {
      if (shape_ == _polygon || shape_ == _triangle || shape_ == _quadrangle || shape_ == _parallelogram || shape_ == _rectangle ||
          shape_ == _square) { return true; }
      return false;
    }

    //! format as string
    virtual string_t asString() const { error("shape_not_handled", words("shape",shape_)); return string_t(); }

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a Surface
    virtual Surface& transform(const Transformation& t);
    //! apply a translation on a Surface (vector version)
    virtual Surface& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a Surface (3 reals version)
    virtual Surface& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a Surface
    virtual Surface& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a Surface
    virtual Surface& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.);
    //! apply a rotation on a Surface
    virtual Surface& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Surface
    virtual Surface& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a Surface
    virtual Surface& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Surface
    virtual Surface& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a Surface
    virtual Surface& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a Surface
    virtual Surface& homothetize(real_t factor);
    //! apply a point reflection on a Surface
    virtual Surface& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a Surface
    virtual Surface& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a Surface
    virtual Surface& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a Surface
    virtual Surface& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a Surface
    virtual Surface& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

/*!
  \class Polygon
  definition of a polygonal geometry in R^3
 */
class Polygon : public Surface
{
  protected:
    //! vertices of the polygon
    std::vector<Point> p_;
    //! local mesh step on each vertex of the polygon
    std::vector<real_t> h_;
    //! number of nodes on each edge
    std::vector<number_t> n_;

  public:
    //! default constructor
    Polygon();
    //! constructor with 1 Parameter
    Polygon(const Parameter& p1);
    //! constructor with 2 Parameter
    Polygon(const Parameter& p1, const Parameter& p2);
    //! constructor with 3 Parameter
    Polygon(const Parameter& p1, const Parameter& p2, const Parameter& p3);
    //! constructor with 4 Parameter
    Polygon(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4);

  protected:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~Polygon() {}

    //! accessor to vertices
    std::vector<Point> p() const { return p_; }
    //! accessor to vertices (writable)
    std::vector<Point>& p() { return p_; }
    //! accessor to vertex i
    Point p(number_t i) const { return p_[i-1]; }
    //! accessor to vertex i (writable)
    Point& p(number_t i) { return p_[i-1]; }
    //! accessor to number of nodes per edge
    std::vector<number_t> n() const {return n_;}
    //! accessor to number of nodes on edge i (read only)
    number_t n(number_t i) const { return n_[i-1];}
    //! accessor to number of nodes on edge i (read/write)
    number_t& n(number_t i) { return n_[i-1];}
    //! accessor to local steps
    std::vector<real_t> h() const {return h_;}
    //! accessor to local step on vertex i (read only)
    real_t h(number_t i) const { return h_[i-1];}
    //! accessor to local step on vertex i (read/write)
    real_t& h(number_t i) { return h_[i-1];}


    //! format as string
    virtual string_t asString() const;

    virtual Geometry* clone() const { return new Polygon(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Polygon(*this); } //!< virtual copy constructor for Surface
    virtual Polygon* clonePG() const { return new Polygon(*this); } //!< virtual copy constructor for Polygon

    bool withNnodes() const { return (h_.size() == 0); } //!< check if Polygon is defined only with _nnodes or with _hsteps option
    std::vector<int_t> nnodesPerBorder(); //!< i-th value is nnodes of the Geometry's i-th border, else -1 if defined with hsteps

    virtual std::vector<const Point*> boundNodes() const; //!< returns list of nodes on boundary (const)
    virtual std::vector<Point*> nodes(); //!< returns list of every node (non const)
    virtual std::vector<const Point*> nodes() const; //!< returns list of every node (const)
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > curves() const; //!< return list of edges (const)
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< return list of surfaces (const)
    virtual number_t nbSides() const { return p_.size(); } //!< returns the number of sides
    virtual void collect(const string_t& n, std::list<Geometry*>&) const; //!< collect in a list all canonical geometry's with name n

    //! computes the minimal box
    virtual void computeMB();

    //! access to child Polygon object (const)
    virtual const Polygon* polygon() const {return this;}
    //! access to child Polygon object
    virtual Polygon* polygon() {return this;}

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a Polygon
    virtual Polygon& transform(const Transformation& t);
    //! apply a translation on a Polygon (vector version)
    virtual Polygon& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a Polygon (3 reals version)
    virtual Polygon& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a Polygon
    virtual Polygon& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a Polygon
    virtual Polygon& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.);
    //! apply a rotation on a Polygon
    virtual Polygon& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Polygon
    virtual Polygon& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a Polygon
    virtual Polygon& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Polygon
    virtual Polygon& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a Polygon
    virtual Polygon& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a Polygon
    virtual Polygon& homothetize(real_t factor);
    //! apply a point reflection on a Polygon
    virtual Polygon& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a Polygon
    virtual Polygon& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a Polygon
    virtual Polygon& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a Polygon
    virtual Polygon& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a Polygon
    virtual Polygon& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

/*!
  \class Triangle
  definition of a triangular geometry in R^3
 */
class Triangle : public Polygon
{
  public:
    //! default constructor
    Triangle();
    //! default constructor with 3 Point
    Triangle(const Point& p1, const Point& p2, const Point& p3, const std::vector<number_t>& n=std::vector<number_t>(3,2),
             const string_t& domName=string_t());
    //! default constructor with 3 Point
    Triangle(const Point& p1, const Point& p2, const Point& p3, const std::vector<real_t>& h, const string_t& domName=string_t());
    //! constructor with 3 Parameter
    Triangle(const Parameter& p1, const Parameter& p2, const Parameter& p3);
    //! constructor with 4 Parameter
    Triangle(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4);
    //! constructor with 5 Parameter
    Triangle(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5);
    //! constructor with 6 Parameter
    Triangle(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
             const Parameter& p6);

  private:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~Triangle() {}

    virtual Geometry* clone() const { return new Triangle(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Triangle(*this); } //!< virtual copy constructor for Surface
    virtual Polygon* clonePG() const { return new Triangle(*this); } //!< virtual copy constructor for Polygon

    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< return list of surfaces (const)
    virtual number_t nbSides() const { return 3; } //!< returns the number of sides
    real_t measure() const;

    //! access to child Triangle object (const)
    virtual const Triangle* triangle() const {return this;}
    //! access to child Triangle object
    virtual Triangle* triangle() {return this;}

    //! format as string
    virtual string_t asString() const;

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a Triangle
    virtual Triangle& transform(const Transformation& t);
    //! apply a translation on a Triangle (vector version)
    virtual Triangle& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a Triangle (3 reals version)
    virtual Triangle& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a Triangle
    virtual Triangle& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a Triangle
    virtual Triangle& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                               real_t angle = 0.);
    //! apply a rotation on a Triangle
    virtual Triangle& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Triangle
    virtual Triangle& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a Triangle
    virtual Triangle& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Triangle
    virtual Triangle& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a Triangle
    virtual Triangle& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a Triangle
    virtual Triangle& homothetize(real_t factor);
    //! apply a point reflection on a Triangle
    virtual Triangle& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a Triangle
    virtual Triangle& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a Triangle
    virtual Triangle& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a Triangle
    virtual Triangle& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a Triangle
    virtual Triangle& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

/*!
  \class Quadrangle
  definition of a quadrangular geometry in R^3
 */
class Quadrangle : public Polygon
{
  public:
    //! default constructor
    Quadrangle();
    //! default constructor with 4 points
    Quadrangle(const Point& p1, const Point& p2, const Point& p3, const Point& p4,
               const std::vector<number_t>& n=std::vector<number_t>(4,2), const string_t& domName=string_t());
    //! default constructor with 4 points
    Quadrangle(const Point& p1, const Point& p2, const Point& p3, const Point& p4,
               const std::vector<real_t>& h, const string_t& domName=string_t());
    //! constructor with 4 Parameter
    Quadrangle(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4);
    //! constructor with 5 Parameter
    Quadrangle(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5);
    //! constructor with 6 Parameter
    Quadrangle(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
             const Parameter& p6);
    //! constructor with 7 Parameter
    Quadrangle(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
             const Parameter& p6, const Parameter& p7);

  protected:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~Quadrangle() {}

    virtual Geometry* clone() const { return new Quadrangle(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Quadrangle(*this); } //!< virtual copy constructor for Surface
    virtual Polygon* clonePG() const { return new Quadrangle(*this); } //!< virtual copy constructor for Polygon

    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< returns list of surfaces (const)
    virtual number_t nbSides() const { return 4; } //!< returns the number of sides
    real_t measure() const;

    //! access to child Quadrangle object (const)
    virtual const Quadrangle* quadrangle() const {return this;}
    //! access to child Quadrangle object
    virtual Quadrangle* quadrangle() {return this;}

    //! format as string
    virtual string_t asString() const;

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a Quadrangle
    virtual Quadrangle& transform(const Transformation& t);
    //! apply a translation on a Quadrangle (vector version)
    virtual Quadrangle& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a Quadrangle (3 reals version)
    virtual Quadrangle& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a Quadrangle
    virtual Quadrangle& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a Quadrangle
    virtual Quadrangle& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                                 real_t angle = 0.);
    //! apply a rotation on a Quadrangle
    virtual Quadrangle& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Quadrangle
    virtual Quadrangle& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a Quadrangle
    virtual Quadrangle& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Quadrangle
    virtual Quadrangle& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a Quadrangle
    virtual Quadrangle& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a Quadrangle
    virtual Quadrangle& homothetize(real_t factor);
    //! apply a point reflection on a Quadrangle
    virtual Quadrangle& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a Quadrangle
    virtual Quadrangle& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a Quadrangle
    virtual Quadrangle& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a Quadrangle
    virtual Quadrangle& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a Quadrangle
    virtual Quadrangle& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

/*!
  \class Parallelogram
  definition of a parallelogram geometry in R^3
 */
class Parallelogram : public Quadrangle
{
  public:
    //! default constructor
    Parallelogram();
    //! constructor with 3 Parameter
    Parallelogram(const Parameter& p1, const Parameter& p2, const Parameter& p3);
    //! constructor with 4 Parameter
    Parallelogram(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4);
    //! constructor with 5 Parameter
    Parallelogram(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5);
    //! constructor with 6 Parameter
    Parallelogram(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
             const Parameter& p6);

  protected:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~Parallelogram() {}

    //! returns the number of nodes on first or third edge
    number_t n1() const {return n_[0];}
    //! returns the number of nodes on second or fourth edge
    number_t n2() const {return n_[1];}

    //! format as string
    virtual string_t asString() const;

    virtual Geometry* clone() const { return new Parallelogram(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Parallelogram(*this); } //!< virtual copy constructor for Surface
    virtual Polygon* clonePG() const { return new Parallelogram(*this); } //!< virtual copy constructor for Polygon

    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< returns list of surfaces (const)
    real_t length1() const; //!< length of first or third side
    real_t length2() const; //!< length of second or fourth side
    real_t measure() const; //!< surface of the parallelogram

    //! access to child Parallelogram object (const)
    virtual const Parallelogram* parallelogram() const {return this;}
    //! access to child Parallelogram object
    virtual Parallelogram* parallelogram() {return this;}

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a Parallelogram
    virtual Parallelogram& transform(const Transformation& t);
    //! apply a translation on a Parallelogram (vector version)
    virtual Parallelogram& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a Parallelogram (3 reals version)
    virtual Parallelogram& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a Parallelogram
    virtual Parallelogram& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a Parallelogram
    virtual Parallelogram& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                                    real_t angle = 0.);
    //! apply a rotation on a Parallelogram
    virtual Parallelogram& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Parallelogram
    virtual Parallelogram& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a Parallelogram
    virtual Parallelogram& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Parallelogram
    virtual Parallelogram& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a Parallelogram
    virtual Parallelogram& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a Parallelogram
    virtual Parallelogram& homothetize(real_t factor);
    //! apply a point reflection on a Parallelogram
    virtual Parallelogram& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a Parallelogram
    virtual Parallelogram& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a Parallelogram
    virtual Parallelogram& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a Parallelogram
    virtual Parallelogram& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a Parallelogram
    virtual Parallelogram& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

/*!
  \class Rectangle
  definition of a rectangular geometry in R^3
 */
class Rectangle : public Parallelogram
{
  protected:
    Point center_;   //!< center of the rectangle
    Point origin_;   //!< "bottom left" vertex of the rectangle (is equal to p_[0])
    bool isCenter_;  //!< true when center key is used
    bool isOrigin_;  //!< true when origin key is used
    real_t xlength_; //!< first length of the rectangle
    real_t ylength_; //!< second length of the rectangle
  private:
    // initialized when rectangle defined with xmin, xmax, ymin, ymax keys
    real_t xmin_, xmax_, ymin_, ymax_;
    // true when key group (_xmin, _xmax, _ymin, _ymax) is given
    bool isBounds_;

  public:
    //! default constructor
    Rectangle();
    //! constructor with 3 Parameter
    Rectangle(const Parameter& p1, const Parameter& p2, const Parameter& p3);
    //! constructor with 4 Parameter
    Rectangle(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4);
    //! constructor with 5 Parameter
    Rectangle(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5);
    //! constructor with 6 Parameter
    Rectangle(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
              const Parameter& p6);
    //! constructor with 7 Parameter
    Rectangle(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
              const Parameter& p6, const Parameter& p7);

  protected:
    //@{
    //! true constructor functions
    void buildP();
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~Rectangle() {}

    //! returns the number of nodes on first or third edge
    number_t nx() const {return n_[0];}
    //! returns the number of nodes on second or fourth edge
    number_t ny() const {return n_[1];}

    //! format as string
    virtual string_t asString() const;

    virtual Geometry* clone() const { return new Rectangle(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Rectangle(*this); } //!< virtual copy constructor for Surface
    virtual Polygon* clonePG() const { return new Rectangle(*this); } //!< virtual copy constructor for Polygon

    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< returns list of surfaces (const)
    real_t xlength() const { return xlength_; }
    real_t ylength() const { return ylength_; }
    real_t measure() const { return (xlength_ * ylength_); }

    virtual void computeMB() { minimalBox= MinimalBox(p_[0],p_[1],p_[3]); } //!< computes the minimal box

    //! access to child Rectangle object (const)
    virtual const Rectangle* rectangle() const {return this;}
    //! access to child Rectangle object
    virtual Rectangle* rectangle() {return this;}

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a Rectangle
    virtual Rectangle& transform(const Transformation& t);
    //! apply a translation on a Rectangle (vector version)
    virtual Rectangle& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a Rectangle (3 reals version)
    virtual Rectangle& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a Rectangle
    virtual Rectangle& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a Rectangle
    virtual Rectangle& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                                real_t angle = 0.);
    //! apply a rotation on a Rectangle
    virtual Rectangle& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Rectangle
    virtual Rectangle& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a Rectangle
    virtual Rectangle& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Rectangle
    virtual Rectangle& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a Rectangle
    virtual Rectangle& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a Rectangle
    virtual Rectangle& homothetize(real_t factor);
    //! apply a point reflection on a Rectangle
    virtual Rectangle& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a Rectangle
    virtual Rectangle& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a Rectangle
    virtual Rectangle& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a Rectangle
    virtual Rectangle& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a Rectangle
    virtual Rectangle& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

/*!
  \class Square
  definition of a square geometry in R^3
 */
class Square : public Rectangle
{
  public:
    //! default constructor
    Square();
    //! constructor with 2 Parameter
    Square(const Parameter& p1, const Parameter& p2);
    //! constructor with 3 Parameter
    Square(const Parameter& p1, const Parameter& p2, const Parameter& p3);
    //! constructor with 4 Parameter
    Square(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4);
    //! constructor with 5 Parameter
    Square(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5);
    //! constructor with 6 Parameter
    Square(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
           const Parameter& p6);

  private:
    //@{
    //! true constructor functions
    void buildP();
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~Square() {}

    //! format as string
    virtual string_t asString() const;

    virtual Geometry* clone() const { return new Square(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Square(*this); } //!< virtual copy constructor for Surface
    virtual Polygon* clonePG() const { return new Square(*this); } //!< virtual copy constructor for Polygon

    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< returns list of surfaces (const)

    //! access to child Square object (const)
    virtual const Square* square() const {return this;}
    //! access to child Square object
    virtual Square* square() {return this;}

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a Square
    virtual Square& transform(const Transformation& t);
    //! apply a translation on a Square (vector version)
    virtual Square& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a Square (3 reals version)
    virtual Square& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a Square
    virtual Square& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a Square
    virtual Square& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                             real_t angle = 0.);
    //! apply a rotation on a Square
    virtual Square& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Square
    virtual Square& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a Square
    virtual Square& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Square
    virtual Square& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a Square
    virtual Square& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a Square
    virtual Square& homothetize(real_t factor);
    //! apply a point reflection on a Square
    virtual Square& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a Square
    virtual Square& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a Square
    virtual Square& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a Square
    virtual Square& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a Square
    virtual Square& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

/*!
  \class Ellipse
  definition of an elliptic geometry in R^3 (surface)
 */
class Ellipse : public Surface
{
  protected:
    Point center_;          //!< center of the ellipse
    Point p1_;              //!< first apogee of the ellipse
    Point p2_;              //!< second apogee of the ellipse
    Point p3_;              //!< first perigee of the ellipse
    Point p4_;              //!< second perigee of the ellipse
    real_t xlength_, ylength_;
    bool isAxis_;
    number_t n1_;           //!< number of nodes on edge [p1_ p2_]
    number_t n2_;           //!< number of nodes on edge [p2_ p3_]
    number_t n3_;           //!< number of nodes on edge [p3_ p4_]
    number_t n4_;           //!< number of nodes on edge [p4_ p1_]
    std::vector<real_t> h_; //!< local mesh step on each vertex of the ellipse

    real_t thetamin_;       //!< first angle defining the sector, given in degrees
    real_t thetamax_;       //!< second angle defining the sector, given in degrees
    dimen_t type_;         //!< indicator to fit curved boundaries (default) or not which gives flat (or plane) boundaries

  public:
    //! default constructor
    Ellipse();
    //! default constructor with 3 points
    Ellipse(const Point& center, const Point& p1, const Point& p2, const std::vector<number_t>& n = std::vector<number_t>(4,2),
            const string_t& domName=string_t());
    //! default constructor with 3 points
    Ellipse(const Point& center, const Point& p1, const Point& p2, const std::vector<real_t>& h, const string_t& domName=string_t());
    //! constructor with 3 Parameter
    Ellipse(const Parameter& p1, const Parameter& p2, const Parameter& p3);
    //! constructor with 4 Parameter
    Ellipse(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4);
    //! constructor with 5 Parameter
    Ellipse(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5);
    //! constructor with 6 Parameter
    Ellipse(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
            const Parameter& p6);
    //! constructor with 7 Parameter
    Ellipse(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
            const Parameter& p6, const Parameter& p7);
    //! constructor with 8 Parameter
    Ellipse(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
            const Parameter& p6, const Parameter& p7, const Parameter& p8);
    //! constructor with 9 Parameter
    Ellipse(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
            const Parameter& p6, const Parameter& p7, const Parameter& p8, const Parameter& p9);

  protected:
    //@{
    //! true constructor functions
    void buildP();
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~Ellipse() {}

    //@{
    //! accessor to point
    Point center() const { return center_; }
    Point p1() const { return p1_; }
    Point p2() const { return p2_; }
    Point p3() const { return p3_; }
    Point p4() const { return p4_; }
    //@}

    //@{
    //! accessor to number of nodes
    number_t n1() const {return n1_;}
    number_t n2() const {return n2_;}
    number_t n3() const {return n3_;}
    number_t n4() const {return n4_;}
    //@}

    real_t length1() const { return xlength_; }    //!< return first axis length
    real_t length2() const { return ylength_; }    //!< return second axis length
    virtual number_t nbSubdiv() const;       //!< returns number of subdivision
    virtual dimen_t type() const { return type_; } //!< returns type of subdivision
    real_t thetamin() const { return thetamin_; } //!< returns the first angle
    real_t thetamax() const { return thetamax_; } //!< returns the second angle

    //! accessor to points
    std::vector<Point> p() const;
    //! accessor to point i (read only)
    Point p(number_t i) const;
    //! accessor to number of nodes per edge
    std::vector<number_t> n() const;
    //! accessor to number of nodes on edge i (read only)
    number_t n(number_t i) const;
    //! accessor to number of nodes on edge i (read/write)
    number_t& n(number_t i);
    //! accessor to local steps
    std::vector<real_t> h() const {return h_;}
    //! accessor to local step on vertex i (read only)
    real_t h(number_t i) const { return h_[i-1];}
    //! accessor to local step on vertex i (read/write)
    real_t& h(number_t i) { return h_[i-1];}

    //! format as string
    virtual string_t asString() const;

    virtual Geometry* clone() const { return new Ellipse(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Ellipse(*this); } //!< virtual copy constructor for Surface

    bool withNnodes() const { return (h_.size() == 0); } //!< check if Ellipse is defined only with _nnodes or with _hsteps option
    std::vector<int_t> nnodesPerBorder(); //!< i-th value is nnodes of the Geometry's i-th border, else -1 if defined with hsteps

    virtual std::vector<const Point*> boundNodes() const; //!< returns list of points on boundary (const)
    virtual std::vector<Point*> nodes(); //!< returns list of every point (non const)
    virtual std::vector<const Point*> nodes() const; //!< returns list of every point (const)
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > curves() const; //!< returns list of edges (const)
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< returns list of surfaces (const)
    virtual number_t nbSides() const { return 4; } //!< returns the number of sides
    real_t measure() const { return 0.25*pi_*xlength_*ylength_; }

    virtual void collect(const string_t& n, std::list<Geometry*>&) const; //!< collect in a list all canonical geometry's with name n

    //! computes the minimal box
    virtual void computeMB() { minimalBox= MinimalBox(3.*center_-p1_-p2_,center_+p1_-p2_,center_+p2_-p1_); }

    //! access to child Ellipse object (const)
    virtual const Ellipse* ellipse() const {return this;}
    //! access to child Ellipse object
    virtual Ellipse* ellipse() {return this;}

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a Ellipse
    virtual Ellipse& transform(const Transformation& t);
    //! apply a translation on a Ellipse (vector version)
    virtual Ellipse& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a Ellipse (3 reals version)
    virtual Ellipse& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a Ellipse
    virtual Ellipse& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a Ellipse
    virtual Ellipse& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.);
    //! apply a rotation on a Ellipse
    virtual Ellipse& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Ellipse
    virtual Ellipse& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a Ellipse
    virtual Ellipse& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Ellipse
    virtual Ellipse& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a Ellipse
    virtual Ellipse& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a Ellipse
    virtual Ellipse& homothetize(real_t factor);
    //! apply a point reflection on a Ellipse
    virtual Ellipse& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a Ellipse
    virtual Ellipse& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a Ellipse
    virtual Ellipse& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a Ellipse
    virtual Ellipse& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a Ellipse
    virtual Ellipse& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

/*!
  \class Disk
  definition of a circular geometry in R^3 (surface)
 */
class Disk : public Ellipse
{
  public:
    //! default constructor
    Disk();
    //! default constructor with 3 points
    Disk(const Point& center, const Point& p1, const Point& p2, const std::vector<number_t>& n=std::vector<number_t>(4,2),
         const string_t& domName=string_t());
    //! default constructor with 3 points
    Disk(const Point& center, const Point& p1, const Point& p2, const std::vector<real_t>& h, const string_t& domName=string_t());
    //! constructor with 2 Parameter
    Disk(const Parameter& p1, const Parameter& p2);
    //! constructor with 3 Parameter
    Disk(const Parameter& p1, const Parameter& p2, const Parameter& p3);
    //! constructor with 4 Parameter
    Disk(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4);
    //! constructor with 5 Parameter
    Disk(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5);
    //! constructor with 6 Parameter
    Disk(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
            const Parameter& p6);
    //! constructor with 7 Parameter
    Disk(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
            const Parameter& p6, const Parameter& p7);
    //! constructor with 8 Parameter
    Disk(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
            const Parameter& p6, const Parameter& p7, const Parameter& p8);
    //! constructor with 9 Parameter
    Disk(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4, const Parameter& p5,
            const Parameter& p6, const Parameter& p7, const Parameter& p8, const Parameter& p9);

  protected:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~Disk() {}

    //! format as string
    virtual string_t asString() const;

    virtual Geometry* clone() const { return new Disk(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Disk(*this); } //!< virtual copy constructor for Surface

    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > curves() const; //!< returns list of edges

    //! access to child Disk object (const)
    virtual const Disk* disk() const {return this;}
    //! access to child Disk object
    virtual Disk* disk() {return this;}

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a Disk
    virtual Disk& transform(const Transformation& t);
    //! apply a translation on a Disk (vector version)
    virtual Disk& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a Disk (3 reals version)
    virtual Disk& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a Disk
    virtual Disk& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a Disk
    virtual Disk& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                           real_t angle = 0.);
    //! apply a rotation on a Disk
    virtual Disk& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Disk
    virtual Disk& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a Disk
    virtual Disk& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a Disk
    virtual Disk& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a Disk
    virtual Disk& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a Disk
    virtual Disk& homothetize(real_t factor);
    //! apply a point reflection on a Disk
    virtual Disk& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a Disk
    virtual Disk& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a Disk
    virtual Disk& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a Disk
    virtual Disk& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a Disk
    virtual Disk& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);

    real_t radius() const { return xlength_/2.; }    //!< return radius
};

typedef Disk Circle;

/*!
   \class SetOfElems
   This class defines an initial mesh from a list of points (pts), a list of elements (elems)
   and a list of boundaries (bounds).
     Each element is defined by N numbers corresponding to the rank of its N vertices
   in the vector pts ; the numbers start from 1. Thus, elems[k][i] is the number of the
   vertex (i+1) of the element number (k+1) (if the elements are numbered starting
   from 1), and 1 <= elems[k][i] <= pts.size().
     The boundaries of the polygonal domain defined by the elements are also defined
   by point numbers with the same convention. Each boundary is defined by the list of
   numbers of the points they belong to, given in any order.
     The last argument nbsubdiv is the number of subdivisions wanted. The initial
   mesh corresponds to nbsubdiv=0.
   -> Important: the subdivision algorithm generate a new point on each edge of the mesh.
      It belongs to boundary B if the two end points of the edge belong to the boundary B ;
      it is an internal point otherwise. Owing to this rule, in order to get a correct result,
      care must be taken in the definition of the boundaries since too few boundaries may
      lead to get boundary points that are in fact inside the domain. For instance,
      subdividing one single triangle leads to mandatory define its three edges as boundaries.
*/
class SetOfElems : public Geometry
{
  private:
    std::vector<Point> pts_; //!< list of points
    const std::vector<std::vector<number_t> >& elems_; //!< list of elements (an element is defined by its list of nodes)
    const std::vector<std::vector<number_t> >& bounds_; //!< list of elements on boundary
    ShapeType elemShape_;             //!< shape of the elements in the mesh
    number_t nbsubdiv_;           //!< number of subdivisions to be performed to get the final mesh
    const number_t nbElems_, nbBounds_;

  public:
    //! default constructor
    SetOfElems(const std::vector<Point>& pts, const std::vector<std::vector<number_t> >& elems,
               const std::vector<std::vector<number_t> >& bounds, const std::vector<string_t>& names,
               const ShapeType esh, const number_t nbsubdiv=1)
    : Geometry(BoundingBox(pts), BoundingBox(pts).dim(), string_t("Omega"), _setofelems),
      pts_(pts), elems_(elems), bounds_(bounds), elemShape_(esh), nbsubdiv_(nbsubdiv),
      nbElems_(elems_.size()), nbBounds_(bounds_.size())
    { sideNames_ = names; }

    //! format as string
    virtual string_t asString() const;

    virtual Geometry* clone() const { return new SetOfElems(*this); } //!< virtual copy constructor

    virtual std::vector<Point*> nodes(); //!< list of every point (non const)
    virtual std::vector<const Point*> nodes() const; //!< list of every point (const)

    // access functions to data members
    std::vector<Point>& pts() { return pts_; } //!< return list of points (non const)
    const std::vector<Point>& pts() const { return pts_; } //!< returns list of points (const)
    const std::vector<std::vector<number_t> >& elems() const { return elems_; } //!< returns list of elements
    const std::vector<std::vector<number_t> >& bounds() const { return bounds_; } //!< returns list of elements on boundary
    ShapeType elemShape() const { return elemShape_; } //!< returns shape of the elements
    virtual number_t nbSubdiv() const { return nbsubdiv_; } //!< returns number of subdivision
    virtual number_t nbSides() const { return bounds_.size(); } //!< returns the number of sides

    //! access to child SetOfElems object (const)
    virtual const SetOfElems* setofelems() const {return this;}
    //! access to child SetOfElems object
    virtual SetOfElems* setofelems() {return this;}

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a SetOfElems
    virtual SetOfElems& transform(const Transformation& t);
    //! apply a translation on a SetOfElems (vector version)
    virtual SetOfElems& translate(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! apply a translation on a SetOfElems (3 reals version)
    virtual SetOfElems& translate(real_t ux, real_t uy = 0., real_t uz = 0.);
    //! apply a rotation on a SetOfElems
    virtual SetOfElems& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.);
    //! apply a rotation on a SetOfElems
    virtual SetOfElems& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                                 real_t angle = 0.);
    //! apply a rotation on a SetOfElems
    virtual SetOfElems& rotate3d(real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a SetOfElems
    virtual SetOfElems& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a rotation on a SetOfElems
    virtual SetOfElems& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! apply a rotation on a SetOfElems
    virtual SetOfElems& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! apply a homothety on a SetOfElems
    virtual SetOfElems& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! apply a homothety on a SetOfElems
    virtual SetOfElems& homothetize(real_t factor);
    //! apply a point reflection on a SetOfElems
    virtual SetOfElems& pointReflect(const Point& c = Point(0.,0.,0.));
    //! apply a reflection2d on a SetOfElems
    virtual SetOfElems& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! apply a reflection2d on a SetOfElems
    virtual SetOfElems& reflect2d(const Point& c, real_t dx, real_t dy = 0.);
    //! apply a reflection3d on a SetOfElems
    virtual SetOfElems& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! apply a reflection3d on a SetOfElems
    virtual SetOfElems& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.);
};

} // end of namespace xlifepp

#endif // GEOMETRIES_2D_HPP
