/*
  XLiFE++ is an extended library of finite elements written in C++
  Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file geometries_utils.hpp
  \authors Y. Lafranche, N. Kielbasiewicz
  \since 18 oct 2012
  \date 30 jul 2015

  \brief Additional functions related to geometries
*/

#ifndef GEOMETRIES_UTILS_HPP
#define GEOMETRIES_UTILS_HPP

#include "Geometry.hpp"

namespace xlifepp
{

number_t nodesDim(const Geometry& g);

string_t oneOfSideNames(const string_t& sn, int i=0);
string_t oneOfSideNames(const std::vector<string_t>& sn, int i=0);
string_t oneOfSideNamesIfVector(const string_t& sn, int i=0);
string_t oneOfSideNamesIfVector(const std::vector<string_t>& sn, int i=0);
number_t nbSubDomainsIfScalar(string_t sn, number_t nbSubDomains);
number_t nbSubDomainsIfScalar(const std::vector<string_t>& sn, number_t nbSubDomains);

//---------------------------------------------------------------------------
// transformations facilities (template external)
//---------------------------------------------------------------------------

//! apply a geometrical transformation on a Geom (template external)
template<class Geom> Geom transform(const Geom& g, const Transformation& t);
//! apply a translation on a Geom (vector version) (template external)
template<class Geom> Geom translate(const Geom& g, std::vector<real_t> u = std::vector<real_t>(3,0.));
//! apply a translation on a Geom (3 reals version) (template external)
template<class Geom> Geom translate(const Geom& g, real_t ux, real_t uy, real_t uz = 0.);
//! apply a rotation on a Geom (template external)
template<class Geom> Geom rotate2d(const Geom& g, const Point& c = Point(0.,0.), real_t angle = 0.);
//! apply a rotation on a Geom (template external)
template<class Geom> Geom rotate3d(const Geom& g, const Point& c = Point(0.,0.,0.),
                                    std::vector<real_t> d = std::vector<real_t>(3,0.), real_t angle = 0.);
//! apply a rotation on a Geom (template external)
template<class Geom> Geom rotate3d(const Geom& g, real_t dx, real_t dy, real_t angle);
//! apply a rotation on a Geom (template external)
template<class Geom> Geom rotate3d(const Geom& g, real_t dx, real_t dy, real_t dz, real_t angle);
//! apply a rotation on a Geom (template external)
template<class Geom> Geom rotate3d(const Geom& g, const Point& c, real_t dx, real_t dy, real_t angle);
//! apply a rotation on a Geom (template external)
template<class Geom> Geom rotate3d(const Geom& g, const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
//! apply a homothety on a Geom (template external)
template<class Geom> Geom homothetize(const Geom& g, const Point& c = Point(0.,0.,0.), real_t factor = 0.);
//! apply a homothety on a Geom (template external)
template<class Geom> Geom homothetize(const Geom& g, real_t factor);
//! apply a point reflection on a Geom (template external)
template<class Geom> Geom pointReflect(const Geom& g, const Point& c = Point(0.,0.,0.));
//! apply a reflection2d on a Geom (template external)
template<class Geom> Geom reflect2d(const Geom& g, const Point& c = Point(0.,0.),
                                     std::vector<real_t> d = std::vector<real_t>(2,0.));
//! apply a reflection2d on a Geom (template external)
template<class Geom> Geom reflect2d(const Geom& g, const Point& c, real_t dx, real_t dy = 0.);
//! apply a reflection3d on a Geom (template external)
template<class Geom> Geom reflect3d(const Geom& g, const Point& c = Point(0.,0.,0.),
                                     std::vector<real_t> n = std::vector<real_t>(3,0.));
//! apply a reflection3d on a Geom (template external)
template<class Geom> Geom reflect3d(const Geom& g, const Point& c, real_t nx, real_t ny, real_t nz = 0.);

} // end of namespace xlifepp

#endif // GEOMETRIES_UTILS_HPP
