/*
XLiFE++ is an extended library of finite elements written in C++
    Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
/*!
  \file init.hpp
  \authors D. Martin, N. Kielbasiewicz
  \since 08 jan 2005
  \date 4 may 2012

  \brief Definition of init function

  Initialisation procedure for error/message handling :

  It creates the XLiFE++ Messages classe loading ascii files of messages,
  see Messages.hpp for a description of these files.
  Those classes have global scope.

  To instantiate an error message, see function msg in Messages.cpp

  To add a new error class :
    - make a new error file (one for each language) : error_file
    - create a new Messages class :
      \code
      my_errors=new Messages(error_file,out_stream,out_file,error_type);
      \endcode
    - or append the new error file in the unique Messages object
      \code
      theMessages_p->appendFromFile(error_file);
      \endcode

  To throw some me_errors call :
    \code
    msg(id_message,data,my_errors)
    msg(id_message,data)
    \endcode

  messages file names are [en | fr]/xxxxx.txt
  where xxxxx is the error family, same as internal Messages object
  for example : for internal error types,
  object name is internal and messages file is [en | fr]/internal.txt

  format of format message in file is of the form :
    \verbatim
    # int_id string_id type_error type_status format_string
    \endverbatim
    where format_string any sequence of strings (excluding character '#') and any combination of %i, %r, %c, %b, %s where :
      - %i stands for an integer parameter,
      - %r stands for a real parameter,
      - %c stands for a complex parameter,
      - %b stands for a boolean parameter,
      - %s stands for a string parameter
  Eol can be used in format_string and is preserved on output
 */
#ifndef INIT_HPP
#define INIT_HPP

#include "config.hpp"

namespace xlifepp
{

//================================================================================
// init functions
//================================================================================

// void init(Language lang = _en); //!< initializes execution of XLiFE++

void init(); //!< initializes execution of XLiFE++
void init(const Parameter& p1); //!< initializes execution of XLiFE++
void init(const Parameter& p1, const Parameter& p2); //!< initializes execution of XLiFE++
void init(const Parameter& p1, const Parameter& p2, const Parameter& p3); //!< initializes execution of XLiFE++
void init(const Parameter& p1, const Parameter& p2, const Parameter& p3, const Parameter& p4); //!< initializes execution of XLiFE++
void init(const std::vector<Parameter>& ps); //!< initializes execution of XLiFE++
void initBuild(Language lang, number_t verboseLevel, bool trackingMode, bool isLogged); //!< initializes execution of XLiFE++

void timerInit(); //!< initialization of engine for time handling
void traceInit(); //!< initialization of engine for trace handling
void msgInit(const string_t&, std::ofstream&); //!< initialization of engine for messages handling

} // end of namespace xlifepp

#endif /* INIT_HPP */
