/*
XLiFE++ is an extended library of finite elements written in C++
    Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file BicgSolver.hpp
  \authors C.Chambeyron, D.Martin, Manh Ha NGUYEN
  \since 24 Nov 2004
  \date  01 Oct 2012

  \brief Definition of the xlifepp::BicgSolver class

  Class xlifepp::BicgSolver implements the BiConjugate Gradient method for solving
  (non-symmetric) linear systems, with or without preconditioner.

  BiCG delivers about the same results as CG for symmetric linear systems
  at twice the cost (use CG instead)
  BiCG is not always suitable for non symetric linear systems (possible //breakdown)
  (use BiCGStab instead)
  In short one should never use BiCG !!! But IT WORKS ...

  Base class : xlifepp::IterativeSolver

  For elementary explanation about BiConjugate Gradient method,
  see e.g. http://mathworld.wolfram.com/BiconjugateGradientMethod.html
*/

#ifndef BICG_SOLVER_HPP
#define BICG_SOLVER_HPP

#include "IterativeSolver.hpp"
#include "Preconditioner.hpp"

namespace xlifepp
{

/*!
     \class BicgSolver
     Implement BiConjugate Gradient algorithm
*/
class BicgSolver : public IterativeSolver
{
public:
  //! Default Constructor
  BicgSolver()
    : IterativeSolver("BiCG") {}

  //! Full constructor
  BicgSolver(const real_t eps,
             const number_t maxOfIt = defaultMaxIterations,
             const number_t vb = 0)
    : IterativeSolver("BiCG", maxOfIt, eps, vb) {}

  ~BicgSolver() {}

  //! Template Solvers without a precondition without argument "solType
  template<class Mat, class VecB, class VecX>
  VecX operator()(Mat& matA, VecB& vecB, VecX& vecX0)
  {
    resetSolver();
    if ( matA.valueType() == _real && vecB.valueType() == _real )
      return (*this)( matA,  vecB,  vecX0,  _real );
    else return (*this)( matA,  vecB,  vecX0,  _complex );
  }

  //! Template Solvers without a precondition
  template<class Mat, class VecB, class VecX>
  VecX operator()(Mat& matA, VecB& vecB, VecX& vecX0, ValueType solType)
  {
    trace_p->push(name_ + "Solver");
    preconditioned_ = false;
    // Default value of maximum of iterations when not defined by user
    maxOfIterations_ = maximumOfIterations(vecX0.size());
    //initialize guess and residue
    printHeader(vecX0.size());
    VecX x(vecX0);
    VecX r;
    multMatrixVector(matA, x, r);
    r*=-1.;
    r+=vecB;
    //run algorithm
    if (_real == solType)  algorithm<real_t>(matA, vecB, x, r);
    else                   algorithm<complex_t>(matA, vecB, x, r);
    //output
    printOutput();
    printResult();
    trace_p->pop();
    return x;
  }

  //! Template Solvers without a preconditionner and vecX0 = vecB
  template<class Mat, class VecX>
  VecX operator()(Mat& matA, VecX& vecB, ValueType solType)
  {
    return (*this)(matA, vecB, vecB, solType);
  }

  //! Template Solvers with a precondition without argument "solType
  template<class Mat, class VecB, class VecX, class VecPC>
  VecX operator()(Mat& matA, VecB& vecB, VecX& vecX0, VecPC& pc)
  {
    resetSolver();
    if ( (_real == matA.valueType()) &&  (_real == vecB.valueType() ) )
    {
      return (*this)( matA,  vecB,  vecX0, pc, _real );
    }
    else
    {
      return (*this)( matA,  vecB,  vecX0, pc, _complex );
    }
  }

  //! Template Solvers with a preconditioner
  template<class Mat, class VecB, class VecX, class Prec>
  VecX operator()(Mat& matA, VecB& vecB, VecX& vecX0, Prec& pc, ValueType solType)
  {
    trace_p->push(name_ + "Solver");

    // Default value of maximum of iterations when not defined by user
    maxOfIterations_ = maximumOfIterations(vecX0.size());
    //print header
    if ( preconditioned_ )
    {
      printHeader(vecX0.size(), pc.name());
      if ( pc.valueType() == _complex ) { solType = _complex; }
    }
    else { printHeader(vecX0.size()); }

    // Define initial residual vector according to Type
    VecX x(vecX0);
    VecX r;
    multMatrixVector(matA, x, r);
    r*=-1.;
    r+=vecB;
    //run algorithm
    if ( _real == solType ) algorithm<real_t>(matA, vecB, x, r, pc);
    else                    algorithm<complex_t>(matA, vecB, x, r, pc);
    //output
    printOutput();
    printResult();
    trace_p->pop();
    return x;
  }

private:
  /*!
      Solve matA.vecX=vecB using BiCG method with initial guess vecX without preconditioning.
         Typename K (K : real or complex) is the type of the initial residual vector
         and thus of auxiliary vectors as matrices matA, vecB
         and vector vecX0 might not share the same type
  */
  template<typename K, class Mat, class VecB, class VecX>
  void algorithm(Mat& matA, VecB& vecB, VecX& vecX, VecX& vecR)
  {
    trace_p->push(name_ + "Algorithm");
    VecX X(vecX);
    VecX Q, Qtilda, P, Ptilda;
    VecX R0 (vecR);    //R0 = B - A*X0;
    VecX R0tilda(R0);
    VecX Z, Ztilda, tmp;
    K beta, alpha, rhoold=0, rho, gamma ;
    number_t iter=0;
    assign(gamma,dotRC(vecR, R0));   //gamma = dotRC(vecR, vecR);
    K gammatilda;
    assign(gammatilda,dotRC(vecR, vecR));  //gammatilda=dotRC(R0tilda,R0tilda);

    while( std::sqrt(std::abs(gamma))>std::abs(epsilon_) &&  std::sqrt(std::abs(gammatilda))>std::abs(epsilon_) && (iter < maxOfIterations_))
    {
      Z=R0; Ztilda=R0tilda;
      assign(rho,dotRC(R0,R0tilda));   //  rho=dotRC(Z,R0tilda);
      if ( std::abs(rho) < theBreakdownThreshold ) breakdown(std::abs(rho), "rho");

      if (iter==0)
      {
        P=Z;
        Ptilda=Ztilda;
      }
      else
      {
        beta=rho/rhoold;
        P*=beta;
        P+=Z;
        Ptilda*=beta;
        Ptilda+=Ztilda;
      }
      multMatrixVector(matA, P, Q);              //Q=A*P;
      multVectorMatrix(Ptilda, matA, Qtilda);    //Qtilda=Ptilda*A;
      assign(alpha, 1. /dotRC(Ptilda, Q) );      //alpha=rho/Ptilda.dot(Q);
      alpha *= rho;
      tmp = P;
      tmp*=alpha;
      X+=tmp;
      tmp = Q; tmp*=-alpha;
      R0+=tmp;                                  //R0-=alpha*Q;
      tmp = Qtilda; tmp*=-alpha;
      R0tilda+=tmp;                             //R0tilda-=alpha*Qtilda;
      assign(gamma,  dotRC(R0, R0));           //gamma = dotRC(R0, R0) ;
      rhoold=rho;
      iter++;
    }
    residue_ = std::sqrt(std::abs(gamma));
    numberOfIterations_ = iter;
    vecX = X;
    trace_p->pop();
  }

  /*!
     Solve matA.vecX=vecB using BiCG method with initial guess vecX  and an a preconditioner.
         Typename K (K : real or complex) is the type of the initial residual vector
         and thus of auxiliary vectors as matrices matA, P, vecB
         and vector vecX0 might not share the same type
  */
  template<typename K, class Mat, class VecB, class VecX, class Prec>
  void algorithm(Mat& matA, VecB& vecB, VecX& vecX, VecX& vecR, Prec& M)
  {
    trace_p->push(name_ + "Algorithm");

    VecX X(vecX);
    VecX Q, Qtilda, P, Ptilda;
    VecX R0 (vecR);     //R0 = B - A*X0;
    VecX R0tilda(R0);
    VecX Z, Ztilda, tmpP, tmpQ;
    K beta, alpha, rhoold=0, rho, gamma ;
    number_t iter=0;
    assign(gamma,dotRC(vecR, R0));  //gamma = dotRC(vecR, R0);
    K gammatilda;
    assign(gammatilda,dotRC(vecR, vecR));  //gammatilda = dotRC(R0tilda,R0tilda);

    while ( std::sqrt(std::abs(gammatilda))>std::abs(epsilon_) && std::sqrt(std::abs(gamma))>std::abs(epsilon_) && (iter<maxOfIterations_))
    {
      M.solve(R0,Z);
      M.transposedSolve(R0tilda,Ztilda);
      assign(rho,dotRC(Z,R0tilda));   //  rho=dotRC(Z,R0tilda);
      if ( std::abs(rho) < theBreakdownThreshold ) breakdown(std::abs(rho), "rho");
      if (iter==0)
      {
        P=Z;
        Ptilda=Ztilda;
      }
      else
      {
        beta=rho/rhoold;
        P*=beta;
        P+=Z;
        Ptilda*=beta;
        Ptilda+=Ztilda;
      }
      multMatrixVector(matA, P, Q);             //Q=A*P;
      multVectorMatrix(Ptilda, matA, Qtilda);   //Qtilda=Ptilda*A; //A^t*Ptilda;
      assign(alpha, 1. /dotRC(Ptilda, Q) );     //alpha=rho/Ptilda.dot(Q);
      alpha *= rho;
      tmpP = P;
      tmpP*=alpha;
      X+=tmpP;
      tmpQ = Q;
      tmpQ*=-alpha;
      R0+=tmpQ;                                //R0-=alpha*Q;
      tmpQ = Qtilda;
      tmpQ*=-alpha;
      R0tilda+=tmpQ;                           //R0tilda-=alpha*Qtilda;
      assign( gamma,  dotRC(R0, R0));          //gamma = dotRC(R0, R0) ;
      rhoold=rho;
      iter++;
    }
    numberOfIterations_ = iter;
    residue_ = std::sqrt(std::abs(gamma));
    vecX = X;
    trace_p->pop();
  }

}; // end of class BicgSolver ===================================================

} // namespace xlifepp

#endif /* BICG_SOLVER_HPP */
