/*
XLiFE++ is an extended library of finite elements written in C++
    Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file CgsSolver.hpp
  \authors C.Chambeyron, D.Martin, Manh Ha NGUYEN
  \since 24 Nov 2004
  \date  16 Oct 2012

  \brief Definition of the xlifepp::CgsSolver class

  Class xlifepp::CgsSolver implements the Conjugate Gradient Squared method for solving
  linear systems, with or without preconditioner.
  CGS is suitable for non symmetric linear systems.
  CGS roughly behaves as BiCG but does not involves transposed matrix vector

  Base class : xlifepp::IterativeSolver

  For elementary explanation about Conjugate Gradient Squared method,
  see http://mathworld.wolfram.com/ConjugateGradientSquaredMethod.html
*/

#ifndef CGS_SOLVER_HPP
#define CGS_SOLVER_HPP

#include "IterativeSolver.hpp"
#include "Preconditioner.hpp"

namespace xlifepp
{

/*!
    \class CgsSolver
    Implement Conjugate Gradient Squared algorithm
*/
class CgsSolver : public IterativeSolver
{
  public:
    //! Default constructor
    CgsSolver()
      : IterativeSolver("CGS") {}

    //! Full constructor
    CgsSolver(const real_t eps,
              const number_t maxOfIt = defaultMaxIterations,
              const number_t vb = 0)
      : IterativeSolver("CGS", maxOfIt, eps, vb) {}

    ~CgsSolver() {}

    //! Template Solvers without a preconditioner without argument "solType
    template<class Mat, class VecB, class VecX>
    VecX operator()(Mat& matA, VecB& vecB, VecX& vecX0)
    {
      if ( matA.valueType() == _real && vecB.valueType() == _real )
           return (*this)( matA,  vecB,  vecX0,  _real );
      else return (*this)( matA,  vecB,  vecX0,  _complex );
    }

    //! Template solvers without a preconditioner
    template<class Mat, class VecB, class VecX>
    VecX operator()(Mat& matA, VecB& vecB, VecX& vecX0, ValueType solType)
    {
      trace_p->push(name_ + "Solver");
      // Default value of maximum of iterations when not defined by user//
      maxOfIterations_ = maximumOfIterations(vecX0.size());
      printHeader(vecX0.size());
      VecX x(vecX0);
      //compute initial residue
      VecX r;
      multMatrixVector(matA, x, r);
      r*=-1.; r+=vecB;
      //run algorithm
      if (_real == solType) algorithm<real_t>(matA, vecB, x, r);
      else algorithm<complex_t>(matA, vecB, x, r);
      //output
      printOutput();
      printResult();
      trace_p->pop();
      return x;
    }

    //! Template solvers without a preconditionner (vecB = vecX0)
    template<class Mat, class VecB, class VecX>
    VecX operator()(Mat& matA, VecB& vecB, ValueType solType)
    {
      return (*this)(matA, vecB, vecB, solType);
    }

    //! Template Solvers with a preconditionner without argument "solType
    template<class Mat, class VecB, class VecX, class VecPC>
    VecX operator()(Mat& matA, VecB& vecB, VecX& vecX0, VecPC& pc)
    {
      if ( matA.valueType() == _real && vecB.valueType() == _real )
           return (*this)( matA,  vecB,  vecX0, pc, _real );
      else return (*this)( matA,  vecB,  vecX0, pc, _complex );
    }

    //! Template solvers with a preconditionner
    template<class Mat, class VecB, class Prec, class VecX>
    VecX operator()(Mat& matA, VecB& vecB, VecX& vecX0, Prec& pc, ValueType solType)
    {
      trace_p->push(name_ + "Solver");
      resetSolver();
      // Default value of maximum of iterations when not defined by user
      maxOfIterations_ = maximumOfIterations(vecX0.size());
      printHeader(vecX0.size(), pc.name());
      if ( pc.valueType() == _complex ) { solType = _complex; }
      // Define initial residual vector according to Type
      VecX x(vecX0);
      VecX r;
      multMatrixVector(matA, x, r);
      r*=-1.; r+=vecB;
      //run algorithm
      if ( _real == solType )  algorithm<real_t>(matA, vecB , x , r, pc);
      else algorithm<complex_t>(matA, vecB , x , r, pc);
      //output
      printOutput();
      printResult();
      trace_p->pop();
      return x;
    }

  private:
    /*!
      Solve matA.vecX=vecB using CGS method without preconditioning.
          Typename K (K : real or complex) is the type of the initial residual vector
          and thus of auxiliary vectors as matrices matA, vecB
          and vector vecX0 might not share the same type
     */
    template<typename K, class Mat, class VecB, class VecX>
    void algorithm(Mat& A, VecB& b, VecX& x, VecX& r)
    {
      trace_p->push(name_ + "algorithm");
      resetSolver();
      K rho_1(1), rho_2(1), alpha(1), beta(1);
      VecX p(r), q(r) , u(r), v(r), t(r);

      real_t normb = b.norm2();
      if ( normb< theBreakdownThreshold ) { normb = 1.; }
      residue_ = r.norm2() / normb;
      if (residue_ < epsilon_)
      {
         trace_p->pop();
         return;
      }
      VecX rtilde = r;
      bool firstIt=true;

      while ( numberOfIterations_ < maxOfIterations_  && residue_  > epsilon_)
      {
         assign(rho_1, dotRC(rtilde, r));
         if (std::abs(rho_1) < theBreakdownThreshold) breakdown(std::abs(rho_1),"rho_1");
         if (firstIt)
            {u = r; p = u; firstIt=false;}
         else
         {
            beta = rho_1 / rho_2;
            u = q ;  u *= beta;  u += r;           // u = r + beta * q;
            p *= beta; p += q; p *= beta; p += u;  // p = u + beta* (q + beta * p);
         }
         multMatrixVector(A, p, v);    // v = A * p
         assign(alpha, rho_1 / dotRC(rtilde, v));
         q=v;  q *= -alpha; q += u;    // q = u - alpha * v;
         u += q;                       // u = u + q
         t = u; t *= alpha; x += t;    // x += alpha * u;
         multMatrixVector(A, u, t);    // t = A * u;
         t *= -alpha; r += t;          // r -= alpha * q;
         rho_2 = rho_1;
         residue_ = r.norm2() / normb;

         // store current value of residue
         storeResidue();
         if ( verboseLevel_ > 0 ) printIteration();
         numberOfIterations_++;
      }
      trace_p->pop();
    }

    /*!
       Solve matA.vecX=vecB using CGS method with an optional preconditioner.
       Typename K (K : real or complex) is the type of the initial residual vector
           and thus of auxiliary vectors as matrices A, P, b
           and vector x might not share the same type
    */
    template<typename K, class Mat, class VecB, class VecX, class Prec>
    void algorithm(Mat& A, VecB& b, VecX& x, VecX& r, Prec& M)
    {
      trace_p->push(name_ + "algorithm");
      resetSolver();
      K rho_1(1), rho_2(1), alpha(1), beta(1);
      VecX p(r), q(r) , u(r), v(r), t(r), phat(r), qhat(r), uhat(r), vhat(r);

      real_t normb = b.norm2();
      if ( normb< theBreakdownThreshold ) { normb = 1.; }
      residue_ = b.norm2() / normb;
      if ((residue_) < epsilon_)
      {
         trace_p->pop();
         return;
      }
      VecX rtilde = r;
      bool firstIt=true;

      while ( ( numberOfIterations_ < maxOfIterations_ ) && (residue_  > epsilon_ ) )
      {
         assign(rho_1, dotRC(rtilde, r));
         if (std::abs(rho_1) < theBreakdownThreshold) breakdown(std::abs(rho_1),"rho_1");
         if (firstIt)
         {u = r; p = u; firstIt=false;}
         else
         {
            beta = rho_1 / rho_2;
            u = q ;  u *= beta;  u += r;           // u = r + beta * q;
            p *= beta; p += q; p *= beta; p += u;  // p = u + beta* (q + beta * p);
         }
         M.solve(p, phat);                  // preconditioner phat = M^{-1}p
         multMatrixVector(A, phat, vhat);   // vhat = A * phat
         assign(alpha, rho_1 / dotRC(rtilde, vhat));
         q=vhat;  q *= -alpha; q += u;      // q = u - alpha * vhat;
         u += q;                            // u = u + q
         M.solve(u, uhat);                  // preconditioner uhat = M^{-1}(u+q)
         uhat *= alpha; x += uhat;          // x += alpha * uhat;
         multMatrixVector(A, uhat, qhat);   // qhat = A * (alpha*uhat);
         r -= qhat;                         // r -= qhat;
         rho_2 = rho_1;
         residue_ = r.norm2() / normb;

         // store current value of residue
         storeResidue();
         if ( verboseLevel_ > 0 ) printIteration();
         numberOfIterations_++;
      }
      trace_p->pop();
    }

//    template<typename K, class Mat, class VecB, class VecX>
//    void algorithm(Mat& matA, VecB& vecB, VecX& vecX, VecX& vecR)
//    {
//      trace_p->push(name_ + "algorithm");
//      resetSolver();
//      real_t normofB = vecB.norm2();
//      if ( normofB < theBreakdownThreshold ) { normofB = 1.; }

//      // Some temporary variables
//      K alpha(1.), beta, theta;
//      VecX u(vecR), p(vecR), q(vecR), w(vecR), v;
//      VecX rtilda(vecR), utilda(vecR), uplusq(vecR);

//      assign(theta, dotRC(rtilda, vecR) );
//      residue_ = std::abs(theta) / normofB;
//      while ( ( numberOfIterations_ < maxOfIterations_ ) && (residue_  > epsilon_ ) )
//      {
//        multMatrixVector(matA, p, v); // v = matA * p
//        assign(alpha, theta / dotRC(rtilda, v) );
//        q=u;
//        v *= -alpha;
//        q+=v;          // q = u - alpha * v
//        uplusq = u;
//        uplusq+=q;     // uplusq = u + q;
//        utilda = uplusq;
//        utilda*=alpha; // utilda = (u+q)*alpha;
//        vecX+=utilda;  // vecX += utilda
//        multMatrixVector(matA, utilda, w); // vecR -= matA * utilda
//        w*=-1.;
//        vecR+=w;
//        beta = 1. / theta;
//        assign(theta, dotRC(rtilda, vecR)); // theta=rtilda.vecR;

//        if (std::abs(theta) < theBreakdownThreshold) breakdown(std::abs(theta),"theta");
//        residue_ = (std::abs(theta) / normofB );
//        if ( residue_  < epsilon_  ) { break; }  // Convergence when abs(theta)/normofB is small enough

//        beta *= theta;
//        q *= beta;
//        u = vecR ;
//        vecR+=q;     //  u = vecR + q;
//        p *= beta*beta;
//        p+=u;
//        p+=q;  //p += (u + q);

//        // store current value of residue
//        storeResidue();

//        if ( verboseLevel_ > 0 ) { printIteration(); }
//        numberOfIterations_++;
//      }
//      trace_p->pop();
//    }

//   /*!
//       Solve matA.vecX=vecB using CGS method with an optional preconditioner.
//       Typename K (K : real or complex) is the type of the initial residual vector
//           and thus of auxiliary vectors as matrices matA, P, vecB
//           and vector vecX0 might not share the same type
//   */
//   template<typename K, class Mat, class VecB, class VecX, class Prec>
//   void algorithm(Mat& matA, VecB& vecB, VecX& vecX, VecX& vecR, Prec& pc)
//   {
//      trace_p->push(name_ + "algorithm");
//      resetSolver();
//      real_t normofB = vecB.norm2();
//      if ( normofB < theBreakdownThreshold ) { normofB = 1.; }
//      K alpha(1.), beta, theta;
//      VecX u(vecR), p(vecR), rtilda(vecR), utilda(vecR), uplusq(vecR), q(vecR), w, v;

//      // In comments, read Identity Matrix for P when not preconditioned
//      VecX phat(vecR);
//      pc.solve(p, phat);

//      assign(theta, dotRC(rtilda, vecR) );
//      residue_ = std::abs(theta) / normofB;

//      while ( numberOfIterations_ < maxOfIterations_ )
//      {
//         //<< v = matA * phat >>
//         multMatrixVector(matA, phat, v);

//         assign(alpha, theta / dotRC(rtilda, v) );
//         //<< q = u - alpha * v >>
// //        multScalarThenAssign(v, alpha);
// //        q = u - v;
// //        multScalarThenAssign(v, -1. );
//         v*=-alpha;
//         q=u;
// //        addVectorThenAssign(q, v);
//         q+=v;

//         //<< uplusq = u + q >>
//         uplusq = u;
// //        addVectorThenAssign(uplusq, q);
//         uplusq+=q;

//         pc.solve(uplusq, utilda);

// //        multScalarThenAssign(utilda, alpha);
//         utilda *= alpha;

//         //<< vecX += utilda >>
// //        addVectorThenAssign(vecX, utilda);
//         vecX+=utilda;

//         //<< vecR -= matA * utilda >>
//         multMatrixVector(matA, utilda, w);
// //        multScalarThenAssign(w, -1.);
// //        addVectorThenAssign(vecR, w);
//         w*=-1.;
//         vecR+=w;

//         beta = 1. / theta;
//         assign(theta, dotRC(rtilda, vecR));  //theta=rtilda.vecR;

//         if ( (residue_ = std::abs(theta) / normofB ) < epsilon_ ) {  break; }

//         // Convergence when abs(theta)/normofB is small enough
//         residue_ = (std::abs(theta) / normofB );
//         if ( residue_  < epsilon_  ) { break; }
//         if ( verboseLevel_ > 0 ) { printIteration(); }

//         beta *= theta;

// //        multScalarThenAssign(q, beta);
//         q *= beta;

//         //<< u = vecR + q >>
//         u = vecR;
// //        addVectorThenAssign(u, q);
//         vecR+=q;
// //        multScalarThenAssign(p, beta * beta);
//         p *= beta * beta;

//         //<< p += u + q >>
// //        addVectorThenAssign(p, u);
// //        addVectorThenAssign(p, q);
//         p+=u;
//         p+=q;

//         pc.solve(p, phat);

//         // store current value of residue
//         storeResidue();
//         numberOfIterations_++;
//      }
//      trace_p->pop();
//   }

}; // end of class CgsSolver

} // namespace xlifepp

#endif /* CGS_SOLVER_HPP */
