/*
XLiFE++ is an extended library of finite elements written in C++
    Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file Space.hpp
  \author E. Lunéville
  \since 24 fev 2012
  \date 13 jun 2013

  \brief Definition of the classes related to spaces

  This is the dicrete spaces defined by a geometrical domain (from mesh)
  and an interpolation (finite element or set of functions)

  xlifepp::Space is the parent class of inherited specialized classes :
  \verbatim
        pointer to        child
  Space ---------> Space -------| FeSpace   (finite element)
                                | SpSpace   (spectral basis)
                                | SubSpace  (subspace of a space) ----> FeSubSpace (specific subspace FE data)
                                | ProdSpace (tensor product of spaces)
  \endverbatim

  A xlifepp::Space object has a pointer to an other xlifepp::Space object which points to the "true" space
  It is a trick to avoid the endusers to explicitely instanciate child space objects
  The spaceInfo_p pointer points to a xlifepp::SpaceInfo object shared by the "user" space and the "true" space
  In this paradigm, any member function of space encapsulating child member function looks like

  \code
  virtual ret_arg Space::member_fun(args)
  {
    if (space_p != this) space_p->member_fun(args); // call the same with the "true" space
    else { ... }                                    // if the "true" space,
  }

  ret_arg ChildSpace::member_fun(args)              // the "true" member function
    { ... }
  \endcode

  Because of polymorphic behaviour, if a child do not propose member_fun it is the else block of xlifepp::Space::member_fun
  which is applied. Do not work with templated member function (no virtual template member function in C++) !!!

  Additionnal class :  xlifepp::SpaceInfo : general space information
*/

#ifndef SPACE_HPP
#define SPACE_HPP

#include "config.h"
#include "Dof.hpp"
#include "geometry.h"
#include "SpectralBasis.hpp"

namespace xlifepp
{

//forward declarations
class FeSpace;
class SpSpace;
class SubSpace;
class FeSubSpace;
class TermVectors;

/*!
   \class SpaceInfo
   structure to store general information of a space
 */
class SpaceInfo
{
  public:
    string_t name;                //!< name of space
    const GeomDomain* domain_p;   //!< geometric domain supporting the space
    SobolevType spaceConforming;  //!< conforming space : L2 or H1 or Hdiv or Hcurl or H2 or Hinf
    dimen_t dimFun;               //!< number of components of the basis functions
    SpaceType spaceType;          //!< type of space definition (FESpace,SPSpace,SubSpace,ProductSpace,OtherSpace)
    SpaceInfo() {}                //!< default constructor
    //! general constructor
    SpaceInfo(const string_t& na, const GeomDomain* gd, SobolevType so, dimen_t dF, SpaceType st)
      : name(na), domain_p(gd), spaceConforming(so), dimFun(dF), spaceType(st) {}
};

/*!
   \class Space
   base class of spaces
 */
class Space
{
  protected:
    Space* space_p;               //!< pointer to the "true" space
    SpaceInfo* spaceInfo_p;       //!< pointer to the space information structure

  public:
    mutable bool global;                        //!< declare space as global, if true forbid deletion of pointer space_p
    Space() {}                                  //!< default constructor
    virtual ~Space();                           //!< destructor
    void printSpaceInfo(std::ostream&) const;   //!< print utility
    void printSpaceInfo(PrintStream& os) const {printSpaceInfo(os.currentStream());}
    virtual void print(std::ostream&) const;    //!< print utility
    virtual void print(PrintStream& os) const {print(os.currentStream());}

    //static definitions
    static std::vector<Space*> theSpaces;                        //!< unique list of defined spaces
    static void clearGlobalVector();                             //!< delete all space objects
    static Space* findSubSpace(const GeomDomain*, Space*);       //!< find subspace in list
    static const Space* findSubSpace(const GeomDomain*, const Space*); //!< find subspace in list
    std::vector<Space*> findSubSpaces(Space*);                   //!< find subspaces linked to a space
    std::vector<Space*> findSubSpaces(const Space*);             //!< find subspaces linked to a space


    // user space constructors
    //-------------------------------------------------------------------
    Space(number_t, const string_t& ="Rn");                                         //!< constructor of trivial space of dim n (say Rn)
    Space(const GeomDomain&, number_t, const string_t& ="Rn");                      //!< constructor of trivial space of dim n with domain

    //constructors and functions addressing SpSpace type
    Space(const SpectralBasis&, const string_t&);                                      //!< constructor of spectral space from explicit spectral basis
    Space(const GeomDomain&, const Function&, number_t, dimen_t, const string_t&);     //!< constructor of spectral space from an analytical function
    Space(const GeomDomain&, const Function&, number_t, dimen_t);                      //!< constructor of spectral space from an analytical function
    Space(const GeomDomain&, const Function&, number_t, const string_t&);              //!< constructor of spectral space from an analytical function
    void buildSpFun(const GeomDomain&, const Function&, number_t, dimen_t, const string_t&); //!< real constructor of analytic functions spectral space
    Space(const TermVectors&, const string_t&);                                        //!< constructor of spectral space from interpolated functions (TermVectors)
    //  implemented in TermVector.cpp

    //constructors and functions addressing FeSpace type
    Space(const GeomDomain&, const Interpolation&, const string_t&, bool opt = true); //!< constructor from GeomDomain and scalar Interpolation (FeSpace)
    Space(const GeomDomain&, FEType, number_t, const string_t&, bool opt = true);     //!< constructor from GeomDomain and Finite Element (FeSpace)
    Space(const GeomDomain&, PolynomType, const string_t&, bool opt = true);          //!< constructor from GeomDomain and Pk polynom (FeSpace)
    Space(const GeomDomain&, QolynomType, const string_t&, bool opt = true);          //!< constructor from GeomDomain and Qk polynom (FeSpace)
    Space(const GeomDomain&, FeFaceType, const string_t&, bool opt = true);           //!< constructor from GeomDomain and FE face type (Hdiv FeSpace)
    Space(const GeomDomain&, FeEdgeType, const string_t&, bool opt = true);           //!< constructor from GeomDomain and FE edge type (Hrot FeSpace)

    //constructors and functions addressing SubSpace type
    Space(const GeomDomain&, Space&, const string_t& na = "");                               //!< constructor from GeomDomain and Space (SubSpace)
    Space(const GeomDomain&, const std::vector<number_t>&, Space&, const string_t& na = ""); //!< constructor from GeomDomain, dofIds list and Space (SubSpace)
    Space(const std::vector<number_t>&, Space&, const string_t& na = "");                    //!< constructor from dofIds list and Space (SubSpace)
    SubSpace* createSubSpace(const GeomDomain&, Space&, const string_t& na);                 //!< actual subspace constructor
    void createSubSpaces(const std::vector<const GeomDomain*>&,std::vector<Space*>&) ;       //!< create subspaces of current space from a list of subdomains
    Space* buildSubspaces(const std::vector<const GeomDomain*>&, std::vector<Space*>&);      //!< build subspaces from a list of subdomains and the largest subspace

    //access to SpaceInfo properties
    //-------------------------------------------------------------------
    string_t name() const {return spaceInfo_p->name;}                           //!< returns space name
    SobolevType conformingSpace() const {return spaceInfo_p->spaceConforming;}  //!< returns space conformity
    const GeomDomain* domain() const {return spaceInfo_p->domain_p;}            //!< returns space related domain
    dimen_t dimDomain() const {return spaceInfo_p->domain_p->dim();}            //!< returns dimension of space related domain
    dimen_t dimPoint() const {return spaceInfo_p->domain_p->spaceDim();}        //!< returns dimension of points
    dimen_t dimFun() const {return spaceInfo_p->dimFun;}                        //!< returns dimension of basis functions
    SpaceType typeOfSpace() const {return spaceInfo_p->spaceType;}              //!< returns type of space
    SpaceType typeOfSubSpace() const;                                           //!< returns type of sub-space
    const Space* space() const {return space_p;}                                //!< returns true space object
    const SpaceInfo* spaceInfo() const {return spaceInfo_p;}                    //!< returns true space object
    virtual number_t dimSpace() const;                            //!< space dimension (number of dofs)
    virtual number_t nbDofs()const;                               //!< number of dofs (a dof may be a vector dof)
    virtual const FeSpace* feSpace() const;                       //!< access to child fespace object (const)
    virtual FeSpace* feSpace();                                   //!< access to child fespace object
    virtual const FeSubSpace* feSubSpace() const;                 //!< access to child fesubspace object (const)
    virtual const SpSpace* spSpace() const;                       //!< access to child spspace object (const)
    virtual const SubSpace* subSpace() const;                     //!< access to child subspace object (const)
    virtual SubSpace* subSpace();                                 //!< access to SubSpace from Space class
    virtual const Space* rootSpace() const;                       //!< access to root space pointer (const)
    virtual Space* rootSpace();                                   //!< access to root space pointer
    virtual number_t dofId(number_t n) const;                     //!< return the n-th DoF id of space
    virtual std::vector<number_t> dofIds() const;                 //!< return the DoF ids on space
    virtual std::pair<number_t, number_t> renumberDofs();         //!< optimize dofs numbering to reduce matrix bandwith
    virtual const Dof& dof(number_t n) const;                     //!< return n-th dof (n=1,...)
    virtual bool include(const Space*) const;                     //!< return true if space is included in current space
    virtual ValueType valueType() const;                          //!< return value type of basis function (real or complex)
    virtual StrucType strucType() const;                          //!< return structure type of basis function (scalar or vector)
    virtual bool isSpectral() const;                              //!< true if spectral space
    virtual bool isFE() const;                                    //!< true if FE space or FeSubspace
    virtual bool extensionRequired() const;                       //!< true if space has no trace space

    //specific virtual functions for FE type space (FeSpace and FeSubSpace), used in FE computation
    virtual const Interpolation* interpolation() const                //! return pointer to interpolation
    {return space_p->interpolation();}
    virtual number_t nbOfElements() const                             //! number of elements for FESpace and FESubspace only
    {return space_p->nbOfElements();}
    virtual const Element* element_p(number_t k) const                //! access to k-th element (pointer) for FESpace and FESubspace only
    {return space_p->element_p(k);}
    virtual const Element& element(number_t k) const                   //! access to k-th (k>=0) element for FESpace and FESubspace only
    {return space_p->element(k);}
    virtual const Element* element_p(GeomElement* gelt) const         //! access to element (pointer) associated to a geomelement
    {return space_p->element_p(gelt);}
    virtual const Element& element(GeomElement* gelt) const           //! access to element associated to a geomelement
    {return space_p->element(gelt);}
    virtual number_t numElement(GeomElement* gelt) const              //! access to element number associated to a geomelement
    {return space_p->numElement(gelt);}
    virtual const std::vector<number_t>& elementDofs(number_t k) const//! access to dofs ranks (local numbering) of k-th element for FESpace and FESubspace only
    {return space_p->elementDofs(k);}
    virtual std::vector<number_t> elementParentDofs(number_t k) const //! access to dofs ranks (parent numbering) of k-th element for FESpace and FESubspace only
    {return space_p->elementParentDofs(k);}
    virtual const std::set<RefElement*>& refElements() const         //! access to set of pointer to RefElement involved in FESpace and FESubspace only
    {return space_p->refElements();}
    virtual void buildgelt2elt() const                               //! construct the map gelt2elt from elements list
    {return space_p->buildgelt2elt();}
    virtual const std::vector<FeDof>& feDofs() const                 //! access to vector of fedofs
    {return space_p->feDofs();}
    virtual const Point& feDofCoords(number_t k) const               //! return coordinates of dot as a point (may be virtual coordinates when not a Lagrange dof)
    {return feDofs()[k].coords();}
    virtual void builddofid2rank() const                             //! construct the map dofid2rank from dofs list
    {return space_p->builddofid2rank();}
    virtual const std::map<number_t, number_t>&  dofid2rank() const  //! return the map dof to rank, built on fly (only for FeSpace and SubSpace)
    {return space_p->dofid2rank();}

    //specific virtual functions for SP type space
    virtual void shiftDofs(number_t n)                               //! shift dofs numbering from n (special utility for SpSpace)
    {return space_p->shiftDofs(n);}

    //print utilities
    static void printAllSpaces(std::ostream&);                        //!< output all spaces in memory
    friend std::ostream& operator<<(std::ostream&, const Space&);
};

std::ostream& operator<<(std::ostream&, const Space&);     //!< output space characteristics

//extern functions
bool compareSpaceUsingSize(const Space*, const Space*);               //!< compare spaces using their sizes
std::vector<number_t> renumber(const Space*, const Space*);           //!< renumber dofs of a space along dof numbering of an another space
Space& subSpace(Space&, const GeomDomain&);                           //!< return sub-space or trace space on domain of a space, created if not exist
inline Space& operator |(Space& V, const GeomDomain& dom)             //!< return sub-space or trace space on domain of a space, created if not exist
{return subSpace(V,dom);}
Space* mergeSubspaces(Space* &, Space* &, bool newSubspaces=false);   //!< merge two subspaces
Space* mergeSubspaces(std::vector<Space*>&, bool newSubspaces=false); //!< merge a list of subspaces
Space* unionOf(std::vector<Space*>&);                                 //!< union of subspaces
std::vector<Point> dofCoords(Space&, const GeomDomain&);              //!< list of coordinates of ponctual dofs of Space on a given domain
void clearTerms(const Space&);                                        //!< clear Terms depending on a space (implemented in Term.cpp)
void clearStorages(const Space&);                                     //!< clear Storages depending on a space (implemented in Term.cpp)


//==================================================================================================
// to deal with collections of spaces
//==================================================================================================
typedef PCollection<Space> Spaces; //!< collection of Space pointers

//@{
//! build a collection of space according to collections of domains and collection of interpolations
inline Spaces spaces(const Domains& doms,const Interpolations& ints, bool opt=true) // few domains, few interpolations
{
  Spaces vs;
  std::vector<GeomDomain*>::const_iterator itd=doms.begin();
  std::vector<Interpolation*>::const_iterator iti=ints.begin();
  for(; itd!=doms.end(); ++itd, ++iti) vs<<(*new Space(**itd,**iti,(*iti)->shortname+"_"+(*itd)->name(),opt));
  return vs;
}

inline Spaces spaces(const Domains& doms,const Interpolation& inte, bool opt=true) // few domains, one interpolation
{
  Spaces vs;
  std::vector<GeomDomain*>::const_iterator itd=doms.begin();
  for(; itd!=doms.end(); ++itd) vs<<(*new Space(**itd, inte, inte.shortname+"_"+(*itd)->name(),opt));
  return vs;
}

inline Spaces spaces(const GeomDomain& dom,const Interpolations& ints, bool opt=true) // few domains, one interpolation
{
  Spaces vs;
  std::vector<Interpolation*>::const_iterator iti=ints.begin();
  for(; iti!=ints.end(); ++iti) vs<<(*new Space(dom, **iti, (*iti)->shortname+"_"+dom.name(),opt));
  return vs;
}
//@}

#ifndef DOXYGEN_SHOULD_SKIP_THIS

/* specialization of PCollectionItem
    when Vs is a Spaces allows to write Vs(i) = Space(...) with no hard copy
*/
template<> class PCollectionItem<Space>
{
  public:
    std::vector<Space*>::iterator itp;
    PCollectionItem(std::vector<Space*>::iterator it) : itp(it) {}
    Space& operator = (const Space& sp)
    {
      //create a copy of sp on the heap, shared pointers
      Space & nsp=*new Space(sp);  //copy constructor
      sp.global=true;   //declare original space as global to disable deletion
      *itp=const_cast<Space*>(&nsp);
      return **itp;
    }
    operator Space&() {return **itp;} //autocast PCollectionItem->T&
};

#endif

//==================================================================================================
/*! a class handling the map between dof numbers of the trace of 2 spaces (V1,V2) on an interface
    the spaces have to be FE spaces on a same mesh with same interpolation !
    This class manages a numbering vector, say m, that relates dof numbers of V1 to dof numbers of V2 :
                    i_th dof of V1 = m(i)_th dof of V2
    This map allows to transport a TermVector defined on V1 to a TermVector defined on V2
*/
//==================================================================================================
class SpaceMap
{
  private:
    const GeomDomain* domain_p;       //!< interface domain
    const Space* V1_p, *V2_p;         //!< spaces involved in the mapping
    void buildMap();                  //!< effective construction of the map

  public:
    std::vector<number_t> dofMap;     //!< numbering vector describing the map V1->V2 (index start from 0)
    SpaceMap(const Space& V1, const Space& V2, const GeomDomain& dom)   // main constructor
      : domain_p(&dom), V1_p(&V1), V2_p(&V2) {buildMap();}

    number_t operator()(number_t i) const //!< return the dof number of V2 of the dof number of V1 (index start from 1)
    {return dofMap[i-1];}

   //io functions
    void print(std::ostream&) const;                           //!< print SpaceMap
    void print(PrintStream& os) const {print(os.currentStream());}
};

inline std::ostream& operator<<(std::ostream& os, const SpaceMap& sm)
{
   sm.print(os);
   return os;
}

} // end of namespace xlifepp

#endif // SPACE_HPP
