/*
XLiFE++ is an extended library of finite elements written in C++
    Copyright (C) 2014  Lunville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file BilinearFormAsLinearForm.hpp
  \author E. Lunville
  \since 23 mar 2012
  \date  01 apr 2012

  \brief Definition of the xlifepp::LinearForm classes
*/
#ifndef BILINEARFORM_AS_LINEARFORM_HPP
#define BILINEARFORM_AS_LINEARFORM_HPP

#include "config.h"
#include "KernelOperatorOnTermVector.hpp"

namespace xlifepp
{
/*!
   \class BilinearFormAsLinearForm
   describes a linear form based on a Bilinearform a(u,v) where u or v are specified as a TermVector
*/
class BilinearFormAsLinearForm : public BasicLinearForm
{
  protected :
      BasicBilinearForm* blf_;     //!< Basic bilinear form a(u,v)
      const TermVector* tv_;       //!< TermVector
  public :
      bool asUnknown_;        //!< true if tv_ is realted to unknown u, false if tv_ is related to test function v

  //! Constructor with a reference of BasicLinearForm
  BilinearFormAsLinearForm(const BasicBilinearForm& blf, const TermVector& tv, bool asUnknown)
     : blf_(blf.clone()), tv_(&tv), asUnknown_(asUnknown) {}
  //! Constructor with a pointer of BasicBilinearForm   
  BilinearFormAsLinearForm(BasicBilinearForm* blf, const TermVector& tv, bool asUnknown)
     :  blf_(blf), tv_(&tv), asUnknown_(asUnknown) {if(asUnknown) u_p=&blf->vp(); else u_p=&blf->up();}
  
  //! copy-constructor
  BilinearFormAsLinearForm(const BilinearFormAsLinearForm& blfaslf)
     {
         u_p = blfaslf.u_p;
         blf_= blfaslf.blf_->clone();
         tv_= blfaslf.tv_;
         asUnknown_=blfaslf.asUnknown_;
     }
  //! operator = (copy-constructor)
  BilinearFormAsLinearForm& operator = (const BilinearFormAsLinearForm& blfaslf)
    {
         if(this == &blfaslf) return *this;
         if(blf_!=0) delete blf_;
         u_p = blfaslf.u_p;
         blf_= blfaslf.blf_->clone();
         tv_= blfaslf.tv_;
         asUnknown_=blfaslf.asUnknown_;
         return *this;
    }
    //! destructor
  ~BilinearFormAsLinearForm()
    {if(blf_!=0) delete blf_;}  
//! return the attribute basicLinearForm_
  BasicBilinearForm* blf() const {return blf_;}
//! return the attribute TermVector tv_
  const TermVector* termVector() const {return tv_;}
  virtual BasicLinearForm* clone() const        //! clone of the linear form
     {return new BilinearFormAsLinearForm(*this);}
  virtual LinearFormType type() const            //! return the type of the linear form
    {return _bilinearAsLinear;}
  virtual ValueType valueType() const            //! return the value type of the linear form
    {return blf_->valueType();}
  virtual StrucType strucType() const            //! return the structure type of the linear form
    {if(asUnknown_) return blf_->vp().strucType(); else return blf_->up().strucType();}
  const GeomDomain* domain() const               //! return the pointer to the domain
    {if(asUnknown_) return &blf_->dom_vp(); else return &blf_->dom_up();}

  virtual string_t asString() const
    {return "";}
  virtual void print(std::ostream&) const;       //!< print utility
  virtual void print(PrintStream& os) const {print(os.currentStream());}  //!< print to current stream
};

//!reinterpretation of double integrals as linear forms
LinearForm intg(const GeomDomain& domx, const GeomDomain& domy,
                const KernelOperatorOnTermVectorAndUnknown& koptvv,
                const IntegrationMethod& im);
LinearForm intg(const GeomDomain& domx, const GeomDomain& domy,
                const KernelOperatorOnTermVectorAndUnknown& koptvv,
                QuadRule qr, number_t qro);
LinearForm intg(const GeomDomain& domx, const GeomDomain& domy,
                const KernelOperatorOnTermVectorAndUnknown& koptvv);

} //end of namespace xlifepp

#endif // BILINEARFORM_AS_LINEARFORM_HPP
