
/*
XLiFE++ is an extended library of finite elements written in C++
  Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file SuTermMatrix.hpp
  \author E. Lunéville
  \since 23 mar 2012
  \date 13 jan 2014

  \brief Definition of the xlifepp::SuTermMatrix class

  xlifepp::SuTermMatrix class handles numerical representation of a single unknown (couple) bilinear form
  and has the following items:
  - the largest subspace of unknown space involved in bilinear form, possibly the whole space
  - the largest subspace of testfunction space involved in bilinear form, possibly the whole space
  - the list of subspaces of unknown/testfunction spaces involved by each basic bilinear form
  - a list of single unknown/testfunction essential boundary conditions

  xlifepp::SuTermMatrix inherits from xlifepp::Term and it is a piece of xlifepp::TermMatrix

  Global organization looks like :

  \verbatim
      child
  Term  --------> SuTermMatrix
          | SuBilinearForm
          | MatrixEntry *
          | ...
  \endverbatim

  xlifepp::SuTermMatrix is an internal object never managed by end user !

  Note : if required it is possible to go to scalar representation of matrix by using scalar_entries_p
     and renumbering vectors cdofs_u (column) and cdofs_v (row).
     This representation has no interest if all the unknowns involved in xlifepp::SuTermMatrix are scalar !
*/


#ifndef SU_TERM_MATRIX_HPP
#define SU_TERM_MATRIX_HPP

#include "config.h"
#include "Term.hpp"
#include "TermVector.hpp"
#include "SpectralBasisInt.hpp"
#include "form.h"
#include "largeMatrix.h"
#include "computation/termUtils.hpp"
#include "TensorKernel.hpp"
#include "BilinearFormAsLinearForm.hpp"
#include "hierarchicalMatrix.h"

#include "computation/SauterSchwabIM.hpp"
#include "computation/DuffyIM.hpp"
#include "computation/LenoirSallesIM.hpp"


namespace xlifepp{

//forward class declaration
class TermMatrix;

/*!
   \class SuTermMatrix
   handles numerical representation
      of single unknowns bilinear forms (SuBilinearForm). Internal class
*/
class SuTermMatrix : public Term
{
  protected :
  SuBilinearForm* sublf_p;           //!< single unknown bilinear form (pointer)
  const Unknown* u_p;                //!< pointer to unknown (column)
  const Unknown* v_p;                //!< pointer to test function (line)
  mutable Space* space_u_p;          //!< largest subspace involved in bilinear form for unknown u, may be the whole space
  mutable Space* space_v_p;          //!< largest subspace involved in bilinear form for testfunction v, may be the whole space
  std::vector<Space *> subspaces_u;  //!< list of subspaces related to each basic bilinear form's GeomDomain
  std::vector<Space *> subspaces_v;  //!< list of subspaces related to each basic bilinear form's GeomDomain
  MatrixEntry* entries_p;            //!< values of SuTermMatrix

  //for scalar representation if required
  MatrixEntry* scalar_entries_p;     //!< values of SuTermMatrix in scalar representation, same as entries_p if scalar unknowns
  std::vector<DofComponent> cdofs_u; //!< scalar numbering of u (in component dofs)
  std::vector<DofComponent> cdofs_v; //!< scalar numbering of v (in component dofs)

  //to deal with constraints
  MatrixEntry* rhs_matrix_p;         //!< pointer to a correction matrix when constraints applied

  //for hierarchical representation
  HMatrixEntry<FeDof>* hm_entries_p; //!< values of SuTermMatrix in a hierarchical matrix representation
  ClusterTree<FeDof>*  cluster_u;    //!< column cluster tree
  ClusterTree<FeDof>*  cluster_v;    //!< row cluster tree
  HMatrixEntry<FeDof>* hm_scalar_entries_p; //!< values of scalar form of SuTermMatrix in a hierarchical representation (=hm_entries_p if scalar SuTermMatrix)

  public :
  SuTermMatrix(const Unknown*, Space*, const Unknown*, Space*, MatrixEntry*, const string_t& na="");  //!< basic constructor
  SuTermMatrix(SuBilinearForm* sublf = 0, const string_t& na="", bool noass=false);                   //!< basic constructor
  SuTermMatrix(SuBilinearForm* sublf, const string_t& na="", ComputingInfo cp= ComputingInfo());      //!< basic constructor
  SuTermMatrix(SuBilinearForm*, const Unknown*, const Unknown*, Space*, Space*,
         const std::vector<Space *>&, const std::vector<Space *>&, const string_t&, MatrixEntry* =0); //!< full constructor

  SuTermMatrix(const Unknown&, const GeomDomain&, const Unknown&, const GeomDomain&,
         const OperatorOnKernel&, const string_t& na="");//!< construtcor of matrix opker(xi,yj) (Lagrange interpolation)
  SuTermMatrix(SuTermVector &, const string_t& na="");         //!< constructor of a diagonal SuTermMatrix from a SuTermVector
  SuTermMatrix(const Unknown*, Space*, const Unknown*, Space*, SuTermVector&, StorageType=_noStorage, AccessType=_noAccess,
         const string_t& na="");                         //!< constructor of a diagonal SuTermMatrix from a SuTermVector
  void diagFromSuTermVector(const Unknown*, Space*, const Unknown*,Space*, SuTermVector&, StorageType=_noStorage, AccessType=_noAccess,
                const string_t& na="");            //!< actual constructor of diagonal SuTermMatrix from a SuTermVector
  SuTermMatrix(const SuTermMatrix&,const string_t& na="");     //!< copy constructor
  SuTermMatrix(const SuTermMatrix&, SpecialMatrix, const string_t& na=""); //!< construct a special matrix from a SuTermMatrix (e.g _IdMatrix)
  virtual ~SuTermMatrix();                                     //!< destructor
  SuTermMatrix& operator=(const SuTermMatrix&);                //!< assign operator
  SuTermMatrix(const LcTerm<SuTermMatrix>&,const string_t& na=""); //!< constructor from a linear combination of SuTermMatrix's
  void copy(const SuTermMatrix&);                              //!< copy SuTermMatrix in current SuTermMatrix
  void initPointers();                                         //!< set all pointers to 0
  void initSuTermVector(SuTermVector&, ValueType, bool col=true) const;    //!< initialize a SuTermVector consistent with matrix rows or columns

  //accessors and properties
  string_t name() const { return Term::name(); }
  //! update the name of a SuTermMatrix
  void name(const string_t& nam)
  { name_=nam; }
  //! return pointer to bilinear form
  SuBilinearForm*& sublfp()
  {return sublf_p;}
  //! return _sutermMatrix
  TermType termType() const
  {return _sutermMatrix;}
  ValueType valueType() const;            //!< return value type (_real,_complex)
  StrucType strucType() const;            //!< return structure type (_scalar,_matrix)
  Space& space_u() const
  {return *space_u_p;}                   //!< return the largest space in u involved in SuBilinearForm
  Space& space_v() const                  //! return the largest space in v involved in SuBilinearForm
  {return *space_v_p;}
  Space* space_up() const                 //! return the pointer to the largest space in u involved in SuBilinearForm
  {return space_u_p;}
  Space* space_vp() const                 //! return the pointer to the largest space in v involved in SuBilinearForm
  {return space_v_p;}
  Space*& space_up()
  {return space_u_p;}                      //!< return the pointer to the largest space in u involved in SuBilinearForm (write)
  Space*& space_vp()
  {return space_v_p;}                      //!< return the pointer to the largest space in v involved in SuBilinearForm (write)
  const Unknown* up() const
  {return u_p;}                                   //!< return pointer to unknown (column)
  const Unknown* vp() const
  {return v_p;}                                  //!< return pointer to testfunction (line)
  const Unknown*& up()
  {return u_p;}                                   //!< return pointer to unknown (column) (write)
  const Unknown*& vp()
  {return v_p;}                                        //!< return pointer to testfunction (line) (write)
  number_t numberOfRows() const;          //!< number of rows (counted in scalar dof)
  number_t numberOfCols() const;          //!< number of columns (counted in scalar dof)
  std::set<const Space*> unknownSpaces() const;          //!< list of involved unknown spaces

  MatrixEntry*& entries()
  {return entries_p;}                               //!< access to entries pointer (r/w)
  const MatrixEntry* entries() const
  {return entries_p;}                               //!< access to entries pointer (r)
  MatrixEntry*& scalar_entries()
  {return scalar_entries_p;}                     //!< access to entries pointer (r/w)
  const MatrixEntry* scalar_entries() const
  {return scalar_entries_p;}                     //!< access to entries pointer (r)
  MatrixEntry* actual_entries() const;           //!< return the actual pointer to entries (priority to scalar entry)
  MatrixEntry*& rhs_matrix()
  {return rhs_matrix_p;}                          //!< access to rhs matrix pointer (r/w)
  const MatrixEntry* rhs_matrix()   const
  {return rhs_matrix_p;}                            //!< access to rhs matrix pointer (r)
  template <typename T>
  LargeMatrix<T>& getLargeMatrix(StrucType str=_undefStrucType) const;  //!<get LargeMatrix representation
  template<typename T>
  HMatrix<T,FeDof>& getHMatrix(StrucType str=_undefStrucType) const;    //!<get HMatrix representation if it exists

  const std::vector<DofComponent>& cdofsu() const
  {return cdofs_u;}                       //!< access to cdofs_u vector (r)
  //! access to cdofs_v vector (r)
  const std::vector<DofComponent>& cdofsv() const
  {return cdofs_v;}
  //! access to cdofs_u vector (r/w)
  std::vector<DofComponent>& cdofsu()
  {return cdofs_u;}
   //! access to cdofs_v vector (r/w)
  std::vector<DofComponent>& cdofsv()
  {return cdofs_v;}

  bool hasHierarchicalMatrix() const
  {return (hm_entries_p!=0 || hm_scalar_entries_p!=0);}

  number_t nnz() const;                     //!< return the number of non zero (scalar non zero)

  SymType symmetry() const;                 //!< return symmetry of submatrix
  FactorizationType factorization() const;  //!< return factorization type if factorized
  StorageType storageType() const;          //!< return storage type of submatrix (_cs,_skyline,_dense)
  AccessType accessType() const;            //!< return access type of submatrix (_row, _col, _sym, _dual)
  MatrixStorage* storagep() const;          //!< return storage of submatrix (0 if no entries or storage)
  MatrixStorage* scalarStoragep() const;    //!< return scalar storage of submatrix (0 if no scalar entries or storage)
  void setStorage(StorageType, AccessType); //!< set the storage of SuTermMatrix
  void toStorage(StorageType, AccessType);  //!< change the storage of SuTermMatrix (real algorithm)

  //utilities
  void buildSubspaces();                   //!< build the required subspaces and largest subspaces
  void clear();                            //!< deallocate memory used by a SuTermMatrix
  void clearScalar();                      //!< deallocate memory used by scalar part of SuTermMatrix

  //compute facilities
  void compute();                                    //!< compute SuTermMatrix from bilinear form
  void compute(const LcTerm<SuTermMatrix>&,const string_t& na=""); //!< compute SuTermMatrix from a linear combination of SuTermMatrix's
  template<number_t CM>
  void compute(const std::vector<SuBilinearForm>&, ValueType, StrucType);   //!< utility for computing SuTermMatrix
  template<typename T, typename K>
  void computeIR(const SuLinearForm& sulf, T* mat, K& vt, const std::vector<Point>& xs,
           const std::vector<Vector<real_t> >* ns=0);                 //!< compute integral representation
  std::vector<SuBilinearForm> getSublfs(ComputationType, ValueType&) const; //!< returns single unknown bilinear forms
  void updateStorageType(const std::vector<SuBilinearForm>&,std::set<number_t>&,std::set<number_t>&,StorageType&) const; //!< update storage type
  void addDenseToStorage(const std::vector<SuBilinearForm>&,MatrixStorage*) const; //!< adds dense matrix to storage
  void toScalar(bool keepEntries=false);  //!< go to scalar representation
  void toVectorUnknown();                 //!< go to vector unknown representation if it has component unknown

  //scalar type moving
  SuTermMatrix& toReal();                 //!< go to real representation getting the real part when complex
  SuTermMatrix& toComplex();              //!< go to complex representation
  SuTermMatrix& toImag();                 //!< go to real representation getting the imag part when complex
  SuTermMatrix& toConj();                 //!< change to conj(U)
  SuTermMatrix& roundToZero(real_t aszero=10*theEpsilon);  //!< round to zero all coefficients close to 0 (|.| < aszero)

  //in/out tools
  void print(std::ostream&) const;        //!< print SuTermMatrix
  void print(PrintStream& os) const {print(os.currentStream());}
  void print(std::ostream&, bool) const;  //!< print SuTermMatrix with header option
  void print(PrintStream& os, bool h) const {print(os.currentStream(),h);}
  void printSummary(std::ostream&) const; //!< print SuTermMatrix in brief
  void printSummary(PrintStream& os) const {printSummary(os.currentStream());}
  void viewStorage(std::ostream&) const;  //!< print storage on stream
  void viewStorage(PrintStream& os) const {viewStorage(os.currentStream());}
  void saveToFile(const string_t&, StorageType, bool=false) const; //!< save TermMatrix to file (dense or Matlab format)

  //various algebraic operations
  template<typename T>
  SuTermMatrix& operator*=(const T&);                                      //!< SutermMatrix *= t
  template<typename T>
  SuTermMatrix& operator/=(const T&);                                      //!< SutermMatrix /= t
  SuTermMatrix& operator+=(const SuTermMatrix&);                           //!< SutermMatrix += SutermMatrix
  SuTermMatrix& operator-=(const SuTermMatrix&);                           //!< SutermMatrix -= SutermMatrix
  friend SuTermVector operator*(const SuTermMatrix&, const SuTermVector&);
  friend SuTermVector operator*(const SuTermVector&, const SuTermMatrix&);
  friend SuTermMatrix operator*(const SuTermMatrix&, const SuTermMatrix&);
  friend SuTermMatrix* mergeSuTermMatrix(const std::list<SuTermMatrix*>&);

  real_t norm2() const;                                                    //!< Frobenius norm of a SuTermMatrix
  real_t norminfty() const;                                                //!< infinite norm of a SuTermMatrix

  friend class TermMatrix;

  private:
  SuTermVector multMVScalarEntries(const SuTermVector& sutV, bool toVector=false) const; //!< product SuTermMatrix * SuTermVector when scalar_entries
  SuTermVector multMVVectorEntries(const SuTermVector& sutV, bool toScalar=false) const; //!< product SuTermMatrix * SuTermVector when entries
  SuTermVector multVMScalarEntries(const SuTermVector& sutV, bool toVector=false) const; //!< product SuTermVector * SuTermMatrix when scalar_entries
  SuTermVector multVMVectorEntries(const SuTermVector& sutV, bool toScalar=false) const; //!< product SuTermVector * SuTermMatrix when entries
};

SuTermVector operator*(const SuTermMatrix&, const SuTermVector&); //!< product SuTermMatrix * SuTermVector
SuTermVector operator*(const SuTermVector&, const SuTermMatrix&); //!< product SuTermVector * SuTermMatrix
SuTermMatrix operator*(const SuTermMatrix&, const SuTermMatrix&); //!< product SuTermMatrix * SuTermMatrix
SuTermMatrix* mergeSuTermMatrix(const std::list<SuTermMatrix*>&); //!< merge SuTerMatrix's referring to the same vector unknown

// ------------------------------------------------------------------------------------------------------------------------
// External functions  declaration
// ------------------------------------------------------------------------------------------------------------------------
void factorize(SuTermMatrix&, SuTermMatrix&, FactorizationType ft=_noFactorization,bool withPermutation=true); //!< factorize matrix as LU or LDLt or LDL*
void factorize(SuTermMatrix&, FactorizationType ft=_noFactorization,bool withPermutation=true);                //!< factorize matrix as LU or LDLt or LDL*
SuTermVector  factSolve(SuTermMatrix&, const SuTermVector&);                              //!< solve AX=B  when A is factorized
SuTermVectors factSolve(SuTermMatrix&, const std::vector<SuTermVector>&);                 //!< solve AXs=Bs  when A is factorized
SuTermMatrix  factSolve(SuTermMatrix&, SuTermMatrix&);                                    //!< create inv(A)*B when A is factorized
SuTermVector gaussSolve(SuTermMatrix&, const SuTermVector&, bool keepA=false);            //!< solve AX=B using Gauss reduction
SuTermVector umfpackSolve(SuTermMatrix&, const SuTermVector&, real_t&, bool keepA =false);//!< solve AX=B using umfpack if available
SuTermVector umfpackSolve(SuTermMatrix&, const SuTermVector&, bool keepA =false);         //!< solve AX=B using umfpack if available

// ------------------------------------------------------------------------------------------------------------------------
// template algebraic operation on SuTermMatrix
// ------------------------------------------------------------------------------------------------------------------------
template<typename T>
SuTermMatrix& SuTermMatrix::operator*=(const T& t)
{
  if(sublf_p!=0) *sublf_p*=t;
  if(computed()) *entries_p*=t;
  return *this;
}

template<typename T>
SuTermMatrix& SuTermMatrix::operator/=(const T& t)
{
  if(sublf_p!=0) *sublf_p/=t;
  if(computed()) *entries_p/=t;
  return *this;
}


/*! templated computation functions
  ComputationAlgorithm class is a trick to template FE/IE computation algorithm
*/
template <number_t CM>
class ComputationAlgorithm
{
  public:
    //! computation algorithm
    template <typename T, typename K>
    static void compute(const SuBilinearForm& sub, LargeMatrix<T>& mat, K& vt, Space * space_u_p,
              Space * space_v_p, const Unknown* u_p, const Unknown* v_p)
    { error("not_handled", "compute(const SuBilinearForm&, LargeMatrix<T>&, K&, ...)");}
};

template<number_t CM>
void SuTermMatrix::compute(const std::vector<SuBilinearForm>& sublfs, ValueType vt, StrucType str)
{
  real_t r = 0;
  complex_t c = 0.;
  std::vector<SuBilinearForm>::const_iterator itsublf;
  for (itsublf = sublfs.begin(); itsublf != sublfs.end(); itsublf++)
    {
      switch(str)
        {
        case _scalar :
          if(vt == _real) ComputationAlgorithm<CM>::compute(*itsublf, *(entries_p->rEntries_p), r, space_u_p, space_v_p, u_p, v_p);
          else ComputationAlgorithm<CM>::compute(*itsublf, *(entries_p->cEntries_p), c, space_u_p, space_v_p, u_p, v_p);
          break;
        case _matrix :
          if(vt == _real) ComputationAlgorithm<CM>::compute(*itsublf, *(entries_p->rmEntries_p), r, space_u_p, space_v_p, u_p, v_p);
          else ComputationAlgorithm<CM>::compute(*itsublf, *(entries_p->cmEntries_p), c, space_u_p, space_v_p, u_p, v_p);
          break;
        default :
          where("SuTerMatrix::compute(Vector<SuBilinearForm, ValueType, StrucType)");
          error("scalar_or_matrix");
        }
    }
}

//!compare SuTermMatrix regarding their numbers of non zero
inline bool compNnzSutermMatrix (SuTermMatrix* sut1, SuTermMatrix* sut2)
{return (sut1->nnz() > sut2->nnz());}

//!compare SuTermMatrix regarding row sizes
inline bool compRowSize (SuTermMatrix* sut1, SuTermMatrix* sut2)
{return (sut1->cdofsv().size() > sut2->cdofsv().size());}

//! compare SuTermMatrix regarding col sizes
inline bool compColSize (SuTermMatrix* sut1, SuTermMatrix* sut2)
{return (sut1->cdofsu().size() > sut2->cdofsu().size());}

/*! get internal matrix (restrictive), i.e LargeMatrix object if available
  if scalar and vector representations both exist, the scalar representation is prior except if StrucType argument is specified
  It returns a reference to the LargeMatrix
*/
template <typename T>
LargeMatrix<T>& SuTermMatrix::getLargeMatrix(StrucType str) const
{
  if (scalar_entries_p!=0 && str!=_matrix) return scalar_entries_p->getLargeMatrix<T>();
  if (entries_p!=0) return entries_p->getLargeMatrix<T>();
  where("SuTermMatrix::getLargeMatrix(StrucType)");
  error("term_no_entries");
  return *(new LargeMatrix<T>());  //fake return
}

/*! get HMatrix representation if it exists, it returns a reference to the HMatrix
  if scalar and vector representations both exist, the scalar representation is prior except if StrucType argument is specified
*/
template<typename T>
HMatrix<T,FeDof>& SuTermMatrix::getHMatrix(StrucType str) const
{
  if (hm_scalar_entries_p!=0 && str!=_matrix) return hm_scalar_entries_p->getHMatrix<T>();
  if (hm_entries_p!=0) return hm_entries_p->getHMatrix<T>();
  where("SuTermMatrix::getHMatrix(StrucType)");
  error("term_no_entries");
  return *(new HMatrix<T,FeDof>());  //fake return
}

} //end of namespace xlifepp

//include computation algorithms
#include "computation/FEMatrixComputation.hpp"
#include "computation/FEextMatrixComputation.hpp"
#include "computation/SPMatrixComputation.hpp"
#include "computation/FESPMatrixComputation.hpp"
#include "computation/IESPMatrixComputation.hpp"
#include "computation/IEMatrixComputation.hpp"
#include "computation/IEHMatrixComputation.hpp"

#endif /* SU_TERM_MATRIX_HPP */
