/*
XLiFE++ is an extended library of finite elements written in C++
    Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file TermMatrix.hpp
  \author E. Lunéville
  \since 23 mar 2012
  \date 5 oct 2012

  \brief Definition of the xlifepp::TermMatrix class

  xlifepp::TermMatrix class handles the numerical representation of general bilinear forms (xlifepp::BilinearForm)
  that is the matrix a(vi,wj), i=1,...m; j=1,..,n with (vi)and (wj) the basis functions.
  It inherits from xlifepp::Term class and it is an end user class.

  A xlifepp::BilinearForm object (see BilinearForm.hpp) is a multiple unknowns/spaces bilinear form
  that is a list of bilinear forms acting on a couple of (sub)spaces (xlifepp::SuBilinearForm object)
  indexed by unknown/testfunction . As a consquence, a xlifepp::TermMatrix object is a list of xlifepp::SuTermMatrix objects
  that are related to xlifepp::SuBilinearForm objects

  \verbatim
          child
  Term  --------> TermMatrix
                    | BilinearForm
                    | MatrixEntry*
                    | map<uvPair,SuTermMatrix>  (uvPair is a pair of unknown, testfunction)
                    | ...
  \endverbatim

  End user has to instanciate only xlifepp::TermMatrix object, xlifepp::SuTermMatrix is an internal object never managed by end user,
  even if its bilinear form is really a single unknown/testfunction bilinear form.

  xlifepp::MatrixEntry pointer of xlifepp::TermMatrix is in general null, may be not in case of multiple uvPair bilinear form.
  xlifepp::TermMatrix being seen as a block matrix, its matrix values are given by xlifepp::MatrixEntry of each xlifepp::SuTermMatrix.
  In some circonstances (direct solvers), xlifepp::MatrixEntry pointer of xlifepp::TermMatrix may be allocated to store xlifepp::TermMatrix
  as a contiguous matrix without block structure.
  Be care, both representations may exist at the same time but with two times memory consuming !!!

  Note : if required it is possible to go to scalar representation of matrix by using scalar_entries_p
         and renumbering vectors cdofs_u (column) and cdofs_v (row).
         This representation has no interest if all the unknowns involved in xlifepp::TemMatrix are scalar !
*/


#ifndef TERM_MATRIX_HPP
#define TERM_MATRIX_HPP

#include "config.h"
#include "Term.hpp"
#include "TermVector.hpp"
#include "SuTermMatrix.hpp"
#include "form.h"
#include "largeMatrix.h"
#include "solvers.h"
#include "computation/termUtils.hpp"
#include "EigenTerms.hpp"

namespace xlifepp
{

//@{
//! useful aliases for TermMatrix class
typedef std::map<uvPair, SuTermMatrix*>::iterator it_mustm;
typedef std::map<uvPair, SuTermMatrix*>::const_iterator cit_mustm;
//@}

extern TermMatrix theDefaultTermMatrix;
extern const TermVector theDefaultTermVector;
extern Preconditioner theDefaultPreconditioner;

/*!
   \class TermMatrix
   end user class handling numerical representation of any
   bilinear form (BilinearForm). May be multiple unknowns
*/
class TermMatrix : public Term
{
  protected:
    BilinearForm bilinForm_;                    //!< bilinear form
    std::map<uvPair, SuTermMatrix*> suTerms_;   //!< list of single unknowns term matrices
    MatrixEntry* entries_p;                     //!< pointer to values of TermMatrix (see explanations)

    //to deal with constraints
    SetOfConstraints* constraints_u_p;         //!< constraints data related to conditions to apply to unknowns  (col)
    SetOfConstraints* constraints_v_p;         //!< constraints data related to conditions to apply to test functions (row), in general equal to constraints_u_p
    MatrixEntry* rhs_matrix_p;                 //!< pointer to a correction matrix when constraints applied

    // for scalar representation
    MatrixEntry* scalar_entries_p;             //!< values of TermMatrix in scalar representation, same as entries_p if scalar unknowns
    std::vector<DofComponent> cdofs_c;         //!< col scalar numbering when scalar_entries_p is allocated
    std::vector<DofComponent> cdofs_r;         //!< row scalar numbering when scalar_entries_p is allocated

  public:
    TermMatrix(const string_t & na="?");                                           //!< default constructor

    //old constructors kept for ascending compatibility
    TermMatrix(const BilinearForm &, const string_t& na, bool noass);   //!< constructor with no essential boundary conditions
    TermMatrix(const BilinearForm &,  const EssentialConditions&,
               const string_t& na, bool noass);                        //!< constructor with blf and essential conditions on u
    TermMatrix(const BilinearForm &, const EssentialConditions&, const EssentialConditions&,
               const string_t& na, bool noass );                       //!< constructor with blf and essential conditions on u and v (if different)
    TermMatrix(const BilinearForm &, const SetOfConstraints&,
               const string_t& na, bool noass);                        //!< constructor with blf and explicit constraints data on u
    TermMatrix(const BilinearForm &, const SetOfConstraints&, const SetOfConstraints&,
               const string_t& na, bool noass);                        //!< constructor with blf and explicit constraints data on u and v (if different)

    //constructors with options and no essential boundary conditions
    TermMatrix(const BilinearForm&, const string_t& na="");
    TermMatrix(const BilinearForm&, TermOption, const string_t& na="");
    TermMatrix(const BilinearForm&, TermOption, TermOption, const string_t& na="");
    TermMatrix(const BilinearForm&, TermOption, TermOption, TermOption, const string_t& na="");

    //constructor with options and one essential boundary condition (same on u and v)
    TermMatrix(const BilinearForm&, const EssentialConditions&, const string_t& na="");
    TermMatrix(const BilinearForm&, const EssentialConditions&, TermOption, const string_t& na="");
    TermMatrix(const BilinearForm&, const EssentialConditions&, TermOption, TermOption, const string_t& na="");
    TermMatrix(const BilinearForm&, const EssentialConditions&, TermOption, TermOption, TermOption, const string_t& na="");

    //constructor with one essential boundary condition (same on u and v) and explicit reduction method
    TermMatrix(const BilinearForm&, const EssentialConditions&, const ReductionMethod&, const string_t& na="");
    TermMatrix(const BilinearForm&, const EssentialConditions&, const ReductionMethod&, TermOption, const string_t& na="");
    TermMatrix(const BilinearForm&, const EssentialConditions&, const ReductionMethod&, TermOption, TermOption, const string_t& na="");
    TermMatrix(const BilinearForm&, const EssentialConditions&, const ReductionMethod&, TermOption, TermOption, TermOption, const string_t& na="");

    //constructor with options and two essential boundary conditions (resp. on u and v)
    TermMatrix(const BilinearForm&, const EssentialConditions&, const EssentialConditions&, const string_t& na="");
    TermMatrix(const BilinearForm&, const EssentialConditions&, const EssentialConditions&, TermOption, const string_t& na="");
    TermMatrix(const BilinearForm&, const EssentialConditions&, const EssentialConditions&, TermOption, TermOption, const string_t& na="");
    TermMatrix(const BilinearForm&, const EssentialConditions&, const EssentialConditions&, TermOption, TermOption, TermOption, const string_t& na="");

    //constructor with options and two essential boundary conditions (resp. on u and v) and explicit reduction method
    TermMatrix(const BilinearForm&, const EssentialConditions&, const EssentialConditions&, const ReductionMethod&, const string_t& na="");
    TermMatrix(const BilinearForm&, const EssentialConditions&, const EssentialConditions&, const ReductionMethod&, TermOption, const string_t& na="");
    TermMatrix(const BilinearForm&, const EssentialConditions&, const EssentialConditions&, const ReductionMethod&, TermOption, TermOption, const string_t& na="");
    TermMatrix(const BilinearForm&, const EssentialConditions&, const EssentialConditions&, const ReductionMethod&, TermOption, TermOption, TermOption, const string_t& na="");

    void build(const BilinearForm&, const EssentialConditions*, const EssentialConditions*, const ReductionMethod&,
               const std::vector<TermOption>&, const string_t&);            //!< effective constructor from bilinear form, essential conditions  and options

    //special constructors
    TermMatrix(const TermMatrix &, const string_t& na="");                  //!< copy constructor (full copy)
    explicit TermMatrix(const TermVector &, const string_t& na="");         //!< constructor of a diagonal TermMatrix from a TermVector
    TermMatrix(const TermMatrix&, SpecialMatrix, const string_t& na="");    //!< construct special matrix from a TermMatrix (e.g _IdMatrix)
    TermMatrix(const SuTermMatrix &, const string_t& na="");                //!< constructor of a TermMatrix from a SuTermMatrix   (full copy)
    TermMatrix(SuTermMatrix *, const string_t& na="");                      //!< constructor of a TermMatrix from a SuTermMatrix * (pointer copy)
    template<typename T>
    TermMatrix(const Unknown&, const GeomDomain&, const Vector<T> &, const string_t& na=""); //!< special constructor of a diagonal TermMatrix from a vector
    TermMatrix(const Unknown&, const GeomDomain&, const Unknown&, const GeomDomain&,
               const OperatorOnKernel&, const string_t& na="");             //!< constructor of matrix opker(xi,yj) (Lagrange interpolation)
    TermMatrix(const LcTerm<TermMatrix>&,const string_t& na="");            //!<constructor from a linear combination of TermMatrix's

    //destructor, assign and construction tools
    virtual ~TermMatrix();                                                //!< destructor
    TermMatrix& operator=(const LcTerm<TermMatrix>&);                     //!< assign operator from a linear combination of TermMatrix's
    TermMatrix& operator=(const TermMatrix&);                             //!< assign operator (full copy)
    void initFromBlf(const BilinearForm &, const string_t& na ="", bool noass = false);  //!< initializer with blf
    void copy(const TermMatrix&);                                         //!< copy a TermMatrix in current TermMatrix
    void clear();                                                         //!< deallocate memory used by a TermMatrix
    void insert(const SuTermMatrix &);                                    //!< insert a SuTermMatrix  with full copy
    void insert(SuTermMatrix *);                                          //!< insert a SuTermMatrix with pointer copy
    void initTermVector(TermVector&, ValueType, bool col=true) const;    //!< initialize a TermVector consistent with matrix rows or columns
    template<typename T>
    void initTermVectorT(TermVector&,const T&, bool col=true) const;     //!< initialize a TermVector from value T, consistent with matrix rows or columns

    //accessors and properties
    string_t name() const { return Term::name(); }
    void name(const string_t& nam);                                       //!< update the name of a TermMatrix and its SuTermMatrixs
    void setStorage(StorageType, AccessType);                             //!< specify the storage of SuTermMatrix's
    StorageType storageType()                                             //! returns storage type (cs, skyline, dense)
    {return computingInfo_.storageType;}
    AccessType storageAccess()                                            //! returns storage access (row, col, sym, dual)
    {return computingInfo_.storageAccess;}
    bool isSingleUnknown() const                                          //!  true if a single unknown matrix
    {return suTerms_.size()==1;}
    std::set<const Unknown*> rowUnknowns() const;                         //!< return set of row Unknowns (say v unknowns)
    std::set<const Unknown*> colUnknowns() const;                         //!< return set of col Unknowns (say u unknowns)
    std::set<const Space*> unknownSpaces() const;                         //!< list of involved unknown spaces
    ValueType valueType() const;                                          //!< return value type (_real,_complex)
    bool isScalar() const;                                                //!< true if all SuTermMatrix's are scalar
    bool hasConstraints() const                                           //! true if one constraints pointer is not null
    {return (constraints_u_p!=0 || constraints_v_p!=0);}
    number_t numberOfRows() const;                                        //!< number of rows (counted in scalar dof)
    number_t numberOfCols() const;                                        //!< number of columns (counted in scalar dof)
    number_t nnz() const;                                                 //!< number of non zero coefficients (counted in scalar)
    void markAsComputed(bool = true);                                     //!< TermMatrix and SutermMatrix's computed state = true or false

    TermMatrix operator()(const Unknown&,const Unknown&) const;           //!< access to single unknowns block matrix as TermMatrix
    SuTermMatrix& subMatrix(const Unknown*,const Unknown*);               //!< access to single unknowns term matrix
    const SuTermMatrix& subMatrix(const Unknown*,const Unknown*) const;   //!< access to single unknowns term matrix (const)
    SuTermMatrix* subMatrix_p(const Unknown*,const Unknown*);             //!< access to single unknowns term matrix pointer
    const SuTermMatrix* subMatrix_p(const Unknown*,const Unknown*) const; //!< access to single unknowns term matrix pointer (const)
    TermVector& getRowCol(number_t, AccessType, TermVector&) const;       //!< return the i-th row or col as a TermVector (i=1,...)
    TermVector row(number_t) const;                                       //!< return the i-th row as a TermVector (i=1,...)
    TermVector column(number_t) const;                                    //!< return the j-th col as a TermVector (j=1,...)
    template <typename T>
    LargeMatrix<T>& getLargeMatrix(StrucType =_undefStrucType) const;     //!< get internal matrix (restrictive), advanced usage
    template <typename T>
    HMatrix<T,FeDof>& getHMatrix(StrucType =_undefStrucType) const;       //!< get internal HMatrix if exists, advanced usage

    cit_mustm begin() const                                               //! iterator to the first element of the SuTermMatrix map (const)
    {return suTerms_.begin();}
    it_mustm begin()                                                      //! iterator to the first element of the SuTermMatrix map
    {return suTerms_.begin();}
    cit_mustm end() const                                                 //! iterator to the last element of the SuTermMatrix map (const)
    {return suTerms_.end();}
    it_mustm end()                                                        //! iterator to the last element of the SuTermMatrix map
    {return suTerms_.end();}
    SuTermMatrix* firstSut() const                         //! first SuTermMatrix as pointer, 0 if void TermMatrix (const)
    {
      if(suTerms_.size()>0) return suTerms_.begin()->second;
      else return 0;
    }
    SuTermMatrix* firstSut()                               //! first SuTermMatrix as pointer, 0 if void TermMatrix
    {
      if(suTerms_.size()>0) return suTerms_.begin()->second;
      else return 0;
    }
    std::map<uvPair, SuTermMatrix*>& suTerms()              //! access to suterms
    {return suTerms_;}
    number_t nbTerms() const                                //! number of suTerms
    {return suTerms_.size();}
    MatrixEntry*& entries()                                 //! access to entries pointer (r/w)
    {return entries_p;}
    const MatrixEntry* entries() const                      //! access to entries pointer (r)
    {return entries_p;}
    MatrixEntry*& scalar_entries()                          //! access to scalar entries pointer (r/w)
    {return scalar_entries_p;}
    const MatrixEntry* scalar_entries() const               //! access to scalar entries pointer (r)
    {return scalar_entries_p;}
    MatrixEntry* actual_entries() const;                    //!< return the actual pointer to entries (priority to scalar entry)
    MatrixEntry*& rhs_matrix()                              //! access to rhs matrix pointer (r/w)
    {return rhs_matrix_p;}
    const MatrixEntry* rhs_matrix()   const                 //! access to rhs matrix pointer (r)
    {return rhs_matrix_p;}
    const std::vector<DofComponent>& cdofsr() const         //! access to cdofs_r vector (r)
    {return cdofs_r;}
    const std::vector<DofComponent>& cdofsc() const         //! access to cdofs_c vector (r)
    {return cdofs_c;}
    std::vector<DofComponent>& cdofsr()                     //! access to cdofs_r vector (r/w)
    {return cdofs_r;}
    std::vector<DofComponent>& cdofsc()                     //! access to cdofs_c vector (r/w)
    {return cdofs_c;}

    SymType symmetry() const;                               //!< return global symmetry property
    FactorizationType factorization() const;                //!< return factorization type if factorized

    //computation functions and related tools
    void compute();                                                       //!< compute TermMatrix
    void compute(const LcTerm<TermMatrix>&,const string_t& na="");        //!< compute TermMatrix from a linear combination of TermMatrix's

    //storage management tools
    std::pair<StorageType,AccessType> findGlobalStorageType() const;      //!< suggest storage type for global representation
    void toScalar(bool keepEntries=false);                                //!< create scalar representation of TermMatrix
    void toGlobal(StorageType, AccessType, SymType=_noSymmetry,
                  bool keepSuTerms=false);                                //!< create global scalar representation of TermMatrix
    void toSkyline();                                                     //!< convert to skyline storage (works only for cs storage single unknown matrix)
    void mergeNumbering(std::map<const Unknown*, std::list<SuTermMatrix* > >&, std::map<SuTermMatrix*, std::vector<number_t > >&,
                        std::vector<DofComponent>&, AccessType);          //!< merging tool used by toGlobal()
    void mergeNumberingFast(std::map<const Unknown*, std::list<SuTermMatrix*> >&, std::map<SuTermMatrix*, std::vector<number_t > >&,
                            std::vector<DofComponent>&, AccessType);      //!< merging tool used by toGlobal(), fast version

    void addIndices(std::vector<std::set<number_t> >&, MatrixStorage*,
                    const std::vector<number_t>&, const std::vector<number_t>&); //!< merging tool used by toGlobal()
    void mergeBlocks();                                                   //!< merge blocks referring to same vector unknowns
    MatrixEntry* matrixData();                                            //!< turns to scalar representation and returns pointer to data

    //essential conditions tools
    void pseudoReduction();                           //!< reduction of essential condition using pseudo reduction method

    //scalar type moving
    TermMatrix& toReal();                             //!< go to real representation getting the real part when complex
    TermMatrix& toComplex();                          //!< go to complex representation
    TermMatrix& toImag();                             //!< go to real representation getting the imag part when complex
    TermMatrix& toConj();                             //!< change to conj(U)
    TermMatrix& roundToZero(real_t aszero=10*theEpsilon);  //!< round to zero all coefficients close to 0 (|.| < aszero)
    friend TermMatrix real(const TermMatrix&);
    friend TermMatrix imag(const TermMatrix&);
    friend TermMatrix toComplex(const TermMatrix&);
    friend TermMatrix roundToZero(const TermMatrix&, real_t aszero);

    real_t norm2() const;                                                 //!< Frobenius norm : sqrt(sum|a_ij|^2)
    real_t norminfty() const;                                             //!< infinite norm : sup|a_ij|

    //in/out utilities
    void print(std::ostream&) const;                                      //!< print TermMatrix
    void print(PrintStream& os) const {print(os.currentStream());}
    void printSummary(std::ostream&) const;                               //!< print summary TermMatrix
    void printSummary(PrintStream& os) const {printSummary(os.currentStream());}
    void viewStorage(std::ostream&) const;                                //!< print storage on stream
    void viewStorage(PrintStream& os) const {viewStorage(os.currentStream());}
    void saveToFile(const string_t&, StorageType, bool=false) const;      //!< save TermMatrix to file (dense or Matlab format)

    //various algebraic operation
    template<typename T> TermMatrix& operator*=(const T&);                //!< TermMatrix *= t
    template<typename T> TermMatrix& operator/=(const T&);                //!< TermMatrix /= t
    TermMatrix& operator+=(const TermMatrix&);                            //!< TermMatrix += TermMatrix
    TermMatrix& operator-=(const TermMatrix&);                            //!< TermMatrix -= TermMatrix
    friend TermVector& multMatrixVector(const TermMatrix&,const TermVector&, TermVector&);
    friend TermVector operator*(const TermMatrix&,const TermVector&);
    friend TermVector& multVectorMatrix(const TermVector&, const TermMatrix&, TermVector&);
    friend TermVector& multVectorMatrix(const TermMatrix&, const TermVector&, TermVector&);
    friend TermVector operator*(const TermVector&, const TermMatrix&);
    friend TermMatrix operator*(const TermMatrix&, const TermMatrix&);

    //solvers
    friend TermVector prepareLinearSystem(TermMatrix&, TermVector&, MatrixEntry*&, VectorEntry*&,
                                          StorageType, AccessType, bool);
    friend void updateRhs(TermMatrix&, TermVector&);
    void ldltSolve(const TermVector&, TermVector&);   //!< solve linear system using LDLt factorization
    void ldlstarSolve(const TermVector&,TermVector&); //!< solve linear system using LDL* factorization
    void luSolve(const TermVector&,TermVector&);      //!< solve linear system using LU factorization
    void umfpackSolve(const TermVector&,TermVector&); //!< solve linear system using umfpack
    void sorSolve(const TermVector& V, TermVector& R, const real_t w, SorSolverType sType ); //!< solve linear system using SOR
    void sorUpperSolve(const TermVector& V, TermVector& R, const real_t w );           //!< solve linear system using SOR for upper triangular matrices
    void sorDiagonalMatrixVector(const TermVector& V, TermVector& R, const real_t w ); //!< solve linear system using SOR for block diagonal matrices
    void sorLowerSolve(const TermVector& V, TermVector& R, const real_t w );           //!< solve linear system using SOR for lower triangular matrices
    void sorDiagonalSolve(const TermVector& V, TermVector& R, const real_t w );        //!< solve linear system using SOR for diagonal matrices
};

TermMatrix real(const TermMatrix&);        //!< return real part as a real TermMatrix
TermMatrix imag(const TermMatrix&);        //!< return imag part as a real TermMatrix
TermMatrix toComplex(const TermMatrix&);   //!< return the complex representation of the given TermMatrix
TermMatrix roundToZero(const TermMatrix&, real_t aszero);   //!< return the 0 rounded TermMatrix

TermVector& multMatrixVector(const TermMatrix&,const TermVector&, TermVector&);   //!< product TermMatrix * TermVector
TermVector operator*(const TermMatrix&,const TermVector&);     //!< product TermMatrix * TermVector
TermVector& multVectorMatrix(const TermVector&, const TermMatrix&, TermVector&);  //!< product TermVector * TermMatrix
TermVector& multVectorMatrix(const TermMatrix&, const TermVector&, TermVector&);  //!< product TermVector * TermMatrix
TermVector operator*(const TermVector&, const TermMatrix&);    //!< product TermVector * TermMatrix
TermMatrix operator*(const TermMatrix&, const TermMatrix&);    //!< product of TermMatrix

//! prepare linear system AX=B (internal tool)
TermVector prepareLinearSystem(TermMatrix&, TermVector&, MatrixEntry*&, VectorEntry*&, StorageType, AccessType, bool);
void updateRhs(TermMatrix&, TermVector&); //!< prepare rhs B of linear system AX=B (internal tool)

// ------------------------------------------------------------------------------------------------------------------------
// External functions declaration
// ------------------------------------------------------------------------------------------------------------------------
#include "decLinSys.hpp"
#include "decEigen.hpp"

inline real_t norm2(const TermMatrix& A)
{return A.norm2();}

inline real_t norminfty(const TermMatrix& A)
{return A.norminfty();}

//in/out utils
inline void saveToFile(const TermMatrix& A, const string_t& fn, StorageType st, bool enc = false)     //! save TermMatrix to file (dense or Matlab format)
{A.saveToFile(fn,st,enc);}

//integral representation
TermMatrix integralRepresentation(const Unknown&, const GeomDomain&, const LinearForm&, string_t nam="IR"); //!< integral representation
TermMatrix integralRepresentation(const GeomDomain&, const LinearForm&, string_t nam="IR");         //!< integral representation, no unknown
TermMatrix integralRepresentation(const std::vector<Point>&, const LinearForm&, string_t nam="IR"); //!< integral representation, no unknown, no domain

//----------------------------------------------------------------------------------------------------------
// template functions
//----------------------------------------------------------------------------------------------------------
//! special constructor of a diagonal TermMatrix from a vector
template<typename T>
TermMatrix::TermMatrix(const Unknown& u, const GeomDomain& dom, const Vector<T> & v, const string_t& na)
{
  trace_p->push("TermMatrix::TermMatrix(Vector<T>)");
  entries_p = 0;
  scalar_entries_p=0;
  rhs_matrix_p=0;
  constraints_u_p = 0;
  constraints_v_p = 0;
  termType_ = _termMatrix;
  computingInfo_.noAssembly = false;
  name_ = na;
  TermVector tv(u,dom,v);
  for(it_mustv it = tv.begin(); it != tv.end(); it++)
    suTerms_[uvPair(&u,&u)]=new SuTermMatrix(*it->second, na+"_"+u.name());
  computingInfo_.isComputed = true;
  trace_p->pop();
}

/*
  multiple algebraic operations on TermMatrix
  produce a LcTerm object (linear combination of terms)
  the computation is done by constructor from LcTerm or assign operation of a LcTerm
*/
const TermMatrix& operator+(const TermMatrix&);         //!< unary operator+
LcTerm<TermMatrix> operator-(const TermMatrix&);                                //!< unary operator- (returns a LcTerm)
LcTerm<TermMatrix> operator+(const TermMatrix&, const TermMatrix&);             //!< addition of TermMatrix
LcTerm<TermMatrix> operator-(const TermMatrix&, const TermMatrix&);             //!< substraction of TermMatrix of a TermMatrix
LcTerm<TermMatrix> operator+(const LcTerm<TermMatrix>&, const TermMatrix&);     //!< addition of a TermMatrix to a LcTerm
LcTerm<TermMatrix> operator-(const LcTerm<TermMatrix>&, const TermMatrix&);     //!< substraction of a TermMatrix of a LcTerm
LcTerm<TermMatrix> operator+(const TermMatrix&, const LcTerm<TermMatrix>&);     //!< addition of a TermMatrix to a LcTerm
LcTerm<TermMatrix> operator-(const TermMatrix&, const LcTerm<TermMatrix>&);     //!< substraction of a LcTerm from a TermMatrix
TermVector operator*(const LcTerm<TermMatrix>&, const TermVector&);             //!< LcTerm * TermVector

//! product of TermMatrix by a real or a complex (template T)
template<typename T>
LcTerm<TermMatrix> operator*(const TermMatrix& tv, const T& t)
{
  return LcTerm<TermMatrix>(&tv,t);
}
//! product of TermMatrix by a real or a complex (template T)
template<typename T>
LcTerm<TermMatrix> operator*(const T& t, const TermMatrix& tv)
{
  return LcTerm<TermMatrix>(&tv,t);
}
//! division of TermMatrix by a real or a complex (template T)
template<typename T>
LcTerm<TermMatrix> operator/(const TermMatrix& tv, const T& t)
{
  return LcTerm<TermMatrix>(&tv,1./t);
}

//! product assignment by a real or a complex (template T)
template<typename T>
TermMatrix& TermMatrix::operator*=(const T& t)
{
  for(it_mustm it = suTerms_.begin(); it != suTerms_.end(); it++) *it->second *= t;
  bilinForm_*=t;
  return *this;
}

//! division assignment by a real or a complex (template T)
template<typename T>
TermMatrix& TermMatrix::operator/=(const T& t)
{
  if(t == 0) error("divBy0");
  for(it_mustm it = suTerms_.begin(); it != suTerms_.end(); it++) *it->second /= t;
  bilinForm_/=t;
  return *this;
}

//! product of TermMatrix by a real or a complex (template T)
template<typename T>
TermMatrix operator*(const TermMatrix & tM, const T& t)
{
  TermMatrix R(tM);
  return R*=t;
}

//! initialize a TermVector consistent with matrix column unknown (col=true), with matrix row unknown (col=false)
template<typename T>
void TermMatrix::initTermVectorT(TermVector& tv, const T& v, bool col) const
{
  string_t na = tv.name(); //preserving only name
  tv.clear();
  tv.name() = na;
  //make list of col unknowns
  std::map<const Unknown*, SuTermMatrix*> terms;
  for(cit_mustm itA = begin(); itA != end(); itA++)
  {
    const Unknown* u;
    if(col) u = itA->first.first;
    else    u = itA->first.second;
    SuTermMatrix* Auv = itA->second;
    if(terms.find(u) == terms.end()) terms[u] = Auv;
  }
  std::map<const Unknown*, SuTermMatrix*>::iterator it;
  for(it = terms.begin(); it != terms.end(); it++)
  {
    const Unknown* u = it->first;
    SuTermMatrix* sut = it->second;
    Space* sp = sut->space_up();
    if(!col) sp = sut->space_vp();
    SuTermVector* sutX = new SuTermVector("", u, sp, v);
    tv.insert(u, sutX);
  }
}

/*! get internal matrix (restrictive), i.e LargeMatrix object if available
    available only for single unknowns TermMatrix and multiple unknowns TermMatrix having a global representation
    if scalar and vector representations both exist, the scalar representation is prior except if StrucType argument is specified
    user has to specify the LargeMatrix type :
        TermMatrix M(intg(omega,u*v));
        LargeMatrix& Alm=M.getMatrix<Real>();
    It returns a reference to the LargeMatrix
    Note : designed for advanced usage
*/
template <typename T>
LargeMatrix<T>& TermMatrix::getLargeMatrix(StrucType str) const
{
  number_t nbsut=suTerms_.size();
  if (nbsut==1) return suTerms_.begin()->second->getLargeMatrix<T>(str);
  if (scalar_entries_p!=0 && str!=_matrix) return scalar_entries_p->getLargeMatrix<T>();
  if (entries_p!=0) return entries_p->getLargeMatrix<T>();
  where("TermMatrix::getLargeMatrix(StrucType)");
  error("term_no_entries");
  return *(new LargeMatrix<T>());  //fake return
}

/*! get internal H-matrix (restrictive), i.e HMatrix object if available
    available only for single unknown TermMatrix
    if scalar and vector representations both exist, the scalar representation is prior except if StrucType argument is specified
    It returns a reference to a HMatrix
    Note : designed for advanced usage
*/
template <typename T>
HMatrix<T,FeDof>& TermMatrix::getHMatrix(StrucType str) const
{
  number_t nbsut=suTerms_.size();
  if (nbsut==1) return suTerms_.begin()->second->getHMatrix<T>(str);
  where("TermMatrix::getHMatrix(StrucType)");
  error("term_no_entries");
  return *(new HMatrix<T,FeDof>());  //fake return
}

/* ================================================================================================
                       computation algorithm for bilinear form as linear form
                        (placed here because it uses TermMatrix's machinery)
   ================================================================================================ */
/*!
  Computation of a linear form defined from a bilinear form and a TermVector to apply to
                   l(v) = a(U,v) or l(u)=a(u,V)   where U,V are some term vectors
  \param lf   : a pair of BasicLinearForm and coefficient
  \param val  : vector of values to fill in
  \param vt   : type of value type

  example :  intg_gamma grad(U)|grad(v) dy  U a termvector
             int_sigma intg_gamma (U*G)*v   U a termvector

  The algorithm is the most lazy one :
      - go to the TermMatrix machinery to compute matrix A from a(u,v)
      - do the product A*U or V*A
      - add the product result to vector val
*/

template<typename T, typename K>
void computeBilAsLin(const std::pair<BasicLinearForm*, complex_t>& lf, Vector<T>& val, K& vt)
{
  if(lf.first->type()!=_bilinearAsLinear)
  {
    where("SuTermVector::computeBilAsLin(...)");
    error("lform_not_handled", words("form type", lf.first->type()));
  }
  BilinearFormAsLinearForm* blfaslf = lf.first->asBilinearAsLinearForm();
  K coef = complexToT<K>(lf.second);
  //create TermMatrix
  SuBilinearForm sublf(std::vector<blfPair>(1, blfPair(blfaslf->blf(), lf.second)));
  BilinearForm blf(sublf);
  TermMatrix S(blf);
  //do product
  TermVector R;
  if(blfaslf->asUnknown_)  R=S* *blfaslf->termVector();
  else                     R=*blfaslf->termVector()*S;
  //add R to vector val
  Vector<T> r;
  R.asVector(r);
  typename  Vector<T>::iterator itv=val.begin(), itr=r.begin();
  for(;itr!=r.end(); ++itr, ++itv) *itv+=*itr;
}

} //end of namespace xlifepp

#endif /* TERM_MATRIX_HPP */
