/*
 XLiFE++ is an extended library of finite elements written in C++
 Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file DuffyIM.hpp
  \authors T.Khaled and N.Salles, E. Lunéville.
  \since 25 aug 2016
  \date 25 aug 2016

  \brief Implementation of a method based on the Duffy transform to compute the FE TermMatrix
             intg_Eu intg Ev opu(y) opk(x,y) opv(x) dy dx
   with Eu and Ev 2 segments.
 */

#ifndef DUFFY_IM_HPP
#define DUFFY_IM_HPP

#include "finiteElements.h"
#include "utils.h"
#include "config.h"
#include "termUtils.hpp"

namespace xlifepp{


/*!
 \class DuffyIM
 integral over a product of 2D geometric elements with singularity using method based on Duffy transform (for elements that shared vertices)
 may manage a quadrature rule for self influence and a a quadrature rule elements that share only one vertex
 Elements must be segment.
*/
class DuffyIM : public DoubleIM
{
  public:
    Quadrature* quadSelf;    //!< quadrature on segment [0,1] for self influence
    Quadrature* quadAdjt;    //!< quadrature on segment [0,1] for adjacent elements
    number_t ordSelf;        //!< order of quadrature on the segment for self influence
    number_t ordAdjt;        //!< order of quadrature on the segment for adjacent elements

    DuffyIM(IntegrationMethodType imt)  //! basic constructor
      : DoubleIM(imt), quadSelf(0), quadAdjt(0), ordSelf(0), ordAdjt(0){name="Duffy_?";}

    DuffyIM(number_t o=6)              //! full constructor from quadrature order
    {ordSelf=o; quadSelf=findQuadrature(_segment,_GaussLegendreRule,ordSelf);
     ordAdjt=ordSelf; quadAdjt= quadSelf;
     name="Duffy_"+tostring(ordSelf);
     imType=_DuffyIM; singularType=_logr; singularOrder=1;}

    DuffyIM(const Quadrature& q)       //! full constructor from a quadrature object
    {
      if (q.shapeType()!=_segment)
      {
        where("DuffyIM::DuffyIM(Quadrature)");
        error("shape_not_handled", words("shape", q.shapeType()));
      }
      quadSelf = const_cast<Quadrature*>(&q); quadAdjt = quadSelf;
      ordSelf = q.degree; ordAdjt = ordSelf;
      name="Duffy_"+tostring(ordSelf);
      imType=_DuffyIM; singularType=_logr; singularOrder=1;
    }

    virtual DuffyIM* clone() const             //covariant
    {return new DuffyIM(*this);}

    virtual std::list<Quadrature*> quadratures() const            //! return the list of (single) quadratures in a list
    {std::list<Quadrature *> lsq;
     if(quadSelf!=0) lsq.push_back(quadSelf);
     if(quadAdjt!=0 && quadAdjt!=quadSelf)  lsq.push_back(quadAdjt);
     return lsq;}

    virtual void print(std::ostream& os) const    //! print IntegrationMethod on stream
    {
      os<<"Duffy integration method with "<< quadSelf->name<<" degree " << quadSelf->degree
        << " ( " << quadSelf->quadratureRule.size() << " points ) " <<eol;
    }
    virtual void print(PrintStream& os) const {print(os.currentStream());}

    //template computeIE function, explicit instanciation because template functions are not inherited
    template<typename K>
    void computeIE(const Element* elt_S, const Element* elt_T, AdjacenceInfo& adj,
                   const KernelOperatorOnUnknowns& kuv, Matrix<K>& res,IEcomputationParameters& ieparams,
                   Vector<K>& val_opu, Vector<K>& val_opv, Vector<K>& val_opk) const;
    void computeIE(const Element* elt_S, const Element* elt_T, AdjacenceInfo& adj,
                   const KernelOperatorOnUnknowns& kuv, Matrix<real_t>& res,IEcomputationParameters& ieparams,
                   Vector<real_t>& val_opu, Vector<real_t>& val_opv, Vector<real_t>& val_opk) const
    {computeIE<>(elt_S, elt_T,adj, kuv, res,ieparams, val_opu, val_opv, val_opk);}
    void computeIE(const Element* elt_S, const Element* elt_T, AdjacenceInfo& adj,
                   const KernelOperatorOnUnknowns& kuv, Matrix<complex_t>& res,IEcomputationParameters& ieparams,
                   Vector<complex_t>& val_opu, Vector<complex_t>& val_opv, Vector<complex_t>& val_opk) const
    {computeIE<>(elt_S, elt_T,adj, kuv, res,ieparams, val_opu, val_opv, val_opk);}

    template<typename K>
    void SelfInfluences(const Element* elt_S, const Element* elt_T, const KernelOperatorOnUnknowns& kuv,
                        const Vector<real_t>* nx, const Vector<real_t>* ny, Matrix<K>& res,
                        IEcomputationParameters& ieparams) const;
    template<typename K>
    void AdjacentSegments(const Element* elt_S, const Element* elt_T, const KernelOperatorOnUnknowns& kuv,
                          const Vector<real_t>* nx, const Vector<real_t>* ny, const Vector<number_t>& indi,
                          const Vector<number_t>& indj, Matrix<K>& res, IEcomputationParameters& ieparams) const;
    template<typename K>
    void k4(real_t x1,real_t x2, const KernelOperatorOnUnknowns& kuv, const Point& S1, const Point & M,
            const Element* elt_S,const Element* elt_T, Matrix<K>& res, const Vector<real_t>* nx, const Vector<real_t>* ny,
            IEcomputationParameters& ieparams) const;
    template<typename K>
    void k5(real_t x1,real_t x2, const KernelOperatorOnUnknowns& kuv, const Point& S1, const Point & MS,
            const Point & T1, const Point & MT, const Element* elt_S,const Element* elt_T, Matrix<K>& res,
            const Vector<real_t>* nx, const Vector<real_t>* ny, IEcomputationParameters& ieparams) const;

};

template<typename K>
void DuffyIM::computeIE(const Element* elt_S, const Element* elt_T, AdjacenceInfo& adj,
                           const KernelOperatorOnUnknowns& kuv, Matrix<K>& res,IEcomputationParameters& ieparams,
                           Vector<K>& val_opu, Vector<K>& val_opv, Vector<K>& val_opk) const
{
  //load normals
  Vector<real_t>* nx=0, *ny=0;
  MeshElement* melt_S=ieparams.melt_u, *melt_T=ieparams.melt_v;
  if(kuv.xnormalRequired()) nx = &melt_T->geomMapData_p->normalVector;
  if(kuv.ynormalRequired()) ny = &melt_S->geomMapData_p->normalVector;

  if(adj.status==_adjacentByElement)  //same geometrical triangle
    {
      SelfInfluences(elt_S, elt_T, kuv, nx, ny, res, ieparams);
      return;
    }
  //ajacent by vertex
  std::vector<number_t>& indi=adj.sharedVertex1, &indj=adj.sharedVertex1;
  if(indi[0]==1) indi[1]=2; else indi[1]=1;
  if(indj[0]==1) indj[1]=2; else indj[1]=1;
  AdjacentSegments(elt_S, elt_T, kuv, nx, ny, indi, indj, res, ieparams);
}

template<typename K>
void DuffyIM::SelfInfluences(const Element* elt_S, const Element* elt_T, const KernelOperatorOnUnknowns& kuv,
                                const Vector<real_t>* nx, const Vector<real_t>* ny,
                                Matrix<K>& res, IEcomputationParameters& ieparams) const
{
  MeshElement* meltu=ieparams.melt_u;
  MeshElement* meltv=ieparams.melt_v;
  Point& S1=*meltu->nodes[0], & S2= *meltu->nodes[1];

  //definition of matrix MS
  Point  M=S2-S1;
  real_t G12=meltu->geomMapData_p->differentialElement; G12*=meltv->geomMapData_p->differentialElement;

  number_t N=quadSelf->numberOfPoints();
  real_t a, b;
  const std::vector< real_t >& coords=quadSelf->coords();
  const std::vector< real_t >& weights=quadSelf->weights();
  Matrix<K> matel=0.*res;

  for(number_t i1=0; i1<N; i1++)
    {
      a=coords[i1];
      real_t w1=weights[i1]*G12;
      for(number_t i2=0; i2<N; i2++)
        {
          real_t w2=w1*weights[i2];
          b=coords[i2];
          matel*=0.;
          k4(a,b,kuv,S1,M,elt_S,elt_T,matel,nx,ny,ieparams);
          res+=w2*matel;
        }
    }
}

template<typename K>
void DuffyIM::AdjacentSegments(const Element* elt_S, const Element* elt_T, const KernelOperatorOnUnknowns& kuv,
                                  const Vector<real_t>* nx, const Vector<real_t>* ny, const Vector< number_t >& indi, const Vector< number_t >& indj,
                                  Matrix<K>& res, IEcomputationParameters& ieparams) const
{
  MeshElement* meltu=ieparams.melt_u;
  Point& S1=*meltu->nodes[indi[0]-1], & S2= *meltu->nodes[indi[1]-1];
  MeshElement* meltv=ieparams.melt_v;
  Point& T1=*meltv->nodes[indj[0]-1], & T2= *meltv->nodes[indj[1]-1];

  //definition of matrices MS, MT
  Point MS=S2-S1;
  Point MT=T2-T1;
  real_t G1G2=meltu->geomMapData_p->differentialElement * meltv->geomMapData_p->differentialElement;

  //return;
  number_t N=quadAdjt->numberOfPoints();
  real_t a,b;
  const std::vector< real_t >& coords=quadAdjt->coords();
  const std::vector< real_t >& weights=quadAdjt->weights();
  Matrix<K> matel(res.numberOfRows(),res.numberOfColumns(),K(0));

  // tmpb=tmpb+w(j)*(a*(f(C(a),Cb(a*(1-b))))+(1-a)*(f(C(a),Cb(a+b*(1-a)))));
  for(number_t i1=0; i1<N; i1++)
    {
      real_t w1=weights[i1];
      a=coords[i1];
      for(number_t i2=0; i2<N; i2++)
        {
          real_t w2=weights[i2]*w1;
          b=coords[i2];
          matel*=0.;
          k5(a, b, kuv, S1, MS,T1,MT, elt_S, elt_T, matel, nx, ny, ieparams);
          res+=w2*matel;
        } // end i2
    } // end i1
  res*=G1G2;
  return;
}

template<typename K>
void DuffyIM::k4(real_t x1,real_t x2, const KernelOperatorOnUnknowns& kuv, const Point& S1, const Point & M,
                    const Element* elt_S,const Element* elt_T, Matrix<K>& res, const Vector<real_t>* nx, const Vector<real_t>* ny,
                    IEcomputationParameters& ieparams) const
{
  Point Xp=S1+x1*M; //Ca(X);
  Point Y1=S1+(x1-x1*x2)*M; // Ca(a(1-b))
  Point Y2=S1+(x1+(1.-x1)*x2)*M; // Ca(a+(a-1)b)

  if(ieparams.isP0)    //P0 shortcut
    {
      if(ieparams.isId)
        {
          K tmp;
          kuv.opker().eval(Xp, Y1, tmp, nx, ny);
          res[0]+=x1*tmp;
          tmp*=0.;
          kuv.opker().eval(Xp, Y2, tmp, nx, ny);
          res[0]+=(1-x1)*tmp;
        }
      else
        {
          ShapeValues shv;
          shv.w=std::vector<real_t>(1,1.);
          Matrix<K> tmp=0.*res;
          kuv.eval(Xp, Y1, shv, shv, K(1.), tmp, nx, ny);
          res[0]+=x1*tmp[0];
          tmp*=0.;
          kuv.eval(Xp, Y2, shv, shv, K(1.), tmp, nx, ny);
          res[0]+=(1.-x1)*tmp[0];
        }
      return;
    }

  if(elt_S==elt_T) //same elements, means same interpolation
    {
      //evaluate shape values
      bool wd= (ieparams.difforder_u>0 || ieparams.difforder_v>0);     //order of operator on unknowns >0
      ShapeValues shv_Sx=elt_S->computeShapeValues(Xp,wd);  //=shv_Tx
      ShapeValues shv_Sy1=elt_T->computeShapeValues(Y1,wd); //=shv_Tx
      ShapeValues shv_Sy2=elt_T->computeShapeValues(Y2,wd); //=shv_Tz

      //evaluate matrix
      Matrix<K> tmp=0.*res;
      Vector<K> val_opu, val_opv, val_opk;
      kuv.evalF(Xp, Y1, shv_Sx, shv_Sy1, K(1.), nx, ny,
                ieparams.dimfp_u, ieparams.dimfp_v, ieparams.opisid_u, ieparams.opisid_v,
                ieparams.hasf_u, ieparams.hasf_v, ieparams.scalar_k, 0, val_opu, val_opv, val_opk, tmp);
      res+=x1*tmp;
      tmp=0.*res;
      Vector<K> val_opub, val_opvb, val_opkb;
      kuv.evalF(Xp, Y2, shv_Sx, shv_Sy2, K(1.), nx, ny,
                ieparams.dimfp_u, ieparams.dimfp_v, ieparams.opisid_u, ieparams.opisid_v,
                ieparams.hasf_u, ieparams.hasf_v, ieparams.scalar_k, 0, val_opub, val_opvb, val_opkb, tmp);
      res+=(1.-x1)*tmp;
      return;
    }

  //different element, means not same interpolation if geometric elements are the same
  //evaluate shape values
  Matrix<K> tmp=0.*res;
  bool wdu=ieparams.difforder_u>0, wdv=ieparams.difforder_v>0;   //order of operator on unknowns >0
  ShapeValues shv_Sx=elt_S->computeShapeValues(Xp,wdu);
  ShapeValues shv_Ty1=elt_T->computeShapeValues(Y1,wdv);
  ShapeValues shv_Ty2=elt_T->computeShapeValues(Y2,wdv);

  //evaluate matrix
  Vector<K> val_opu, val_opv, val_opk;
  //kuv.evalF(Xp, Y1, shv_Sx, shv_Ty1, K(1.), nx, ny,
  kuv.evalF(Xp, Y1, shv_Sx, shv_Ty1, K(1.), nx, ny,
            ieparams.dimfp_u, ieparams.dimfp_v, ieparams.opisid_u, ieparams.opisid_v,
            ieparams.hasf_u, ieparams.hasf_v, ieparams.scalar_k, 0, val_opu, val_opv, val_opk, tmp);
  res+=x1*tmp;
  tmp=0.*res;
  Vector<K> val_opub, val_opvb, val_opkb;
  //kuv.evalF(Xp, Y2, shv_Sx, shv_Ty2, K(1.), nx, ny,
  kuv.evalF(Xp, Y2, shv_Sx, shv_Ty2, K(1.), nx, ny,
            ieparams.dimfp_u, ieparams.dimfp_v, ieparams.opisid_u, ieparams.opisid_v,
            ieparams.hasf_u, ieparams.hasf_v, ieparams.scalar_k, 0, val_opu, val_opv, val_opk, tmp);
  res+=(1.-x1)*tmp;
  return;
}


template<typename K>
void DuffyIM::k5(real_t x1,real_t x2, const KernelOperatorOnUnknowns& kuv, const Point& S1, const Point & MS,
                    const Point & T1, const Point & MT, const Element* elt_S,const Element* elt_T, Matrix<K>& res,
                    const Vector<real_t>* nx, const Vector<real_t>* ny, IEcomputationParameters& ieparams) const
{
  Point Y1=S1+x1*MS; //Ca(X);
  Point X1=T1+(x1*x2)*MT; // Ca(a(1-b))
  Point Y2=S1+x1*x2*MS;
  Point X2=T1+x2*MT;
  if(ieparams.isP0)    //P0 shortcut
    {
      if(ieparams.isId)
        {
          K tmp;
          kuv.opker().eval(X1, Y1, tmp, nx, ny);
          res[0]+=x1*tmp;
          tmp*=0.;
          kuv.opker().eval(X2, Y2, tmp, nx, ny);
          res[0]+=(x2)*tmp;
          //	  res*=2.;
        }
      else
        {
          ShapeValues shv;
          shv.w=std::vector<real_t>(1,1.);
          Matrix<K> tmp=0.*res;
          kuv.eval(X1, Y1, shv, shv, K(1.), tmp, nx, ny);
          res+=x1*tmp;
          tmp=0.*res;
          kuv.eval(X2, Y2, shv, shv, K(1.), tmp, nx, ny);
          res+=x2*tmp;
          //res*=2.;
        }
      return;
    } // end if isP0

  if(elt_S==elt_T) //same elements, means same interpolation
    {
      //evaluate shape values
      Matrix<K> tmp=0.*res;
      bool wd= (ieparams.difforder_u>0 || ieparams.difforder_v>0);     //order of operator on unknowns >0
      ShapeValues shv_Sy1=elt_S->computeShapeValues(Y1,wd); //=shv_Tx
      ShapeValues shv_Tx1=elt_T->computeShapeValues(X1,wd); //=shv_Tx
      ShapeValues shv_Sy2=elt_S->computeShapeValues(Y2,wd); //=shv_Tz
      ShapeValues shv_Tx2=elt_T->computeShapeValues(X2,wd); //=shv_Tz

      //evaluate matrix
      Vector<K> val_opu, val_opv, val_opk;
      //kuv.evalF(X1, Y1, shv_Sy1, shv_Tx1, K(1.), nx, ny,
      kuv.evalF(X1, Y1, shv_Tx1, shv_Sy1, K(1.), nx, ny,
                ieparams.dimfp_u, ieparams.dimfp_v, ieparams.opisid_u, ieparams.opisid_v,
                ieparams.hasf_u, ieparams.hasf_v, ieparams.scalar_k, 0, val_opu, val_opv, val_opk, tmp);
      res+=x1*tmp;
      tmp=0.*res;
      Vector<K> val_opub, val_opvb, val_opkb;
      //kuv.evalF(X2, Y2, shv_Sy2, shv_Tx2, K(1.), nx, ny,
      kuv.evalF(X2, Y2, shv_Tx2, shv_Sy2, K(1.), nx, ny,
                ieparams.dimfp_u, ieparams.dimfp_v, ieparams.opisid_u, ieparams.opisid_v,
                ieparams.hasf_u, ieparams.hasf_v, ieparams.scalar_k, 0, val_opub, val_opvb, val_opkb, tmp);
      res+=x2*tmp;
      //res*=2.;
      return;
    }

  //different element, means not same interpolation if geometric elements are the same
  //evaluate shape values
  Matrix<K> tmp=0.*res;
  bool wdu=ieparams.difforder_u>0, wdv=ieparams.difforder_v>0;   //order of operator on unknowns >0
  ShapeValues shv_Sy1=elt_S->computeShapeValues(Y1,wdu);
  ShapeValues shv_Tx1=elt_T->computeShapeValues(X1,wdv);
  ShapeValues shv_Sy2=elt_S->computeShapeValues(Y2,wdu);
  ShapeValues shv_Tx2=elt_T->computeShapeValues(X2,wdv);


  //evaluate matrix
  Vector<K> val_opu, val_opv, val_opk;
  kuv.evalF(X1, Y1, shv_Sy1, shv_Tx1, K(1.), nx, ny,
            ieparams.dimfp_u, ieparams.dimfp_v, ieparams.opisid_u, ieparams.opisid_v,
            ieparams.hasf_u, ieparams.hasf_v, ieparams.scalar_k, 0, val_opu, val_opv, val_opk, tmp);
  res+=x1*tmp;
  tmp=0.*res;
  Vector<K> val_opub, val_opvb, val_opkb;
  kuv.evalF(X2, Y2, shv_Sy2, shv_Tx2, K(1.), nx, ny,
            ieparams.dimfp_u, ieparams.dimfp_v, ieparams.opisid_u, ieparams.opisid_v,
            ieparams.hasf_u, ieparams.hasf_v, ieparams.scalar_k, 0, val_opub, val_opvb, val_opkb, tmp);
  res+=x2*tmp;
  return;
}

}
#endif // DUFFY_IM_HPP



