/*
 XLiFE++ is an extended library of finite elements written in C++
 Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file SauterSchwabIM.hpp
  \authors S.Roman and N.Salles, E. Lunéville.
  \since 16 jun 2014
  \date 16 jun 2017

  \brief Implementation of the Sauter and Schwab's method to compute the FE TermMatrix
             intg_Eu intg Ev opu(y) opk(x,y) opv(x) dy dx
 */


#ifndef SAUTER_SCHWAB_IM_NEW_HPP
#define SAUTER_SCHWAB_IM_NEW_HPP

#include "finiteElements.h"
#include "utils.h"
#include "config.h"
#include "termUtils.hpp"

namespace xlifepp
{

/*!
  \class SauterSchwabIM
  integral over a product of geometric elements with singularity using Sauter-Schwab technique (for elements that shared vertices)
  Stefan A. Sauter, Christoph Schwab, "Boundary Element Methods", Springer, 2010
*/
class SauterSchwabIM : public DoubleIM
{
  public:
    Quadrature* quadSelf;    //!< quadrature on segment [0,1] for self influence (same element)
    Quadrature* quadEdge;    //!< quadrature on segment [0,1] for elements adjacent only by an edge
    Quadrature* quadVertex;  //!< quadrature on segment [0,1] for elements adjacent only by a vertex
    number_t ordSelf;        //!< order of quadrature for self influence (same element)
    number_t ordEdge;        //!< order of quadrature for for elements adjacent only by an edge
    number_t ordVertex;      //!< order of quadrature for for elements adjacent only by a vertex

    SauterSchwabIM(number_t ord =3) //! basic constructor (order 3 on segment)
    {
      ordSelf=ord; ordEdge=ordSelf; ordVertex=ordSelf;
      quadSelf=findQuadrature(_segment,_GaussLegendreRule,ordSelf); quadEdge=quadSelf; quadVertex=quadSelf;
      name="Sauter-Schwab_"+tostring(ordSelf);
      imType=_SauterSchwabIM; singularType=_r; singularOrder=-1;
    }

    SauterSchwabIM(Quadrature& q) //! full constructor from quadrature object
    {
      ordSelf=q.degree; ordEdge=ordSelf; ordVertex=ordSelf;
      quadSelf=&q; quadEdge=quadSelf; quadVertex=quadSelf;
      imType=_SauterSchwabIM; singularType=_r; singularOrder=-1;
      name="Sauter-Schwab_"+tostring(ordSelf);
    }

    virtual SauterSchwabIM* clone() const
    {return new SauterSchwabIM(*this);}

    virtual std::list<Quadrature*> quadratures() const  //! return the list of (single) quadratures in a list
    {
      std::list<Quadrature*> lsq;
      if(quadSelf!=0) lsq.push_back(quadSelf);
      if(quadEdge!=0 && quadEdge!=quadSelf)  lsq.push_back(quadEdge);
      if(quadVertex!=0 && quadVertex!=quadSelf && quadVertex!=quadEdge) lsq.push_back(quadVertex);
      return lsq;
    }

    virtual void print(std::ostream& os) const          //! print IntegrationMethod on stream
    {
      os<<"Sauter-Schwab integration method with "<< quadSelf->name<<" degree " << quadSelf->degree
        << " ( " << quadSelf->quadratureRule.size() << " points ) " <<eol;
    }
    virtual void print(PrintStream& os) const {print(os.currentStream());}

    //template computeIE function, explicit instanciation because template functions are not inherited
    template<typename K>
    void computeIE(const Element* elt_S, const Element* elt_T, AdjacenceInfo& adj,
                   const KernelOperatorOnUnknowns& kuv, Matrix<K>& res,IEcomputationParameters& ieparams,
                   Vector<K>& val_opu, Vector<K>& val_opv, Vector<K>& val_opk) const;
    void computeIE(const Element* elt_S, const Element* elt_T, AdjacenceInfo& adj,
                   const KernelOperatorOnUnknowns& kuv, Matrix<real_t>& res,IEcomputationParameters& ieparams,
                   Vector<real_t>& val_opu, Vector<real_t>& val_opv, Vector<real_t>& val_opk) const
    {computeIE<>(elt_S, elt_T,adj, kuv, res,ieparams, val_opu, val_opv, val_opk);}
    void computeIE(const Element* elt_S, const Element* elt_T, AdjacenceInfo& adj,
                   const KernelOperatorOnUnknowns& kuv, Matrix<complex_t>& res,IEcomputationParameters& ieparams,
                   Vector<complex_t>& val_opu, Vector<complex_t>& val_opv, Vector<complex_t>& val_opk) const
    {computeIE<>(elt_S, elt_T,adj, kuv, res,ieparams, val_opu, val_opv, val_opk);}

    //quadrature tools
    template<typename K>
    void selfInfluences(const Element* elt_S, const KernelOperatorOnUnknowns& kuv,
                        const Vector<real_t>* nx, const Vector<real_t>* ny,
                        Matrix<K>& res, IEcomputationParameters& ieparams) const;
    template<typename K>
    void adjacentTrianglesbyEdge(const Element* elt_S, const Element* elt_T, const KernelOperatorOnUnknowns& kuv,
                                 const Vector<real_t>* nx, const Vector<real_t>* ny, const Vector< number_t >& indi, const Vector< number_t >& indj,
                                 Matrix<K>& res, IEcomputationParameters& ieparams) const;
    template<typename K>
    void adjacentTrianglesbyVertex(const Element* elt_S, const Element* elt_T, const KernelOperatorOnUnknowns& kuv, const Vector<real_t>* nx,
                                   const Vector<real_t>* ny, const Vector< number_t >& indi, const Vector< number_t >& indj,
                                   Matrix<K>& res, IEcomputationParameters& ieparams) const;

    template<typename K>
    void k2(const Point & Xp, const Point & Zp,const KernelOperatorOnUnknowns& kuv, const Point& S1, const Vector< Point >& M,
            const Element* elt_S,const Element* elt_T, Matrix<K>& res, const Vector<real_t>* nx, const Vector<real_t>* ny,
            IEcomputationParameters& ieparams) const;
    template<typename K>
    void k3(const Point & Ca, const Point & Cb, const KernelOperatorOnUnknowns& kuv, const Point& S1, const Point& T1,
            const Vector< Point >& MS, const Vector< Point >& MT, const Element* elt_S, const Element* elt_T, Matrix<K>& res,
            const Vector<real_t>* nx, const Vector<real_t>* ny, IEcomputationParameters& ieparams) const;
};

// implementation of template member functions
template<typename K>
void SauterSchwabIM::computeIE(const Element* elt_S, const Element* elt_T, AdjacenceInfo& adj,
                                  const KernelOperatorOnUnknowns& kuv, Matrix<K>& res, IEcomputationParameters& ieparams,
                                  Vector<K>& val_opu, Vector<K>& val_opv, Vector<K>& val_opk) const
{
  if (ieparams.sh_u!=_triangle)
    error("shape_not_handled", words("shape", ieparams.sh_u));
  if (ieparams.sh_v!=_triangle)
    error("shape_not_handled", words("shape", ieparams.sh_v));

  //load normals
  Vector<real_t>* nx=0, *ny=0;
  MeshElement* melt_S=ieparams.melt_u, *melt_T=ieparams.melt_v;
  if (kuv.xnormalRequired()) nx = &melt_T->geomMapData_p->normalVector;
  if (kuv.ynormalRequired()) ny = &melt_S->geomMapData_p->normalVector;

  //  compute quadrature regarding triangle position
  std::vector<number_t>& indi=adj.sharedVertex1, &indj=adj.sharedVertex2;
  switch (adj.status)
  {

    case _adjacentByElement : selfInfluences(elt_S, kuv, nx, ny, res, ieparams); return;
    case _adjacentBySide :
    {
      indi[2]=6-indi[0]-indi[1]; indj[2]=6-indj[0]-indj[1];
      adjacentTrianglesbyEdge(elt_S, elt_T, kuv, nx, ny, indi, indj, res, ieparams);
      return;
    }
    case _adjacentByVertex :
    {
      switch (indi[0])
      {
        case 1  : indi[1]=2; indi[2]=3; break;
        case 2  : indi[1]=3; indi[2]=1; break;
        default : indi[1]=1; indi[2]=2;
      }
      switch (indj[0])
      {
        case 1  : indj[1]=2; indj[2]=3; break;
        case 2  : indj[1]=3; indj[2]=1; break;
        default : indj[1]=1; indj[2]=2;
      }
      adjacentTrianglesbyVertex(elt_S, elt_T, kuv, nx, ny, indi, indj, res, ieparams);
      return;
    }
    default:
      error("adjacence_status_not_handled");
  }
}

/*================================================================================================================
                          Same triangles, singular integral -> use Sauter-Schwab formulae

  up to now, it works only for scalar cases

  arguments :
     elt_S, elt_T : elements involved in integral
     kuv : integrand op(u)*op(ker)*op(v)
     nx,ny : pointers to normal of elt_T and elt_S (0 if not used)
     ieparams : some parameters of computation
     indi, indj : vertex numbering
     quad :  1d quadrature method used inSauter-Schwabb method
     res : matrix int_elt_S x elt_T op(wj)*op(ker)*op(wi)

  ================================================================================================================*/

template<typename K>
void SauterSchwabIM::selfInfluences(const Element* elt_S, const KernelOperatorOnUnknowns& kuv,
                                       const Vector<real_t>* nx, const Vector<real_t>* ny,
                                       Matrix<K>& res, IEcomputationParameters& ieparams) const
{
  MeshElement* meltu=ieparams.melt_u;
  Point& S1=*meltu->nodes[0], & S2= *meltu->nodes[1],  & S3= *meltu->nodes[2];

  //definition of matrix MS
  Vector< Point > M(2);
  M[0]=S2-S1;
  M[1]=S3-S2;
  real_t G12=meltu->geomMapData_p->differentialElement; G12*=G12;

  number_t N=quadSelf->numberOfPoints();
  real_t a, b, c, d;
  const std::vector< real_t >& coords=quadSelf->coords();
  const std::vector< real_t >& weights=quadSelf->weights();
  Matrix<K> matel(res.numberOfRows(),res.numberOfColumns(),K(0));

  Point U=M[0];
  Point V=M[1];
  Point UV=U+V;
  for(number_t i1=0; i1<N; i1++)
    {
      a=coords[i1];
      Point X0=S1+a*U; //M[0];
      Point Z0=X0+a*V;
      for(number_t i2=0; i2<N; i2++)
        {
          real_t w2=weights[i1]*weights[i2];
          b=coords[i2];
          real_t ab=a*b;
          Point abV=ab*V;
          Point Z10=Z0-abV;
          Point X30=X0+abV; //M[1];
          for(number_t i3=0; i3<N; i3++)
            {
              c=coords[i3];
              real_t w3=w2*weights[i3];
              real_t a3b2c=a*a*a*b*b*c*G12, abc=ab*c;
              Point abcV=abc*V;
              Point Z3=X30-abcV;
              Point X1=Z10+abcV;
              Point Z2=X30-abcV-abc*U;
              for(number_t i4=0; i4<N; i4++)
                {
                  real_t w4=w3*weights[i4];
                  d=coords[i4];
                  real_t abcd=abc*d;
                  matel*=0.;
                  Point tU=abcd*U;
                  Point tV=abcd*V;
                  Point Z1=Z10-tU;
                  Point X2=Z3+tV;
                  Point X3=X30-tU-tV;
                  k2(X1,Z1,kuv, S1, M, elt_S, elt_S, matel, nx, ny, ieparams);
                  k2(X2,Z2, kuv, S1, M, elt_S, elt_S, matel, nx, ny, ieparams);
                  k2(X3,Z3, kuv, S1, M, elt_S, elt_S, matel, nx, ny, ieparams);
                  res+=w4*a3b2c*matel;

                }
            }
        }
    }
}

/*================================================================================================================
                          Adjacent triangles by edge, singular integral -> use Sauter-Schwab formulae

  up to now, it works only for scalar cases

  arguments :
     elt_S, elt_T : elements involved in integral
     kuv : integrand op(u)*op(ker)*op(v)
     nx,ny : pointers to normal of elt_T and elt_S (0 if not used)
     ieparams : some parameters of computation
     indi, indj : vertex numbering
     quad :  1d quadrature method used inSauter-Schwabb method
     res : matrix int_elt_Sy x elt_Tx op(wj)(y)*op(ker)(x,y)*op(wi)(x)

  ================================================================================================================*/

template<typename K>
void SauterSchwabIM::adjacentTrianglesbyEdge(const Element* elt_S, const Element* elt_T, const KernelOperatorOnUnknowns& kuv,
    const Vector<real_t>* nx, const Vector<real_t>* ny, const Vector< number_t >& indi, const Vector< number_t >& indj,
    Matrix<K>& res, IEcomputationParameters& ieparams) const
{
  MeshElement* meltu=ieparams.melt_u;
  Point& S1=*meltu->nodes[indi[0]-1], & S2= *meltu->nodes[indi[1]-1],  & S3= *meltu->nodes[indi[2]-1];
  MeshElement* meltv=ieparams.melt_v;
  Point& T1=*meltv->nodes[indj[0]-1], & T2= *meltv->nodes[indj[1]-1],  & T3= *meltv->nodes[indj[2]-1];

  //definition of matrices MS and MT
  Vector< Point > MS(2), MT(2);
  MS[0]=S2-S1;   MS[1]=S3-S2;
  MT[0]=T2-T1;   MT[1]=T3-T2;

  real_t G1G2=meltu->geomMapData_p->differentialElement * meltv->geomMapData_p->differentialElement;

  // Definition des Points et Poids de quadrature
  number_t N=quadEdge->numberOfPoints();
  real_t a,b,c,d;
  const std::vector< real_t >& coords=quadEdge->coords();
  const std::vector< real_t >& weights=quadEdge->weights();
  Matrix<K> matel1(res.numberOfRows(),res.numberOfColumns(),K(0)), matel(res.numberOfRows(),res.numberOfColumns(),K(0));

  Point Uxy=MS[0]+MS[1];
  Point Vxy=MT[0]+MT[1];
  for(number_t i1=0; i1<N; i1++)
    {
      a=coords[i1];
      Point X0=S1+a*MS[0];
      Point Y0=T1+a*MT[0];
      for(number_t i2=0; i2<N; i2++)
        {
          real_t w2=weights[i1]*weights[i2];
          b=coords[i2];
          real_t a3b2=a*a*a*b*b*G1G2, ab=a*b;
          Point X2=X0+ab*MS[1];
          Point Y4=Y0+ab*MT[1];
          Point Y1t=Y0+ab*MT[1];
          for(number_t i3=0; i3<N; i3++)
            {
              real_t w3=w2*weights[i3];
              c=coords[i3];
              real_t abc=ab*c;
              Point X3=X2-abc*Uxy;
              Point X4t=X0+abc*MS[1];
              Point Y5=Y0+abc*MT[1];
              Point Y1=Y1t-abc*Vxy;
              for(number_t i4=0; i4<N; i4++)
                {
                  real_t w4=w3*weights[i4];
                  d=coords[i4];
                  real_t abcd=abc*d;
                  real_t abd=ab*d;
                  Point tmpX=abcd*Uxy;
                  Point X4=X4t-tmpX;
                  Point X1=X0+abd*MS[1];
                  Point X5=X2-tmpX;
                  Point Y2=Y5-abcd*Vxy;
                  Point Y3=Y0+abcd*MT[1];
                  matel*=0.; matel1*=0.;
                  k3(X1,Y1, kuv, S1, T1, MS, MT, elt_S, elt_T, matel1, nx, ny, ieparams);
                  k3(X2,Y2,kuv, S1, T1, MS, MT, elt_S, elt_T, matel, nx, ny, ieparams);
                  k3(X3,Y3, kuv, S1, T1, MS, MT, elt_S, elt_T, matel, nx, ny, ieparams);
                  k3(X4,Y4, kuv, S1, T1, MS, MT, elt_S, elt_T, matel, nx, ny, ieparams);
                  k3(X5,Y5, kuv, S1, T1, MS, MT, elt_S, elt_T, matel, nx, ny, ieparams);
                  res+=a3b2*w4*(c*matel+matel1);
                }
            }
        }
    }
}

/*================================================================================================================
                          Adjacent triangles by vertex, singular integral -> use Sauter-Schwab formulae

  up to now, it works only for scalar cases

  arguments :
     elt_S, elt_T : elements involved in integral
     kuv : integrand op(u)*op(ker)*op(v)
     nx,ny : pointers to normal of elt_T and elt_S (0 if not used)
     ieparams : some parameters of computation
     indi, indj : vertex numbering
     quad :  1d quadrature method used inSauter-Schwabb method
     res : matrix int_elt_Sy x elt_Tx op(wj)(y)*op(ker)(x,y)*op(wi)(x)

  ================================================================================================================*/
template<typename K>
void SauterSchwabIM::adjacentTrianglesbyVertex(const Element* elt_S, const Element* elt_T, const KernelOperatorOnUnknowns& kuv,
    const Vector<real_t>* nx, const Vector<real_t>* ny, const Vector< number_t >& indi, const Vector< number_t >& indj,
    Matrix<K>& res, IEcomputationParameters& ieparams) const
{
  MeshElement* meltu=ieparams.melt_u;
  Point& S1=*meltu->nodes[indi[0]-1], & S2= *meltu->nodes[indi[1]-1],  & S3= *meltu->nodes[indi[2]-1];
  MeshElement* meltv=ieparams.melt_v;
  Point& T1=*meltv->nodes[indj[0]-1], & T2= *meltv->nodes[indj[1]-1],  & T3= *meltv->nodes[indj[2]-1];

  //definition of matrices MS, MT
  Vector< Point > MS(2), MT(2);
  MS[0]=S2-S1;  MS[1]=S3-S2;
  MT[0]=T2-T1;  MT[1]=T3-T2;
  real_t G1G2=meltu->geomMapData_p->differentialElement * meltv->geomMapData_p->differentialElement;

  number_t N=quadVertex->numberOfPoints();
  real_t a,b,c,d;
  const std::vector< real_t >& coords=quadVertex->coords();
  const std::vector< real_t >& weights=quadVertex->weights();
  Matrix<K> matel(res.numberOfRows(),res.numberOfColumns(),K(0));

  for(number_t i1=0; i1<N; i1++)
    {
      a=coords[i1];
      Point X0=S1+a*MS[0];
      Point Y0=T1+a*MT[0];
      for(number_t i2=0; i2<N; i2++)
        {
          real_t w2=weights[i2]*weights[i1];
          b=coords[i2];
          real_t ab=a*b;
          Point X1=X0+ab*MS[1];
          Point Y2=Y0+ab*MT[1];
          for(number_t i3=0; i3<N; i3++)
            {
              real_t w3=w2*weights[i3];
              c=coords[i3];
              real_t a3c=a*a*a*c*G1G2, ac=a*c;
              Point Xt=S1+ac*MS[0];
              Point Yt=T1+ac*MT[0];
              for(number_t i4=0; i4<N; i4++)
                {
                  real_t w4=w3*weights[i4];
                  d=coords[i4];
                  real_t acd=ac*d;
                  Point X2=Xt+acd*MS[1];
                  Point Y1=Yt+acd*MT[1];
                  matel*=0.;
                  k3(X1,Y1, kuv, S1, T1, MS, MT, elt_S, elt_T, matel, nx, ny, ieparams);
                  k3(X2,Y2, kuv, S1, T1, MS, MT, elt_S, elt_T, matel, nx, ny, ieparams);
                  res+=a3c*w4*matel;
                } // end i4
            } // end i3
        } // end i2
    } // end i1
}

/*================================================================================================================
                                       member functions k2 ans k3
  ================================================================================================================*/
template<typename K>
void SauterSchwabIM::k2(const Point & Xp, const Point & Zp,const KernelOperatorOnUnknowns& kuv, const Point& S1, const Vector< Point >& M,
                           const Element* elt_S,const Element* elt_T, Matrix<K>& res, const Vector<real_t>* nx, const Vector<real_t>* ny,
                           IEcomputationParameters& ieparams) const
{
  if(ieparams.isP0 &&  ieparams.dimf_u==1 &&  ieparams.dimf_v==1)    //scalar P0 shortcut
    {
      if(ieparams.isId)
        {
          K tmp;
          kuv.opker().eval(Xp, Zp, tmp, nx, ny);
          res[0]+=tmp;
          kuv.opker().eval(Zp, Xp, tmp, nx, ny);
          res[0]+=tmp;
        }
      else
        {
          ShapeValues shv;
          shv.w=std::vector<real_t>(1,1.);
          kuv.eval(Xp, Zp, shv, shv, K(1.), res, nx, ny);
          kuv.eval(Zp, Xp, shv, shv, K(1.), res, nx, ny);
        }
      return;
    }

  if(elt_S==elt_T) //same elements, means same interpolation
    {
      //evaluate shape values
      bool wd= (ieparams.difforder_u>0 || ieparams.difforder_v>0);     //order of operator on unknowns >0
      ShapeValues shv_Sx=elt_S->computeShapeValues(Xp,wd,ieparams.sign_u); //=shv_Tx
      ShapeValues shv_Sz=elt_S->computeShapeValues(Zp,wd,ieparams.sign_u); //=shv_Tz
      //evaluate matrix
      Vector<K> val_opu, val_opv, val_opk;
      kuv.evalF(Xp, Zp, shv_Sx, shv_Sz, K(1.), nx, ny,
                ieparams.dimfp_u, ieparams.dimfp_v, ieparams.opisid_u, ieparams.opisid_v,
                ieparams.hasf_u, ieparams.hasf_v, ieparams.scalar_k, 0, val_opu, val_opv, val_opk, res);
      kuv.evalF(Zp, Xp, shv_Sz, shv_Sx, K(1.), nx, ny,
                ieparams.dimfp_u, ieparams.dimfp_v, ieparams.opisid_u, ieparams.opisid_v,
                ieparams.hasf_u, ieparams.hasf_v, ieparams.scalar_k, 0, val_opu, val_opv, val_opk, res);
      return;
    }

  //different element, means not same interpolation if geometric elements are the same
  //evaluate shape values
  bool wdu=ieparams.difforder_u>0, wdv=ieparams.difforder_v>0;   //order of operator on unknowns >0
  ShapeValues shv_Sx=elt_S->computeShapeValues(Xp,wdu,ieparams.sign_u);
  ShapeValues shv_Sz=elt_S->computeShapeValues(Zp,wdu,ieparams.sign_u);
  ShapeValues shv_Tx=elt_T->computeShapeValues(Xp,wdv,ieparams.sign_v);
  ShapeValues shv_Tz=elt_T->computeShapeValues(Zp,wdv,ieparams.sign_v);
  //evaluate matrix
  Vector<K> val_opu, val_opv, val_opk;
  kuv.evalF(Xp, Zp, shv_Sx, shv_Tz, K(1.), nx, ny,
            ieparams.dimfp_u, ieparams.dimfp_v, ieparams.opisid_u, ieparams.opisid_v,
            ieparams.hasf_u, ieparams.hasf_v, ieparams.scalar_k, 0, val_opu, val_opv, val_opk, res);
  kuv.evalF(Zp, Xp, shv_Sz, shv_Tx, K(1.), nx, ny,
            ieparams.dimfp_u, ieparams.dimfp_v, ieparams.opisid_u, ieparams.opisid_v,
            ieparams.hasf_u, ieparams.hasf_v, ieparams.scalar_k, 0, val_opu, val_opv, val_opk, res);
  return;
}

template<typename K>
void SauterSchwabIM::k3(const Point & Ca, const Point & Cb, const KernelOperatorOnUnknowns& kuv, const Point& S1, const Point& T1,
                           const Vector<Point>& MS, const Vector<Point>& MT, const Element* elt_S, const Element* elt_T, Matrix<K>& res,
                           const Vector<real_t>* nx, const Vector<real_t>* ny, IEcomputationParameters& ieparams) const
{

  if(ieparams.isP0 &&  ieparams.dimf_u==1 &&  ieparams.dimf_v==1)
    {
      if(ieparams.isId)
        {
          K tmp=K(0.);
          kuv.opker().eval(Cb, Ca, tmp, nx, ny);
          res[0]+=tmp;
        }
      else
        {
          ShapeValues shv;
          shv.w=std::vector<real_t>(1,1.);
          kuv.eval(Cb, Ca, shv, shv, K(1.), res, nx, ny);
        }
      return;
    }
  //general case
  ShapeValues shv_Sx=elt_S->computeShapeValues(Ca,ieparams.difforder_u>0,ieparams.sign_u);
  ShapeValues shv_Tz=elt_T->computeShapeValues(Cb,ieparams.difforder_v>0,ieparams.sign_v);
  Vector<K> val_opu, val_opv, val_opk;
  kuv.evalF(Cb, Ca, shv_Sx, shv_Tz, K(1.), nx, ny,
            ieparams.dimfp_u, ieparams.dimfp_v, ieparams.opisid_u, ieparams.opisid_v,
            ieparams.hasf_u, ieparams.hasf_v, ieparams.scalar_k, 0, val_opu, val_opv, val_opk, res);
}

}
#endif // SAUTER_SCHWAB_IM_HPP
