/*
XLiFE++ is an extended library of finite elements written in C++
    Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file termUtils.hpp
  \author N. Kielbasiewicz
  \since 27 feb 2013
  \date 1 mar 2013

  \brief Definition of term utility functions
*/

#ifndef TERM_UTILS_HPP
#define TERM_UTILS_HPP

#include "config.h"
#include "largeMatrix.h"
#include "space.h"

namespace xlifepp
{

//data structure to store IE computation parameters
class IEcomputationParameters
{
  public:
    dimen_t dim_sp;                   //!< space dimension (elt->geomElt_p->mesh()->spaceDim())
    bool extend_u, extend_v;          //!< true if u/v lives on a extended domain
    number_t side_u, side_v;          //!< side numbers when a extended domain (0 when not used)
    FEMapType femt_u, femt_v;         //!< type of map type (_standardMap,_contravariantPiolaMap,_covariantPiolaMap)
    ShapeType sh_u, sh_v;             //!< shape of elements
    number_t ord_u, ord_v;            //!< degree of u/v interpolation
    dimen_t dimf_u, dimf_v;           //!< dimension of u/v shape functions of reference element
    dimen_t dimfp_u, dimfp_v;         //!< dimension of u/v shape functions of physical element
    bool opisid_u, opisid_v;          //!< true if operator on u/v is identity
    bool hasf_u, hasf_v;              //!< true if operator on u/v requires to evaluate a function
    dimen_t difforder_u, difforder_v; //!< order of differential operator
    bool scalar_k;                    //!< true if scalar kernel
    bool isP0;                        //!< true if u and v interpolations are P0
    bool isId;                        //!< true if operator on u and v are identity
    number_t nb_blf;                  //!< number of bilinear forms
    MeshElement* melt_u, *melt_v;     //!< pointer to mesh elements
    bool rotsh_u, rotsh_v;            //!< true if dof rotation has to be done
    bool changeSign_u, changeSign_v;  //!< if true change signs of u/v dofs according to sign_u/sign_v
    Vector<real_t>* sign_u, *sign_v;  //!< pointer to dof signs
    bool knor_x, knor_y;              //!< true if x-normal/y-normal is required by kernel;
    bool isUptodate;                  //!< true if up to date

    IEcomputationParameters()
      : dim_sp(2), extend_u(false), extend_v(false), side_u(0), side_v(0),femt_u(_standardMap), femt_v(_standardMap), sh_u(_noShape), sh_v(_noShape),
        ord_u(0), ord_v(0), dimf_u(1), dimf_v(1), dimfp_u(1), dimfp_v(1), opisid_u(true), opisid_v(true), hasf_u(false), hasf_v(false),
        difforder_u(0), difforder_v(0), scalar_k(true), isP0(true), isId(true), nb_blf(1), melt_u(0), melt_v(0),
        rotsh_u(false), rotsh_v(false), changeSign_u(false), changeSign_v(false), sign_u(0), sign_v(0), knor_x(false), knor_y(false), isUptodate(false) {}
};

//storage stuff
MatrixStorage* buildStorage(const Space &rs, const Space &cs, StorageType st, AccessType at); //!< build a storage from a pair of Space
MatrixStorage* buildStorage(StorageType st, AccessType at, number_t dimr, number_t dimc); //!<build a storage from  type and dimension, no allocation of pointers (void matrix)
MatrixStorage* buildStorage(StorageType st, AccessType at, number_t dimr, number_t dimc, const std::vector<std::vector<number_t> >& indices); //!< build a storage from  type, dimension and column indices (vector of vectors)
MatrixStorage* buildStorage(StorageType st, AccessType at, number_t dimr, number_t dimc, const std::vector<std::set<number_t> >& indices); //!< build a storage from  type, dimension and column indices (vector of sets)
MatrixStorage* updateStorage(MatrixStorage& stm, const std::vector<number_t> rows, const std::vector<number_t> cols,
                             StorageType st, AccessType at, bool overwrite=false);  //!< update storage for adding dense submatrix given by its row and column indices

//interpolation stuff
void interpolatedNormals(Space& sp, std::vector<Vector<real_t> >& ns); //!< compute interpolated normals on Lagrange Dofs of space sp
TermVector normalsOn(GeomDomain& dom, const Unknown& u); //!< compute normals of a side domain using interpolation given by an unknown

// special cast function used by FEComputation's
real_t toRealComplex(const Vector<real_t>&,const complex_t&);
complex_t toRealComplex(const Vector<complex_t>&,const complex_t&);
real_t toRealComplex(const Vector<Vector<real_t> >&,const complex_t&);
complex_t toRealComplex(const Vector<Vector<complex_t> >&,const complex_t&);

//precomputation
void preComputationIE(const GeomDomain& dom, const Space* subsp, std::set<Quadrature*>& quads,
                      dimen_t nbc, bool der, bool nor, bool extendedDomain=false); //!< precompute geometry data and shape values on any element for any quadrature involved

} //end of namespace xlifepp

#endif // TERM_UTILS_HPP

