/*
XLiFE++ is an extended library of finite elements written in C++
   Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file ThreadData.hpp
  \author E. Luneville
  \since 25 apr 2018
  \date 25 apr 2018

  \brief ThreadData class handling some possible shared data in omp computation(normal vectors, element, dof, ...)

  When doing FE computation, some local informations related to one elements (normal vectors, element itself, dof, ...)
  have to be passed to user functions. When multithreading (say omp) is enable, these informations may depend on the thread,
  so multiple data (one by thread) has to be stored somewhere. It is the purpose of the ThreadData class that handles the following vectors

    vector<Vector<real_t>*> theCurrentNxs   -> normal or x-normal pointer vectors
    vector<Vector<real_t>*> theCurrentNys   -> y-normal pointer vectors
    vector<GeomElement*> theCurrentElements -> geom element pointer vectors
    vector<Dof*> theCurrentDofs             -> dof pointer vectors
    GeomDomain* currentDomainx              -> domain or domainx pointer vectors (not omp)
    GeomDomain* currentDomainy              -> domainy pointer vectors (not omp)

  and proposes some functions to set or get data

    setNx, getNx, setNy, getNy, setElement, getElement, setDof, getDof, setDomain,getDomain

  At the initialization, XLiFE++ instances one ThreadData object : theThreadData (see globalScopeData.hpp, config.hpp and init())
  To make the user life easier, the users can use the extern functions getNx, getN, getNy, getElement, getFeDof, getDof, getDomain
  that addresses the theThreadData object

  Note : when single thread computation, only the first element of vectors is used

  To users : This machinery replaces the transmission of normal vectors by the parameter associated to Function or Kernel
             How it works now? Suppose you want to use a function involving the normal vector at point x

             a) define a standard C++ XLiFE++ function
                Real fun(const Point& x, Parameters& pa=defaultParameters)
                  { ...
                    Vector<Real>& n=getN();  //to get the normal vector at x if computed!
                    ...
                  }

             b) create a Function object (with or without parameters) in your program
                  ...
                  Function F(fun);          // Function object with no parameter handling the C++ function fun
                  F.require(_n);            // tell to XLiFE++ that F will use the normal n
                                            // may be disable F.require(_n, false)

             Remember that the normal will be updated, only if your function is involved in a FE computation on a manifold (boundary for instance)

*/

#ifndef THREAD_DATA_HPP
#define THREAD_DATA_HPP

#include "config.h"
#include "Vector.hpp"

namespace xlifepp
{
 //forward classes
 class GeomElement;
 class Dof;
 class FeDof;
 class GeomDomain;

//=================================================================================
/*!
   \class ThreadData
   class handling some possible shared data in omp computation(normal vectors, element, dof, ...)
*/
//=================================================================================

class ThreadData
{
public:
  std::vector<Vector<real_t>*> theCurrentNxs;     //!< normal pointers or x-normal pointers
  std::vector<Vector<real_t>*> theCurrentNys;     //!< y-normal pointers
  std::vector<GeomElement*> theCurrentElements;   //!< element pointers
  std::vector<Dof*> theCurrentDofs;               //!< dof pointers
  GeomDomain* currentDomainx;                     //!< domain pointer or domain_x pointer
  GeomDomain* currentDomainy;                     //!< domain_y pointer

  ThreadData(number_t s=1) {resize(s);}
  void resize(number_t); //! resize global vectors theCurrentNxs, theCurrentNys, ... to size s

  //set data
  void setNx(Vector<real_t>* np)  {theCurrentNxs[currentThread()]=np;}
  void setNy(Vector<real_t>* np)  {theCurrentNys[currentThread()]=np;}
  void setElement(GeomElement* ep){theCurrentElements[currentThread()]=ep;}
  void setDomainx(GeomDomain* dp) {currentDomainx=dp;}
  void setDomainy(GeomDomain* dp) {currentDomainy=dp;}
  void setDof(Dof* dp)            {theCurrentDofs[currentThread()]=dp;}

  void setNx(Vector<real_t>* np, number_t t)  {theCurrentNxs[t]=np;}
  void setNy(Vector<real_t>* np, number_t t)  {theCurrentNys[t]=np;}
  void setElement(GeomElement* ep, number_t t){theCurrentElements[t]=ep;}
  void setDof(Dof* dp, number_t t)            {theCurrentDofs[t]=dp;}

  //get data
  Vector<real_t>& getNx(number_t t) const
  {
    Vector<real_t>* p=theCurrentNxs[t];
    if(p==0) p = theCurrentNxs[0];
    if(p==0) error("omp_no_data","n/nx", t);
    return *p;
  }
  Vector<real_t>& getNx() const {return getNx(currentThread());}
  Vector<real_t>& getNy(number_t t) const
  {
    Vector<real_t>* p=theCurrentNys[t];
    if(p==0) p = theCurrentNys[0];
    if(p==0) error("omp_no_data","ny", t);
    return *p;
  }
  Vector<real_t>& getNy() const {return getNy(currentThread());}

  GeomElement& getElement(number_t t) const
  {
    GeomElement* p=theCurrentElements[t];
    if(p==0) p = theCurrentElements[0];
    if(p==0) error("omp_no_data","elt", t);
    return *p;
  }
  GeomElement& getElement() const {return getElement(currentThread());}

  GeomDomain& getDomainx() const
  {
    if(currentDomainx==0) error("omp_no_data","domain or domain_x", 0);
    return *currentDomainx;
  }
  GeomDomain& getDomainy() const
  {
    if(currentDomainy==0) error("omp_no_data","domain_y", 0);
    return *currentDomainy;
  }

  Dof& getDof(number_t t) const
  {
    Dof* p=theCurrentDofs[t];
    if(p==0) p = theCurrentDofs[0];
    if(p==0) error("omp_no_data","domain", t);
    return *p;
  }
  Dof& getDof() const {return getDof(currentThread());}

  FeDof& getFeDof(number_t t) const;
  FeDof& getFeDof() const {return getFeDof(currentThread());}

  void print(std::ostream&) const;
  void print(PrintStream& os) const {print(os.currentStream());}
};

inline void setNx(Vector<real_t>* p, number_t t)  {theThreadData.setNx(p,t);}
inline void setNy(Vector<real_t>* p, number_t t)  {theThreadData.setNy(p,t);}
inline void setElement(GeomElement* p, number_t t){theThreadData.setElement(p,t);}
inline void setDof(Dof* p,number_t t)             {theThreadData.setDof(p,t);}

inline void setNx(Vector<real_t>* p)  {theThreadData.setNx(p,currentThread());}
inline void setNy(Vector<real_t>* p)  {theThreadData.setNy(p,currentThread());}
inline void setElement(GeomElement* p){theThreadData.setElement(p,currentThread());}
inline void setDomain(GeomDomain* p)  {theThreadData.setDomainx(p);}
inline void setDomainx(GeomDomain* p) {theThreadData.setDomainx(p);}
inline void setDomainy(GeomDomain* p) {theThreadData.setDomainy(p);}
inline void setDof(Dof* p)            {theThreadData.setDof(p,currentThread());}

inline Vector<real_t>&    getN(number_t t) {return theThreadData.getNx(t);}
inline Vector<real_t>&   getNx(number_t t) {return theThreadData.getNx(t);}
inline Vector<real_t>&   getNy(number_t t) {return theThreadData.getNy(t);}
inline GeomElement& getElement(number_t t) {return theThreadData.getElement(t);}
inline Dof&             getDof(number_t t) {return theThreadData.getDof(t);}
inline FeDof&         getFeDof(number_t t) {return theThreadData.getFeDof(t);}

inline Vector<real_t>&    getN() {return theThreadData.getNx(currentThread());}
inline Vector<real_t>&   getNx() {return theThreadData.getNx(currentThread());}
inline Vector<real_t>&   getNy() {return theThreadData.getNy(currentThread());}
inline GeomElement& getElement() {return theThreadData.getElement(currentThread());}
inline GeomDomain&   getDomain() {return theThreadData.getDomainx();}
inline GeomDomain&   getDomainx(){return theThreadData.getDomainx();}
inline GeomDomain&   getDomainy(){return theThreadData.getDomainy();}
inline Dof&             getDof() {return theThreadData.getDof(currentThread());}
inline FeDof&         getFeDof() {return theThreadData.getFeDof(currentThread());}
inline Vector<real_t>& getVector(UnitaryVector un)
{
    if(un==_n || un==_nx)  return getNx();
    if(un==_ny) return getNy();
    where("Parameters::getVector(String)");
    error("unitaryvector_not_handled",un);
    return getNx();  // dummy return
}

}
#endif // THREAD_DATA_HPP

