/*
XLiFE++ is an extended library of finite elements written in C++
    Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file Transformation.hpp
  \author N. Kielbasiewicz
  \since 8 oct 2014
  \date 08 feb 2016

  \brief Definition of the xlifepp::Transformation class

  xlifepp::Transformation class represents geometric transformations applied to Mesh or Geometry objects
*/

#ifndef TRANSFORMATIONS_HPP
#define TRANSFORMATIONS_HPP

#include "config.h"
#include "Point.hpp"
#include "Messages.hpp"
#include "Environment.hpp"

namespace xlifepp
{

// forward declarations
class Translation;
class Rotation2d;
class Rotation3d;
class Homothety;
class PointReflection;
class Reflection2d;
class Reflection3d;

/*!
   \class Transformation
    handles mesh transformations
 */
class Transformation
{
  protected:
    string_t name_;             //!< name of the transformation
    TransformType transformType_; //!< geometrical transformation
  private:
    std::vector<const Transformation*> components_;   //!< composite or loop Transformation if not empty

  public:
    Transformation(const string_t& nam = "", TransformType trt=_noTransform); //!< basic constructor
    Transformation(const Transformation& t);                                //!< copy constructor

    virtual Transformation* clone() const { return new Transformation(*this); } //!< virtual copy constructor

    virtual ~Transformation(); //!< destructor

    Transformation & operator=(const Transformation& t); //!< assign operator
    TransformType transformType() const      //! return the geometrical transformation
      { return transformType_; }
    void transformType(TransformType tr) //! sets the geometrical shape
      { transformType_=tr; }

    string_t name() const { return name_; }       //!< get the Transformation name
    void name(const string_t& nm) { name_ = nm; } //!< set the Transformation name

    const std::vector<const Transformation*> components() const { return components_;} //!< accessor to components
    const Transformation* component(number_t i) const { return components_[i];} //!< accessor to i-th component

    virtual void print(std::ostream&) const;           //!< print Transformation
    virtual void print(PrintStream& os) const {print(os.currentStream());}

    virtual const Translation* translation() const //! access to child Translation object (const)
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _translation)); return 0; }
    virtual Translation* translation() //! access to child Translation object
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _translation)); return 0; }

    virtual const Rotation2d* rotation2d() const //! access to child Rotation2d object (const)
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _rotation2d)); return 0; }
    virtual Rotation2d* rotation2d() //! access to child Rotation2d object
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _rotation2d)); return 0; }

    virtual const Rotation3d* rotation3d() const //! access to child Rotation3d object (const)
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _rotation3d)); return 0; }
    virtual Rotation3d* rotation3d() //! access to child Rotation3d object
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _rotation3d)); return 0; }

    virtual const Homothety* homothety() const //! access to child Homothety object (const)
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _homothety)); return 0; }
    virtual Homothety* homothety() //! access to child Homothety object
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _homothety)); return 0; }

    virtual const PointReflection* pointReflection() const //! access to child PointReflection object (const)
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _ptReflection)); return 0; }
    virtual PointReflection* pointReflection() //! access to child PointReflection object
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _ptReflection)); return 0; }

    virtual const Reflection2d* reflection2d() const //! access to child Reflection object (const)
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _reflection2d)); return 0; }
    virtual Reflection2d* reflection2d() //! access to child Reflection object
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _reflection2d)); return 0; }

    virtual const Reflection3d* reflection3d() const //! access to child Reflection3d object (const)
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _reflection3d)); return 0; }
    virtual Reflection3d* reflection3d() //! access to child Reflection3d object
      { error("bad_transform", name_, words("transform", transformType_), words("transform", _reflection3d)); return 0; }

    //! computes the image of a point by the transformation
    virtual Point apply(const Point& p) const;

    //----------------------------------
    // composition of 2 transformations
    //----------------------------------
    //! composition of transformations (general case)
    Transformation& operator*=(const Transformation& t);
    friend Transformation composeCompositeAndComposite(const Transformation& t1, const Transformation& t2);
    friend Transformation composeCompositeAndCanonical(const Transformation& t1, const Transformation& t2);
    friend Transformation composeCanonicalAndComposite(const Transformation& t1, const Transformation& t2);
    friend Transformation composeCanonicalAndCanonical(const Transformation& t1, const Transformation& t2);
    friend Transformation operator*(const Transformation& t1, const Transformation& t2);

    friend Transformation toComposite(const Transformation& t);
    friend std::ostream& operator<<(std::ostream&, const Transformation&);
};

//----------------------------------
// composition of 2 transformations
//----------------------------------
//! composition of composite transformations
Transformation composeCompositeAndComposite(const Transformation& t1, const Transformation& t2);
//! composition of a composite transformation and a canonical transformation
Transformation composeCompositeAndCanonical(const Transformation& t1, const Transformation& t2);
//! composition of a canonical transformation and a composite transformation
Transformation composeCanonicalAndComposite(const Transformation& t1, const Transformation& t2);
//! composition of canonical transformations
Transformation composeCanonicalAndCanonical(const Transformation& t1, const Transformation& t2);
Transformation operator*(const Transformation& t1, const Transformation& t2); //!< composition of transformations (general case)
Transformation toComposite(const Transformation& t); //!< return a composite Transformation with only one component : t
std::ostream& operator<<(std::ostream&, const Transformation&); //!< output Transformation

/*!
  \class Translation
  representation of a translation of vector u
*/
class Translation : public Transformation
{
  protected:
    std::vector<real_t> u_; //!< vector of the translation

  public:
    //! default constructor with a vector of real numbers
    Translation(std::vector<real_t> u = std::vector<real_t>(3,0.));
    //! constructor from 3 values
    Translation(real_t ux, real_t uy = 0., real_t uz = 0.);
    virtual ~Translation() {}

    std::vector<real_t> u() const { return u_; } //!< accessor to vector u

    virtual Transformation* clone() const { return new Translation(*this); } //!< virtual copy constructor

    //! access to child Translation object (const)
    virtual const Translation* translation() const {return this;}
    //! access to child Translation object
    virtual Translation* translation() {return this;}

    virtual Point apply(const Point& p) const;

    virtual void print(std::ostream&) const;           //!< print Transformation
    virtual void print(PrintStream& os) const {print(os.currentStream());}
};

/*!
  \class Rotation2d
  representation of a rotation from a point and an angle
*/
class Rotation2d : public Transformation
{
  protected:
    Point c_;    //!< center of the rotation
    real_t angle_; //!< angle of the rotation (in radians)

  public:
    //! default constructor with a vector of real numbers
    Rotation2d(const Point& c = Point(0.,0.), real_t angle = 0);
    //! destructor
    virtual ~Rotation2d() {}

    Point center() const { return c_; } //!< accessor to rotation center
    real_t angle() const { return angle_; } //!< accessor to rotation angle

    virtual Transformation* clone() const { return new Rotation2d(*this); } //!< virtual copy constructor

    //! access to child Rotation2d object (const)
    virtual const Rotation2d* rotation2d() const {return this;}
    //! access to child Rotation2d object
    virtual Rotation2d* rotation2d() {return this;}

    virtual Point apply(const Point& p) const;

    virtual void print(std::ostream&) const;           //!< print Transformation
    virtual void print(PrintStream& os) const {print(os.currentStream());}
};

/*!
  \class Rotation3d
  representation of a rotation from an axis ( a point and a vector ) and an angle
*/
class Rotation3d : public Transformation
{
  protected:
    Point c_;             //!< one point of the rotation axis
    std::vector<real_t> d_; //!< direction vector of the rotation axis
    real_t angle_;          //!< angle of the rotation (in radians)

  public:
    //! default constructor with a vector of real numbers
    Rotation3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.), real_t angle = 0);
    //! constructor with direction given with 2 reals
    Rotation3d(const Point& c, real_t dx, real_t dy, real_t angle);
    //! constructor with direction given with 3 reals
    Rotation3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
    //! destructor
    virtual ~Rotation3d() {}

    Point axisPoint() const { return c_; } //!< accessor to rotation axis point
    std::vector<real_t> axisDirection() const { return d_; } //!< accessor to rotation axis direction
    real_t angle() const { return angle_; } //!< accessor to rotation angle

    virtual Transformation* clone() const { return new Rotation3d(*this); } //!< virtual copy constructor

    //! access to child Rotation3d object (const)
    virtual const Rotation3d* rotation3d() const {return this;}
    //! access to child Rotation3d object
    virtual Rotation3d* rotation3d() {return this;}

    virtual Point apply(const Point& p) const;

    virtual void print(std::ostream&) const;           //!< print Transformation
    virtual void print(PrintStream& os) const {print(os.currentStream());}
};

/*!
  \class Homothety
  representation of an homothety from a center and a factor
*/
class Homothety : public Transformation
{
  protected:
    Point c_;     //!< center of the homothety
    real_t factor_; //!< scale factor of the homothety

  public:
    //! default constructor from a center and a factor
    Homothety(const Point& c = Point(0.,0.,0.), real_t factor = 1.);
    //! destructor
    virtual ~Homothety() {}

    Point center() const { return c_; } //!< returns the center of the homothety
    real_t factor() const { return factor_; } //!< returns the scale factor of the homothety

    virtual Transformation* clone() const { return new Homothety(*this); } //!< virtual copy constructor

    //! access to child Homothety object (const)
    virtual const Homothety* homothety() const {return this;}
    //! access to child Homothety object
    virtual Homothety* homothety() {return this;}

    virtual Point apply(const Point& p) const;

    virtual void print(std::ostream&) const;           //!< print Transformation
    virtual void print(PrintStream& os) const {print(os.currentStream());}
};

/*!
  \class PointReflection
  representation of a point symmetry
  it is an homothety of factor -1.
*/
class PointReflection : public Homothety
{
  public:
    //! constructor from a center
    PointReflection(const Point& c = Point(0.,0.,0.));
    //! destructor
    virtual ~PointReflection() {}

    virtual Transformation* clone() const { return new PointReflection(*this); } //!< virtual copy constructor

    //! access to child PointReflection object (const)
    virtual const PointReflection* pointReflection() const {return this;}
    //! access to child PointReflection object
    virtual PointReflection* pointReflection() {return this;}

    virtual Point apply(const Point& p) const;

    virtual void print(std::ostream&) const;           //!< print Transformation
    virtual void print(PrintStream& os) const {print(os.currentStream());}
};

/*!
  \class Reflection2d
  representation of a reflection
*/
class Reflection2d : public Transformation
{
  protected:
    Point c_;             //!< one point of the reflection axis
    std::vector<real_t> d_; //!< direction vector of the reflection axis
  public:
    //! constructor from a center and a direction vector
    Reflection2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
    //! constructor from a center and coordinates of the direction vector
    Reflection2d(const Point& c, real_t dx, real_t dy);
    //! destructor
    virtual ~Reflection2d() {}

    Point axisPoint() const { return c_; } //!< returns the point on the reflection axis
    std::vector<real_t> axisDirection() const { return d_; } //!< returns the direction vector of the reflection axis


    virtual Transformation* clone() const { return new Reflection2d(*this); } //!< virtual copy constructor

    //! access to child Reflection2d object (const)
    virtual const Reflection2d* reflection2d() const {return this;}
    //! access to child Reflection2d object
    virtual Reflection2d* reflection2d() {return this;}

    virtual Point apply(const Point& p) const;

    virtual void print(std::ostream&) const;           //!< print Transformation
    virtual void print(PrintStream& os) const {print(os.currentStream());}
};

/*!
  \class Reflection3d
  representation of a reflection on a plane
*/
class Reflection3d : public Transformation
{
  protected:
    Point c_;             //!< one point od the reflection plane
    std::vector<real_t> n_; //!< normal vector of the reflection plane
  public:
    //! constructor from a center and a normal vector of the plane
    Reflection3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
    //! constructor from a center and coordinates of the normal vector of the plane
    Reflection3d(const Point& c, real_t nx, real_t ny, real_t nz);
    virtual ~Reflection3d() {}

    Point planePoint() const { return c_; } //!< returns the point on the reflection plane
    std::vector<real_t> planeNormal() const { return n_; } //!< returns the normal vector of the reflection plane

    std::vector<real_t> u() const; //!< returns a vector "in" the plane
    std::vector<real_t> v() const; //!< returns another vector "in" the plane
    virtual Transformation* clone() const { return new Reflection3d(*this); } //!< virtual copy constructor

    //! access to child Reflection3d object (const)
    virtual const Reflection3d* reflection3d() const {return this;}
    //! access to child Reflection3d object
    virtual Reflection3d* reflection3d() {return this;}

    virtual Point apply(const Point& p) const;

    virtual void print(std::ostream&) const;           //!< print Transformation
    virtual void print(PrintStream& os) const {print(os.currentStream());}
};

//=============================================================
//  external functions to compute transformations on a Point
//=============================================================

//! apply a geometrical transformation on a Point
inline Point transform(const Point& p, const Transformation& t) { return t.apply(p); }
//! apply a translation on a Point (vector version)
Point translate(const Point& p, std::vector<real_t> u = std::vector<real_t>(3,0.));
//! apply a translation on a Point (3 reals version)
Point translate(const Point& p, real_t ux, real_t uy, real_t uz = 0.);
//! apply a rotation on a Point
Point rotate2d(const Point& p, const Point& c = Point(0.,0.), real_t angle = 0.);
//! apply a rotation on a Point
Point rotate3d(const Point& p, const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.), real_t angle = 0.);
//! apply a rotation on a Point
Point rotate3d(const Point& p, real_t dx, real_t dy, real_t angle);
//! apply a rotation on a Point
Point rotate3d(const Point& p, real_t dx, real_t dy, real_t dz, real_t angle);
//! apply a rotation on a Point
Point rotate3d(const Point& p, const Point& c, real_t dx, real_t dy, real_t angle);
//! apply a rotation on a Point
Point rotate3d(const Point& p, const Point& c, real_t dx, real_t dy, real_t dz, real_t angle);
//! apply a homothety on a Point
Point homothetize(const Point& p, const Point& c = Point(0.,0.,0.), real_t factor = 0.);
//! apply a homothety on a Point
Point homothetize(const Point& p, real_t factor);
//! apply a point reflection on a Point
Point pointReflect(const Point& p, const Point& c = Point(0.,0.,0.));
//! apply a reflection2d on a Point
Point reflect2d(const Point& p, const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.));
//! apply a reflection2d on a Point
Point reflect2d(const Point& p, const Point& c, real_t dx, real_t dy = 0.);
//! apply a reflection3d on a Point
Point reflect3d(const Point& p, const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.));
//! apply a reflection3d on a Point
Point reflect3d(const Point& p, const Point& c, real_t nx, real_t ny, real_t nz = 0.);

} // end of namespace xlifepp

#endif // TRANSFORMATIONS_HPP

