#include "xlife++.h"
using namespace xlifepp;
using namespace std;

// find = eta(x)
Real find(const Point & M, Parameters & pa = defaultParameters)
{
  Real res=1.;
  if(std::max(std::abs(M[0]),std::abs(M[1]))<0.5)
    res=std::exp(-((M[0]*M[0]-0.25)*(M[1]*M[1]-0.25))/(2.*0.05));
  return res;
}
Real eta2(const Point & M, Parameters & pa = defaultParameters)
{
  Real tmp=find(M);
  return tmp*tmp;
}
Complex g1(const Point& M, Parameters& pa = defaultParameters)
{
  Real k=real(pa("k"));
  Point d(1.,0.);
  return exp(i_*(k*dot(M,d)));
}

int main(int argc, char** argv)
{
  init(_lang=en);   // mandatory initialization of xlifepp
  verboseLevel(10);
  Real k=10.;
  //meshing
  Real hsize=(2*pi_/k)/15.;
  Square sp(_center=Point(0.,0.), _length=1., _hsteps=hsize, _domain_name="Omega", _side_names="Gamma");
  Mesh m1=Mesh(sp, _triangle, 1, _gmsh);
  Domain omega = m1.domain("Omega");
  Domain gamma = m1.domain("Gamma");
  theCout << "Mesh size = " << hsize << eol;
  theCout << "Number of Triangles = " << m1.nbOfElements() << eol;
  //defining parameter and kernel
  Parameters pars;
  pars << Parameter(k,"k");
  Vector<Real> nv(2);
  pars << Parameter(&nv,"_n");
  Kernel G=Helmholtz2dKernel(pars);
  Function finc(g1,pars);
  //defining space, unknown and test function
  Space V1(omega,P1,"V1",false);
  Space V0(gamma,P1,"V0",false);
  Unknown u1(V1,"u1"); TestFunction v1(u1,"v1");
  Unknown l0(V0,"l0"); TestFunction lt0(l0,"lt0");
  theCout<<"Nb dofs BEM= " << V0.nbDofs() << " Nb dofs FEM= " << V1.nbDofs() << eol;
  //defining bilinear and linear form
  IntegrationMethods ims(Duffy, 15, 0., _defaultRule, 12, 1., _defaultRule, 10, 2., _defaultRule, 8);
  BilinearForm blf=intg(omega,grad(u1)|grad(v1))-k*k*intg(omega,eta2*u1*v1)
                  - intg(gamma,l0*v1) + 0.5*intg(gamma,u1*lt0)
                  - intg(gamma,gamma,u1*ndotgrad_y(G)*lt0,ims)
                  + intg(gamma,gamma,l0*G*lt0,ims);
  LinearForm lf=intg(gamma,finc*lt0);
  //computing FEM/BEM matrix and right hand side vector
  TermMatrix lhs(blf,"lhs");
  TermVector rhs(lf);
  //solving linear system using direct method
  TermVector sol=directSolve(lhs,rhs);

  // Representing the solution FEM and BEM
  Square Sint(_center=Point(0.,0.), _length=1, _hsteps=hsize, _domain_name="S_int");
  Square Sext(_center=Point(0.,0.), _length=3, _hsteps=1.5*hsize, _domain_name="S_ext");
  Mesh mrep(Sext+Sint, _triangle, 1, _gmsh);
  Domain S_ext=mrep.domain("S_ext"), S_int=mrep.domain("S_int");
  Domain S=merge(S_ext,S_int,"S");
  Space Vrep(S,P1,"Vrep",false);
  Unknown ur(Vrep,"ur");
  Function Find(find,pars);
  TermVector findex(ur,S,Find);
  saveToFile("findex",findex,_vtu); // Representing eta
  TermVector Uint=interpolate(ur,S_int,sol(u1)); // FEM solution (total field)
  saveToFile("Uint",Uint,_vtu);

  // Representing of the BEM part
  IntegrationMethods imr(_GaussLegendreRule, 20, 1., _GaussLegendreRule, 10, 2., _GaussLegendreRule,5);
  TermVector Uext =
      - integralRepresentation(ur, S_ext, intg(gamma,G*sol(l0),imr))
      + integralRepresentation(ur, S_ext, intg(gamma,ndotgrad_y(G)*sol(u1),imr));

  TermVector Uinc(ur,S_ext,finc);
  saveToFile("Uinc",Uinc,_vtu); // Incident field
  saveToFile("Uext",Uext,_vtu); // scattered field in exterior domain
  TermVector Uext_t = Uext + Uinc;
  saveToFile("Uext_t",Uext_t,_vtu); // Total field in exterior domain
  TermVector U=merge(Uint,Uext_t); // Merged FEM and BEM solutions
  saveToFile("U",U,_vtu);
  theCout << "Program finished" << eol;
  return 0;
}

